"""
@file box.py
Box-shaped domains definition.

"""
from domain import Domain
from parmepy.constants import np, PARMES_REAL, PARMES_INTEGER, PERIODIC, \
    debug


class Box(Domain):
    """
    Box-shaped domain description.
    @note FP : BC as parameter may be better?
    @todo Have different BC
    """

    @debug
    def __init__(self, dimension=3, length=[1.0, 1.0, 1.0],
                 origin=[0., 0., 0.]):
        """
        Create a Periodic Box from a dimension, length and origin.
        Parameters dimensions must coincide. Raise ValueError
        in case of inconsistent parameters dimensions.

        By defaults, it creates a \f$[0;1]^3\f$ Box.

        @param dimension : Box dimension. Default: 3
        @param length : Box length. Default [1.0, 1.0, 1.0]
        @param origin : Box minimum position. Default [0., 0., 0.]
        \code
        >>> import parmepy as pp
        >>> import numpy as np
        >>> b = pp.Box()
        >>> (b.max == np.asarray([1.0, 1.0, 1.0])).all()
        True

        \endcode
        """
        if not (dimension == len(length) and dimension == len(origin)):
            raise ValueError("Box parameters inconsistents dimensions")
        Domain.__init__(self, dimension)
        ##  Box length.
        self.length = np.asarray(length, dtype=PARMES_REAL)
        ##  Box origin
        self.origin = np.asarray(origin, dtype=PARMES_REAL)
        ## Maximum Box position. max = origin + length
        self.max = self.origin + self.length
        ## Boundary conditions type
        self.boundaries = np.zeros((self.dimension), dtype=PARMES_INTEGER)
        self.boundaries[:] = PERIODIC

    def __str__(self):
        """
        Informations display.
        @return Informations
        """
        s = str(self.dimension) + \
            "D box (parallelepipedic or rectangular) domain : \n"

        s += "   origin : " + str(self.origin) + ", maxPosition :" \
             + str(self.max) + ", lengths :" + str(self.length) + "."
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Box."
    print Box.__doc__
