"""
Cylinder obstacle description
"""
from parmepy.domain.obstacle.obstacle import Obstacle
import math
import numpy as np


class Cylinder(Obstacle2D):
    """
    Discrete obstacles representation.
    """

    def __init__(self, obst, radius=0.2):
        """
        Creates the cylinder obsctacle and y-boundaries and 
        returns 3 arrays corresponding to the characteristic 
        functions of three different porous media.

        @param obstacle2D : Two dimensional obstacle description.
        @param radius : Cylinder radius
        """
        ## 2D parent obstacle
        self.obst = obst
        ## Radius of the cylinder
        self.radius = radius
        ## Characteristic function (array) for y-boundaries
        self.chiBoundary = None
        ## Characteristic function (array) for the solid
        self.chiSolid = None
        ## Characteristic function (array) for the porous area
        self.chiPorous = None
        print 'obstacle =', self.obst.obstacleName

    def setUp(self, topology):
        """
        Compute the characteristic functions associated 
        to y-boundaries and cylinder
        """
        ## Temporary arrays
        chiBoundary_i = []
        chiBoundary_j = []
        chiSolid_i = []
        chiSolid_j = []
        chiPorous_i = []
        chiPorous_j = []

        step = topology.mesh.space_step
        ghosts = topology.ghosts
        local_start = topology.mesh.local_start
        local_end = topology.mesh.local_end
        coord_start = topology.mesh.origin + (ghosts * step)
        coord_end = topology.mesh.end - (ghosts * step)
        layerMin = coord_start[1] + self.obst.zlayer
        layerMax = coord_end[1] - self.obst.zlayer
        if not (self.obst.porousLayerThickn <= self.radius):
            raise ValueError("Error, porous layer thickness" +
                             "is higher than cylinder radius.")
        radiusMinuslayer = self.radius- self.obst.porousLayerThickn

        print 'step, ghosts, local_start, local_end, zlayer', \
              step, ghosts, local_start, local_end, self.obst.zlayer
        print 'start, end, layerMin, layerMax, radiusMinuslayer', \
              coord_start, coord_end, layerMin, layerMax, radiusMinuslayer

        for j in xrange (local_start[1], local_end[1] + 1):
            cy = coord_start[1] + j * step[1]
            for i in xrange (local_start[0], local_end[0] + 1):
                if (cy >= layerMax or cy <= layerMin):
                # we are in the y-layer boundary:
                    chiBoundary_i.append(i)
                    chiBoundary_j.append(j)
                else :
                    cx = coord_start[0] + i * step[0]
                    dist = np.sqrt((cx - self.obst.center[0]) ** 2 +
                                   (cy - self.obst.center[1]) ** 2)
                    if (radiusMinuslayer < dist 
                        and dist <= self.radius + 1E-12 
                        and self.obst.porousLayerThickn != 0.):
                        # we are in the porous region of the cylinder:
                        chiPorous_i.append(i)
                        chiPorous_j.append(j)
                    if (dist <= radiusMinuslayer + 1E-12):
                        # we are in the solid region of the cylinder:
                        chiSolid_i.append(i)
                        chiSolid_j.append(j)

        ## Characteristic function of penalized boundaries
        chiBoundary_i = np.asarray(chiBoundary_i, dtype=PARMES_INTEGER)
        chiBoundary_j = np.asarray(chiBoundary_j, dtype=PARMES_INTEGER)
        self.chiBoundary = tuple([chiBoundary_i, chiBoundary_j])

        ## Characteristic function of solid areas
        chiSolid_i = np.asarray(chiSolid_i, dtype=PARMES_INTEGER)
        chiSolid_j = np.asarray(chiSolid_j, dtype=PARMES_INTEGER)
        self.chiSolid = tuple([chiSolid_i, chiSolid_j])

        ## Characteristic function of porous areas
        chiPorous_i = np.asarray(chiPorous_i, dtype=PARMES_INTEGER)
        chiPorous_j = np.asarray(chiPorous_j, dtype=PARMES_INTEGER)
        self.chiPorous = tuple([chiPorous_i, chiPorous_j])

    def __str__(self):
        """ToString method"""
        return "Cylinder"

if __name__ == "__main__" :
    print "This module defines the following classe:"
    print "Cylinder: ", Cylinder.__doc__
