"""
@file cylinder2d.py
Rigid disk.
"""
from parmepy.domain.obstacle.sphere import Sphere, HemiSphere
import math
import numpy as np


class Cylinder2D(Sphere):
    """
    Disk in a 2D domain.
    """

    def __init__(self, domain, position, radius=1.0, vd=0.0, porousLayers=[]):
        """
        Description of a disk in a domain.
        @param domain : the physical domain that contains the sphere.
        @param position : position of the center
        @param radius : sphere radius, default = 1
        @param porousLayers : a list of thicknesses
        for successive porous layers
        radius is the inside sphere radius and thicknesses are given from
        inside layer to outside one.
        @param vd : velocity of the disk (considered as a rigid body),
        default = 0.
        """
        Sphere.__init__(self, domain, position, radius, vd, porousLayers)
        assert self.domain.dimension == 2

        def dist(x, y, R):
            """
            """
            return math.sqrt((x - self.position[0]) ** 2
                             + (y - self.position[1]) ** 2) - R
        self.chi = [np.vectorize(dist)]
        ## List of thicknesses for porous layers

    ## def discretize(self, topo):
    ##     # first check if we have already compute indices for
    ##     # this topology
    ##     if topo not in self.ind.keys():
    ##         currentRadius = self.radius
    ##         self.ind[topo] = []
    ##         # for each indicator function
    ##         for thickness in self.layers:
    ##             # apply indicator function on topo local mesh
    ##             args = (currentRadius,)
    ##             condA = self.chi[0](*(topo.mesh.coords + args)) <= 0
    ##             args = (currentRadius - thickness,)
    ##             condB = self.chi[0](*(topo.mesh.coords + args)) > 0
    ##             self.ind[topo].append(np.logical_and(condA, condB))
    ##             # update current radius
    ##             currentRadius = currentRadius - thickness
    ##         # and finally the 'internal' sphere
    ##         args = (currentRadius,)
    ##         self.ind[topo].append(self.chi[0](*(topo.mesh.coords + args)) <= 0)
    ##     return self.ind[topo]

    def __str__(self):
        s = '2D Cylinder of radius ' + str(self.radius)
        s += ' and center position ' + str(self.position)
        return s


class SemiCylinder2D(HemiSphere):
    """
    Half disk in a 2D domain.
    """
    def __init__(self, domain, position, radius=1.0, vd=0.0, porousLayers=[]):
        """
        Constructor for the semi-disk.
        @param domain : the physical domain that contains the sphere.
        @param position : position of the center
        @param radius : sphere radius, default = 1
        (if box ...)
        @param vd : velocity of the disk (considered as a rigid body),
        default = 0.
        """
        HemiSphere.__init__(self, domain, position, radius, vd, porousLayers)
        assert self.domain.dimension == 2

        def dist(x, y, R):
            """
            """
            return math.sqrt((x - self.position[0]) ** 2
                             + (y - self.position[1]) ** 2) - R
        self.chi = [np.vectorize(dist)]

        def LeftBox(x, y):
            return x - self.position[0]

        self.LeftBox = np.vectorize(LeftBox)

    ## def discretize(self, topo):
    ##     # first check if we have already compute indices for
    ##     # this topology
    ##     if topo not in self.ind.keys():
    ##         currentRadius = self.radius
    ##         self.ind[topo] = []
    ##         # First cond : we must be in the left half plane
    ##         cond0 = self.HalfPlane(*(topo.mesh.coords)) <= 0
    ##         # for each indicator function
    ##         for thickness in self.layers:
    ##             # apply indicator function on topo local mesh
    ##             args = (currentRadius,)
    ##             condA = self.chi[0](*(topo.mesh.coords + args)) <= 0
    ##             args = (currentRadius - thickness,)
    ##             condB = self.chi[0](*(topo.mesh.coords + args)) > 0
    ##             np.logical_and(condA, condB, condA)
    ##             np.logical_and(condA, cond0, condA)
    ##             self.ind[topo].append(condA)
    ##             # update current radius
    ##             currentRadius = currentRadius - thickness
    ##         # and finally the 'internal' sphere
    ##         args = (currentRadius,)
    ##         condA = self.chi[0](*(topo.mesh.coords + args)) <= 0
    ##         self.ind[topo].append(np.logical_and(condA, cond0))
    ##     return self.ind[topo]

    def __str__(self):
        s = '2D semi-cylinder of radius ' + str(self.radius)
        s += ' and center position ' + str(self.position)
        return s

if __name__ == "__main__":
    print "This module defines the following classe:"
    print "Disk: ", Cylinder2D.__doc__
    print "Half-disk: ", SemiCylinder2D.__doc__
