"""
@file sphere.py
Plate-like sub-domains.
"""
from parmepy.domain.obstacle.obstacle import Obstacle
import numpy as np


class Plates(Obstacle):
    """
    Top and down (z-axis) plate-like sub-domains.
    Defines two areas on the min and max z position of
    the main domain.
    zplates:
    \f$ \{x,y,z\} \ for \ z_{max} - \epsilon \leq z \leq z_{max} + \epsilon
    \ or \ z_{min} - \epsilon \leq z \leq z_{min}\f$
    """

    def __init__(self, domain, normal_dir, epsilon=0.1, vd=0.0):
        """
        Description of a sphere in a domain.
        @param domain : the physical domain that contains the sphere.
        @param epsilon : thickness/2 of the plates
        @param vd : velocity of obstacle (considered as a rigid body),
        default = 0.
        """
        Obstacle.__init__(self, domain, vd=vd)
        assert epsilon > 0.0, 'Plate thickness must be positive'
        ## Thickness/2
        self.epsilon = epsilon
        ## Direction of the normal to the plate (0:x, 1:y, 2:z))
        self.normal = normal_dir
        assert self.normal < domain.dimension and self.normal >= 0

        def Upper(*args):
            return domain.max[self.normal] - self.epsilon - args[self.normal]

        def Lower(*args):
            return args[self.normal] - self.epsilon - \
                domain.origin[self.normal]

        self.Upper = np.vectorize(Upper)
        self.Lower = np.vectorize(Lower)

    def discretize(self, topo):
        # first check if we have already compute indices for
        # this topology

        if topo not in self.ind.keys():
            self.ind[topo] = []
            # apply indicator function on topo local mesh
            condUpper = self.Upper(*topo.mesh.coords) <= 0
            condLower = self.Lower(*topo.mesh.coords) <= 0
            self.ind[topo].append(np.logical_or(condUpper, condLower))

        return self.ind[topo]

    def __str__(self):
        """ToString method"""
        s = 'Upper and lower (z-axis) plane areas of thickness '
        s += str(2 * self.epsilon) + '.'
        return s

