!> @file fftw2py.f90 
!! Fortran to python interface file.

!> Interface to mpi-fftw (fortran) utilities
module fftw2py

  use parmesparam
  !> 2d case
   use fft2d
  !> 3d case
  use fft3d
  use mpi
  implicit none

contains
  
  !> Initialisation of fftw context : create plans and memory buffers
  !! @param[in] resolution global resolution of the discrete domain
  !! @param[in] lengths width of each side of the domain
  !! @param[out] datashape local dimension of the input/output field
  !! @param[out] offset absolute index of the first component of the local field
  subroutine init_fftw_solver(resolution,lengths,datashape,offset,dim,fftw_type_real)

    integer, intent(in) :: dim 
    integer, dimension(dim),intent(in) :: resolution
    real(pk),dimension(dim), intent(in) :: lengths
    integer(ik), dimension(dim), intent(out) :: datashape
    integer(ik), dimension(dim), intent(out) :: offset
    logical, optional :: fftw_type_real
    !f2py optional :: dim=len(resolution)
    !f2py intent(hide) dim
    !f2py logical optional, intent(in) :: fftw_type_real = 1 
    
    if(fftw_type_real) then
       if(dim == 2) then 
          print*, "Init fftw/poisson solver for a 2d problem"
          call init_r2c_2d(resolution,lengths)
       else
          print*, "Init fftw/poisson solver for a 3d problem"
          call init_r2c_3d(resolution,lengths)
       end if
    else
       if(dim == 2) then 
          print*, "Init fftw/poisson solver for a 2d problem"
          call init_c2c_2d(resolution,lengths)
       else
          print*, "Init fftw/poisson solver for a 2d problem"
          call init_c2c_3d(resolution,lengths)
       end if
    end if

    if(dim==2) then
       call getParamatersTopologyFFTW2d(datashape,offset)
    else
       call getParamatersTopologyFFTW3d(datashape,offset)
    end if
  end subroutine init_fftw_solver

  !> Free memory allocated for fftw-related objects (plans and buffers)
  subroutine clean_fftw_solver(dim)

    integer, intent(in) :: dim 
    if(dim == 2) then     
       call cleanFFTW_2d()
    else
       call cleanFFTW_3d()
    end if
  end subroutine clean_fftw_solver
  
  !> Solve 
  !! \f[ \nabla (\nabla \times velocity) = - \omega \f]
  !! velocity being a 2D vector field and omega a 2D scalar field. 
  subroutine solve_poisson_2d(omega,velocity_x,velocity_y)
    real(pk),dimension(:,:),intent(in):: omega
    real(pk),dimension(size(omega,1),size(omega,2)),intent(out) :: velocity_x,velocity_y
    real(pk) :: start
    !f2py intent(in,out) :: velocity_x,velocity_y
    start = MPI_WTime()

    call r2c_2d(omega)
    
    call filter_poisson_2d()

    call c2r_2d(velocity_x,velocity_y)
    print *, "fortran resolution time : ", MPI_WTime() - start
  
  end subroutine solve_poisson_2d

  !> Solve 
  !! \f{eqnarray*} \frac{\partial \omega}{\partial t} &=& \nu \Delta \omega \f}
  !! omega being a 2D scalar field. 
  subroutine solve_diffusion_2d(nudt, omega)
    real(pk), intent(in) :: nudt
    real(pk),dimension(:,:),intent(inout):: omega
    !f2py intent(in,out) :: omega

    call r2c_2d(omega)
    
    call filter_diffusion_2d(nudt)

    call c2r_scalar_2d(omega)
  
  end subroutine solve_diffusion_2d

  !> Solve 
  !! \f{eqnarray*} \Delta \psi &=& - \omega \\ velocity = \nabla\times\psi \f}
  !! velocity and omega being 3D vector fields.
  subroutine solve_poisson_3d(omega_x,omega_y,omega_z,velocity_x,velocity_y,velocity_z)
    real(pk),dimension(:,:,:),intent(in):: omega_x,omega_y,omega_z
    real(pk),dimension(size(omega_x,1),size(omega_y,2),size(omega_z,3)),intent(out) :: velocity_x,velocity_y,velocity_z
    real(pk) :: start
    !f2py intent(in,out) :: velocity_x,velocity_y,velocity_z    
    start = MPI_WTime()
    call r2c_3d(omega_x,omega_y,omega_z)
    
    call filter_poisson_3d()
    
    call c2r_3d(velocity_x,velocity_y,velocity_z)
    print *, "fortran resolution time : ", MPI_WTime() - start

  end subroutine solve_poisson_3d

  !> Solve 
  !! \f{eqnarray*} \Delta \psi &=& - \omega \\ velocity = \nabla\times\psi \f}
  !! velocity being a 2D complex vector field and omega a 2D complex scalar field. 
  subroutine solve_poisson_2d_c(omega,velocity_x,velocity_y)
    complex(pk),dimension(:,:),intent(in):: omega
    complex(pk),dimension(size(omega,1),size(omega,2)),intent(out) :: velocity_x,velocity_y
    !f2py intent(in,out) :: velocity_x,velocity_y

    call c2c_2d(omega,velocity_x,velocity_y)
      
  end subroutine solve_poisson_2d_c

  !> Solve 
  !!  \f{eqnarray*} \Delta \psi &=& - \omega \\ velocity = \nabla\times\psi \f}
  !! velocity and omega being 3D complex vector fields.
  subroutine solve_poisson_3d_c(omega_x,omega_y,omega_z,velocity_x,velocity_y,velocity_Z)
    complex(pk),dimension(:,:,:),intent(in):: omega_x,omega_y,omega_z
    complex(pk),dimension(size(omega_x,1),size(omega_y,2),size(omega_z,3)),intent(out) :: velocity_x,velocity_y,velocity_z
    !f2py intent(in,out) :: velocity_x,velocity_y,velocity_z

    call c2c_3d(omega_x,omega_y,omega_z,velocity_x,velocity_y,velocity_z)
      
  end subroutine solve_poisson_3d_c

  !> Solve 
  !! \f{eqnarray*} \omega &=& \nabla \times v \\ \frac{\partial \omega}{\partial t} &=& \nu \Delta \omega \f}
  !! velocity and omega being 3D vector fields.
  subroutine solve_curl_diffusion_3d(nudt,velocity_x,velocity_y,velocity_z,omega_x,omega_y,omega_z)
    real(pk), intent(in) :: nudt
    real(pk),dimension(:,:,:),intent(in):: velocity_x,velocity_y,velocity_z
    real(pk),dimension(size(velocity_x,1),size(velocity_y,2),size(velocity_z,3)),intent(out) :: omega_x,omega_y,omega_z
    !f2py intent(in,out) :: omega_x,omega_y,omega_z
    
    call r2c_3d(velocity_x,velocity_y,velocity_z)
    
    call filter_curl_diffusion_3d(nudt)
    
    call c2r_3d(omega_x,omega_y,omega_z)

  end subroutine solve_curl_diffusion_3d

  !> Solve 
  !! \f{eqnarray*} \frac{\partial \omega}{\partial t} &=& \nu \Delta \omega \f}
  !! omega being 3D vector field.
  subroutine solve_diffusion_3d(nudt,omega_x,omega_y,omega_z)
    real(pk), intent(in) :: nudt
    real(pk),dimension(:,:,:),intent(inout):: omega_x,omega_y,omega_z
    !f2py intent(in,out) :: omega_x,omega_y,omega_z
    
    call r2c_3d(omega_x,omega_y,omega_z)
    
    call filter_diffusion_3d(nudt)
    
    call c2r_3d(omega_x,omega_y,omega_z)

  end subroutine solve_diffusion_3d

  !> Perform solenoidal projection to ensure divergence free vorticity field
  !! \f{eqnarray*} \omega ' &=& \omega - \nabla\pi \f}
  !! omega being a 3D vector field.
  subroutine projection_om_3d(omega_x,omega_y,omega_z)
    real(pk),dimension(:,:,:),intent(inout):: omega_x,omega_y,omega_z
    !f2py intent(in,out) :: omega_x,omega_y,omega_z

    call r2c_3d(omega_x,omega_y,omega_z)
    
    call filter_projection_om_3d()
    
    call c2r_3d(omega_x,omega_y,omega_z)

  end subroutine projection_om_3d

  !> Projects vorticity values from fine to coarse grid :
  !! @param[in] dxf, dyf, dzf: grid filter size = domainLength/(CoarseRes-1)
  !! in the following, omega is the 3D vorticity vector field.
  subroutine multires_om_3d(dxf, dyf, dzf, omega_x,omega_y,omega_z)
    real(pk), intent(in) :: dxf, dyf, dzf
    real(pk),dimension(:,:,:),intent(inout):: omega_x,omega_y,omega_z
    !f2py intent(in,out) :: omega_x,omega_y,omega_z

    call r2c_3d(omega_x,omega_y,omega_z)
    
    call filter_multires_om_3d(dxf, dyf, dzf)
    
    call c2r_3d(omega_x,omega_y,omega_z)

  end subroutine multires_om_3d

end module fftw2py
