## @package parmepy.fields
# Tools to build and use 1, 2 or 3D fields (scalar or vector)
#
# Fields are used to described the physical continuous variables involved in
# a problem.
# Each field is defined on a physical domain and may have one or more
# discretisations.
# A user-defined functions may also be used to describe the field.
#
# \section fieldsTuto Fields usage tutorial :
#
# \subsection fieldsTutoContinuous Continuous fields description
#
# \b Classes : parmepy.fields.continuous.Field,
# parmepy.fields.analytical.AnalyticalField
#
# First of all, one must defined (described) the continuous fields
# involved in the problem:
# \code
# from parmepy.domain.box import Box
# from parmepy.fields.continuous import Field
#
# # First, the domain : a 3D box
# dom = Box()
# # A scalar field on this domain
# rho = Field(dom)
# # A vector field
# velocity = Field(dom, isVector=True)
# # A vector field, named "reference", defined by a function
# # 'compute_ref'
# def compute_ref(x,y,z):
#     valx = sin(x)
#     valy = sin(y)
#     valz = sin(z)
#     return valx, valy, valz
#
# vorticity = Field(dom, isVector=True, name='reference,
#     formula=compute_ref)
#
# # Compute the value of vorticity at point of coordinates (1,2,0) :
# wx, wy, wz = vorticity.value(1.,2.,0.)
#
# # Modify the formula and add some extra arguments:
# 
# def compute_ref2(x,y,z,t,dt):
#     valx = sin(x)*t
#     valy = sin(y) + dt
#     valz = sin(z)
#     return valx, valy, valz
#
# vorticity.setFormula(compute_ref2)
#
# t = 4
# dt = 8.3
# # Compute vorticity using compute_ref2 for all the available
# # discretizations.
# # Warning : space coordinates are implicit but all extra-parameters
# # must be explicitly given, in the same order as defined in compute_ref2.
# # No keywords arguments allowed.
# vorticity.initialize(t,dt)
# \endcode
#
# Up to now, these are just descriptions of the fields. No discretisations, no
# memory allocations have been done for any data. We have just said that
# we will use one scalar field, rho, and two vectors fields velocity and
# vorticity, the last one defined as some sinus of the space coordinates.
#
# \subsection fieldsTutoDiscretisation Discretisation of the fields
# \b classes : parmepy.fields.scalar.ScalarField,
# parmepy.fields.vector.VectorField
#
# These fields must now be discretized, which means
# that memory must be allocated for each required resolution.
# Since the problem may be distributed among mpi process, the
# basic object used to described the resolution is a
# parmepy.mpi.topology.Cartesian, which describes the mpi process
# distribution and the local meshes resolution on each process.
#
# As an example, suppose that we have a first problem with
# vorticity and velocity at a 'coarse' resolution (r1) and another
# one with a more precise resolution (r2), involving velocity and rho.
# It means that vorticity must be discretized with r1, rho with r2 and
# velocity with r1 and r2
# \code
# # First we set two global resolutions :
# r1 = [64, 64, 64]
# r2 = [257, 257, 257]
# # Then we create two topologies, one for each global resolution
# from parmepy.mpi.topology import Cartesian
# topo1 = Cartesian(dom, 3, r1)
# topo2 = Cartesian(dom, 3, r2)
# # Now  let us discretize the fields:
# rho.discretize(topo2)
# vorticity.discretize(topo1)
# v1 = velocity.discretize(topo1)
# v2 = velocity.discretize(topo2)
# \endcode
# This means that :
# @li both rho and vorticity have one discrete field
# that contains numpy arrays of the proper size
# (depends of the number of mpi process)
# @li velocity has two discrete fields, accessed in the example with v1 and v2.
#
# \subsection fieldsTutoDataAccess Fields' data access
#
# Some examples on how to get access or set values of the fields:
#
# \code
# # Get a discretisation for a given continuous field:
# v3 = velocity.discreteFields[topo1]
# # Note that in this case v3 == v1
#
# # Get all the discretisations :
# for topo in velocity.discreteFields.keys()
#     print velocity.discreteFields[topo]
#
# # Get second component value of v2 at point of index 8,8,8
# # in the local mesh of topo2:
# print v2[1][8,8,8]
# # use vorticity formula to initialize v1 values :
# v1.initialize(formula=compute_ref)
# # norm 2 of v1
# nx, ny, nz = v1.norm()
# \endcode
#
#
# \b Remark : usually, in a 'complete' problem, the user does not have
# to bother with the topologies' stuff and the fields discretisations.
# All this is hidden from user : one just has to define
# some operators, bricks of the global problem, associated with
# some required global resolutions and some solving methods.
# Then the setup process of the problem/operators will automatically
# performs the topologies creation and the discretisation of the fields.
# See operators/problem documentation for more details or
# the examples in Parmes/Examples.
#
#
# To be continued ...
