"""
@file fields/continuous.py

Continuous variable description.
"""
from parmepy.constants import debug
from parmepy.fields.discrete import DiscreteField
#from parmepy.fields.vector import VectorField
from parmepy.mpi import main_rank


class Field(object):
    """ Continuous field defined on a physical domain.

    This object handles a dictionnary of discrete fields
    (from 0 to any number).
    Each discrete field is uniquely defined by the topology used to
    discretize it.

    Example :
    if topo1 and topo2 are two parmepy.mpi.topology.Cartesian objects:
    \code
    scal = Field(domain=dom, name='Scalar')
    # Discretize scal on two different topologies
    scal.discretize(topo1)
    scal.discretize(topo2)
    # Access to the discrete fields :
    scal_discr1 = scal.discreteFields[topo1]
    scal_discr2 = scal.discreteFields[topo2]
    # or in a more  straightforward way:
    scal_discr2 = scal.discretize(topo2)
    \endcode

    """

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, domain, formula=None, name=None, isVector=False,
                 nbComponents=None):
        """
        Create a scalar of vector field which dimension is determined
        by its domain of definition.

        @param domain : physical domain where this field is defined.
        @param formula : a user-defined python function
        (default = 0 all over the domain)
        @param name : name of the variable.
        @param isVector : true if the field is a vector.
        """
        ## Physical domain of definition of the variable.
        self.domain = domain
        ## dimension (1, 2 or 3D), equal to domain dimension.
        self.dimension = self.domain.dimension
        ## Id (optional)
        if(name is not None):
            self.name = name
        else:
            self.name = 'unamed'
        ## Dictionnary of all the discretizations of this field.
        ## Key = parmepy.mpi.topology.Cartesian,
        ## value =parmepy.fields.discrete.DiscreteField.
        ## Example : vel = discreteFields[topo]
        self.discreteFields = {}
        ## Is this field a vector field?
        self.isVector = isVector
        ## Analytic formula.
        self._formula = formula
        ## A list of extra parameters, used in _formula
        self.extraParameters = tuple([])
        ## Number of components of the field
        if nbComponents is None:
            self.nbComponents = self.dimension if self.isVector else 1
        else:
            self.nbComponents = nbComponents

    @debug
    def discretize(self, topo):
        """
        Discretization of the field on a topology.
        In the case of this field already discretized whith the same topology,
        a discrete object is not created.

        @param topo : topology specification for discretization
        @return a discrete field
        """

        if topo in self.discreteFields.keys():
            return self.discreteFields[topo]
        else:
            nameD = self.name + '_' + str(topo.getId())
            self.discreteFields[topo] = DiscreteField(
                topo,
                isVector=self.isVector,
                name=nameD,
                nbComponents=self.nbComponents)

        return self.discreteFields[topo]

    def setFormula(self, formula):
        """
        Set (or reset) a user-defined function to compute field
        values.
        @param formula : a user-defined python function
        Warning: if set, this formula will
        overwrite the (optional) formula given during
        construction of this field.

        """
        self._formula = formula

    @debug
    def initialize(self, *args):
        """
        Initialize all the discrete fields associated to this continuous field.
        using the formula set during construction or with setFormula method.
        If formula is not set, field values are set to zero.

        @param *args : see parmepy.fields
        """
        for dF in self.discreteFields.values():
            dF.initialize(self._formula, *(args + self.extraParameters))

    def value(self, *pos):
        """
        Evaluation of the field at a given position, according
        to the analytic formula given during construction.

        @param pos : Position to evaluate.
        @return formula(pos): value of the formula at the given position.
        """
        if self._formula is not None:
            return self._formula(*pos)
        else:
            return 0.0

    def __str__(self):
        """ Field info display """

        s = '[' + str(main_rank) + '] " ' + self.name + ' ", '
        s += str(self.dimension) + 'D '
        if self._formula is not None:
            s += 'an analytic formula is associated to this field:'
            s += str(self._formula.func_name)
        if self.isVector:
            s += 'vector field '
        else:
            s += 'scalar field '
        s += 'with the following (local) discretisations :\n'
        for f in self.discreteFields.values():
            s += f.__str__(True)
        return s

    def norm(self, topo=None):
        """
        Compute the p-norm of the discretisation of the
        current field corresponding to topology topo (input arg).
        Default = norm-2.
        @param topo : the topology corresponding to the
        required discretisation. If topo == None, returns
        the norm of the first discretization.
        @return norm of the field.
        """
        if topo is not None:
            return self.discreteFields[topo].norm()
        else:
            return self.discreteFields.values()[0].norm()

    def setExtraParameters(self, *args):
        """
        Set values for (optional) list of extra parameters
        that may be required in formula.
        @param params : a list of parameters to set.
        """
        self.extraParameters = args

    def dump(self, filename, topo=None, mode=None):
        """
        Dump (serialize) the data of the field into filename.
        serialization process.
        @param filename : name of the file in which data are serialized
        @param topo : topology that identify a discrete field to be saved.
        @param mode : set mode='append' to add data to an existing file.
        if None, the first discrete field in the list will be saved.
        """
        if topo is not None:
            assert topo in self.discreteFields.keys()
            self.discreteFields[topo].dump(filename, mode)
        else:
            # dump all discr or only the first one?
            self.discreteFields.values()[0].dump(filename, mode)

    def load(self, filename, topo=None, fieldname=None):
        """
        load data from a file build
        with cPickle serialization.
        @param filename : name of the file in which data are serialized
        @param topo : topology that identify a discrete field to be saved.
        if None, the first discrete field in the list will be saved.
        """
        if topo is not None:
            assert topo in self.discreteFields.keys()
            self.discreteFields[topo].load(filename, fieldname=fieldname)
        else:
            # load all discr or only the first one?
            self.discreteFields.values()[0].load(filename, fieldname=fieldname)

    def zero(self, topo=None):
        """
        reset to 0.0 all components of this field.
        @param topo : if given, only discreteFields[topo] is
        set to zero
         """
        if topo is not None:
            self.discreteFields[topo].zero()
        else:
            for dfield in self.discreteFields.values():
                dfield.zero()

if __name__ == "__main__":
    print __doc__
    print "- Provided class : ContinuousField"
    print Field.__doc__
