"""
@file fields/discrete.py

Discrete fields (scalars or vectors) descriptions.
"""
from parmepy.constants import debug
from scitools.NumPyDB import NumPyDB_cPickle
from itertools import count
from parmepy.mpi import main_rank
from parmepy.tools.timers import Timer, timed_function
import numpy as np
from parmepy.constants import ORDER, PARMES_REAL
from numpy import linalg as la
from parmepy.mpi import main_comm


class DiscreteField(object):
    """Interface for discrete fields (scalar or vector).

    A discrete field represents the data layout on each mpi process for
    a given global resolution.

    More precisely a discrete field :
    - is the discrete representation of a continuous field for some
    local resolution given by a parmepy.mpi.topology.Cartesian topology.
    - can be a vector or a scalar

    Thus, this object is the one which deals with memory allocation
    on each processor (self.data member)

    Constructor (init) process sets the associated topology and allocate
    memory for the data (self.data object, numpy arrays).

    Data are initialized to zero by default. It's possible to provide
    a user-defined function to compute another set of initial values
    by using discrete.initialize(formula=your_function_name).

    Example 1 : direct build of a discrete field
    (topo being a pre-defined parmepy.mpi.topology.Cartesian object) :
    \code
    velocity = ScalarField(topo)
    def myfunc(x,y,z):
        velx = sin(x)
        vely = cos(y)
        velz = sin(z)
        return velx, vely, velz

    velocity.initialize(formula=myfunc)
    \endcode

    Example 2 : description of a continuous and discretisation
    \code
    velocity = Field()
    # discretisation of the continuous field on two different topologies
    # i.e. creation of two discrete fields
    velocity.discretize(topo)
    velocity.discretize(topo2)
    # Initialisation of each discrete fields of velocity with
    # velocity(x,y,z) = myfunc(x,y,z) for each point of the grid
    velocity.initialize(formula=myfunc)
    \endcode
    """
    ## Counter for discrete field instanciations.
    ## Used to set default name and for serialization ids.
    __field_counter = count(0)

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, topology, isVector=False, name=None, nbComponents=None):
        """ Discretize a field on a given topology.

        @param topology : a topology (mpi topo + mesh, parmepy.mpi.topology)
        @param name : field name
        """
        ## Topology informations.
        self.topology = topology

        ## Id (unique) for the field
        self.__id = self.__field_counter.next()
        ## Field name.
        if name is not None:
            self.name = name
        else:
            self.name = 'unamed'
        ## Field dimension.
        self.dimension = topology.domain.dimension
        ## Field resolution.
        self.resolution = self.topology.mesh.resolution
        ## Application domain of the variable.
        self.domain = self.topology.domain
        ## Is self.data contains data
        self.contains_data = False
        ## Scalar or vector
        self.isVector = isVector
        ## Object to store computational times
        self.timer = Timer(self)
        ## Number of components of the field
        if nbComponents is None:
            self.nbComponents = self.dimension if self.isVector else 1
        else:
            self.nbComponents = nbComponents
        ## The memory space for data ...
        self.data = [np.zeros((self.resolution),
                              dtype=PARMES_REAL, order=ORDER)
                     for d in xrange(self.nbComponents)]
 #      assert i == 0, 'Scalar field has only one component.'

    def __getitem__(self, i):
        """ Access to the content of the field.
        @param i : requested index.
        @return component i of the field.
        """
        #print "get item for vector field"
        return self.data[i]

    def __setitem__(self, i, value):
        """
        Set the content of component i of the field.
        """
        self.data[i][...] = value

    @debug
    @timed_function
    def initialize(self, formula=None, *args):
        """
        Initialize values with a given formula.

        Formula need to be vectorized by numpy.
        Therefore formula must have the following signature:
        @li 3D : float, float, float formula(float x, float y, float z, *args)
        @li 2D : float, float formula(float x, float y, *args)
        where x,y,z are point coordinates.

        @param formula : formula to apply on coordinates.
        @param args : Extra parameters for formula
        """
        if formula is not None:
            if isinstance(formula, np.lib.function_base.vectorize):
                v_formula = formula
            else:
                v_formula = np.vectorize(formula)
            if len(self.data) == 1:
                self.data[0][...] = v_formula(*(self.topology.mesh.coords
                                                + args))
            elif len(self.data) == 2:
                self.data[0][...], self.data[1][...] = \
                    v_formula(*(self.topology.mesh.coords
                                + args))
            else:
                self.data[0][...], self.data[1][...], self.data[2][...] = \
                    v_formula(*(self.topology.mesh.coords + args))
            for i in range(self.nbComponents):
                self.data[i][...] = np.asarray(self.data[i],
                                               dtype=PARMES_REAL,
                                               order=ORDER)
        else:
            for d in xrange(self.dimension):
                self.data[d][...] = 0.0

        self.contains_data = True

    def norm(self):
        """
        Compute a 2-norm for the discrete field
        p-norm = (sum data[d](i,...)**p)**1/p for d = 1..dim
        sum on all grid points excluding ghosts.
        Warning : this is probably not a proper way
        to compute the 'real' norm of the field but
        just a way to check field distribution. To be improved.
        """
        result = np.zeros((self.nbComponents))
        gResult = np.zeros((self.nbComponents))
        indNoGhost = self.topology.mesh.iCompute
        for d in range(self.nbComponents):
            result[d] += la.norm(self.data[d][indNoGhost]) ** 2
        main_comm.Allreduce(result, gResult)
        return gResult ** (1. / 2)

    def get_data_method(self):
        pass

    @debug
    @timed_function
    def dump(self, filename, mode=None):
        """
        Dump (serialize) the data of the field into filename.
        @param filename : name of the file in which data are serialized
        @param mode : set mode='append' to add data to an existing file
        """
        filename += '_rk_'
        filename += str(main_rank)
        # create a new db
        if mode is None:
            db = NumPyDB_cPickle(filename, mode='store')
        elif mode is 'append':
            # use an existing db
            db = NumPyDB_cPickle(filename, mode='load')

        #for dim in xrange(self.dimension):
        #    idd = self.name + '_' + str(dim)
        db.dump(self.data, self.name)

    @debug
    @timed_function
    def load(self, filename, fieldname=None):
        """
        load data from a file build
        with cPickle serialization.
        @param filename : name of the file in which data are serialized
        """
        if fieldname is None:
            fieldname = self.name
        filename += '_rk_'
        filename += str(main_rank)
        db = NumPyDB_cPickle(filename, mode='load')
        for dim in xrange(self.nbComponents):
            self.data[dim] = db.load(fieldname)[0][dim]

    def zero(self):
        """ set all components to zero"""
        for dim in xrange(self.nbComponents):
            self.data[dim][...] = 0.0

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Domain."
    print DiscreteField.__doc__





