"""
Testing parmepy.field.scalar.ScalarField,
 parmepy.field.scalar.VectorField
"""
from parmepy.constants import np
from parmepy.fields.continuous import Field
from parmepy.domain.box import Box
from parmepy.mpi.topology import Cartesian


def test_discretization():
    """Test multiple discretizations"""
    dom = Box()
    csf = Field(dom)
    cvf = Field(dom, isVector=True)
    resolTopo = [33, 33, 17]
    topo = Cartesian(dom, 3, resolTopo)
    csf.discretize(topo)
    cvf.discretize(topo)
    assert np.equal(csf.discreteFields[topo].resolution,
                    csf.discreteFields[topo].data[0].shape).all()
    assert not csf.discreteFields[topo].contains_data
    assert np.equal(cvf.discreteFields[topo].resolution,
                    cvf.discreteFields[topo].data[0].shape).all()
    assert np.equal(cvf.discreteFields[topo].resolution,
                    cvf.discreteFields[topo].data[1].shape).all()
    assert np.equal(cvf.discreteFields[topo].resolution,
                    cvf.discreteFields[topo].data[2].shape).all()
    assert not cvf.discreteFields[topo].contains_data


def test_init_scalar():
    dom = Box()
    csf = Field(dom)
    resolTopo = [33, 11, 17]
    topo = Cartesian(dom, 3, resolTopo)
    topo.setUp()
    csfd = csf.discretize(topo)
    assert not csfd.contains_data
    csfd_data_id = id(csfd.data[0])
    scalar = lambda x, y, z: x * y * z
    true_data = np.zeros_like(csf.discreteFields[topo].data[0])
    for i in np.arange(resolTopo[0] - 1):
        for j in np.arange(resolTopo[1] - 1):
            for k in np.arange(resolTopo[2] - 1):
                true_data[i, j, k] = scalar(i / ((resolTopo[0] - 1) * 1.),
                                            j / ((resolTopo[1] - 1) * 1.),
                                            k / ((resolTopo[2] - 1) * 1.))
    csf.discreteFields[topo].initialize(scalar)
    assert csfd.contains_data
    assert np.allclose(csf.discreteFields[topo].data[0], true_data)
    # testing data pointer unchanged
    assert csfd_data_id == id(csfd.data[0])


def test_init_vector():
    dom = Box()
    cvf = Field(dom, isVector=True)
    resolTopo = [33, 11, 17]
    topo = Cartesian(dom, 3, resolTopo)
    topo.setUp()
    cvfd = cvf.discretize(topo)
    assert not cvfd.contains_data
    cvfd_data_id = [id(d) for d in cvfd.data]
    vector = lambda x, y, z: (x ** 2, 1.0 - y, 10 * z + 10.)
    true_data_x = np.zeros_like(cvf.discreteFields[topo].data[0])
    true_data_y = np.zeros_like(cvf.discreteFields[topo].data[1])
    true_data_z = np.zeros_like(cvf.discreteFields[topo].data[2])
    for i in np.arange(resolTopo[0] - 1):
        for j in np.arange(resolTopo[1] - 1):
            for k in np.arange(resolTopo[2] - 1):
                true_data_x[i, j, k], \
                    true_data_y[i, j, k], \
                    true_data_z[i, j, k] = \
                    vector(i / ((resolTopo[0] - 1) * 1.),
                           j / ((resolTopo[1] - 1) * 1.),
                           k / ((resolTopo[2] - 1) * 1.))
    cvf.discreteFields[topo].initialize(vector)
    assert cvfd.contains_data
    assert np.allclose(cvf.discreteFields[topo].data[0], true_data_x)
    assert np.allclose(cvf.discreteFields[topo].data[1], true_data_y)
    assert np.allclose(cvf.discreteFields[topo].data[2], true_data_z)
    # testing data pointer unchanged
    assert cvfd_data_id == [id(d) for d in cvfd.data]


def test_accessing_vector_element():
    """Vector element access"""
    dom = Box()
    cvf = Field(dom, isVector=True)
    resolTopo = [33, 11, 17]
    topo = Cartesian(dom, 3, resolTopo)
    topo.setUp()
    cvf.discretize(topo)
    vector = lambda x, y, z: (x ** 2, 1.0 - y, 10 * z + 10.)
    true_data_x = np.zeros_like(cvf.discreteFields[topo].data[0])
    true_data_y = np.zeros_like(cvf.discreteFields[topo].data[1])
    true_data_z = np.zeros_like(cvf.discreteFields[topo].data[2])
    cvf.discreteFields[topo].initialize(vector)
    tshape = (resolTopo[0] - 1, resolTopo[1] - 1, resolTopo[2] - 1)
    assert np.equal(cvf.discreteFields[topo][0][:, :, :].shape, tshape).all()
    for i in np.arange(resolTopo[0] - 1):
        for j in np.arange(resolTopo[1] - 1):
            for k in np.arange(resolTopo[2] - 1):
                true_data_x[i, j, k], \
                    true_data_y[i, j, k], \
                    true_data_z[i, j, k] = \
                    vector(i / ((resolTopo[0] - 1) * 1.),
                           j / ((resolTopo[1] - 1) * 1.),
                           k / ((resolTopo[2] - 1) * 1.))
                assert np.allclose(cvf.discreteFields[topo][0][i, j, k],
                                   true_data_x[i, j, k])
                assert np.allclose(cvf.discreteFields[topo][1][i, j, k],
                                   true_data_y[i, j, k])
    assert np.allclose(cvf.discreteFields[topo][0], true_data_x)
    assert np.allclose(cvf.discreteFields[topo][1], true_data_y)
    assert np.allclose(cvf.discreteFields[topo][2], true_data_z)
    assert np.allclose(cvf.discreteFields[topo][0][:, 1, 2],
                       true_data_x[:, 1, 2])
    #set item
    cvf.discreteFields[topo][0][0, 1, :] = 5.
    cvf.discreteFields[topo][1][0, 1, :] = 5.
    cvf.discreteFields[topo][2][0, 1, :] = 5.
    true_data = [np.ones_like(cvf.discreteFields[topo].data[0][0, 1, :]) * 5.,
                 np.ones_like(cvf.discreteFields[topo].data[1][0, 1, :]) * 5.,
                 np.ones_like(cvf.discreteFields[topo].data[2][0, 1, :]) * 5.]
    assert np.allclose(cvf.discreteFields[topo][0][0, 1, :], true_data[0])
    assert np.allclose(cvf.discreteFields[topo][1][0, 1, :], true_data[1])
    assert np.allclose(cvf.discreteFields[topo][2][0, 1, :], true_data[2])
    cvf.discreteFields[topo][0][..., 9] = -1.
    assert np.allclose(cvf.discreteFields[topo][0][..., 9],
                       np.ones_like(true_data_x[..., 9]) * -1.)
