/**
 * @file builtin.cl
 * Advection function, vectorized version.
 */

float__N__ advection(uint i, float dt, float dx, float invdx, __local float* gvelo_loc);


/**
 * Compute the position of a particle with a RK2 integration scheme. Velocity is linearly interpolated from the global field.
 * Use of builtin OpenCL functions fma and mix. Computations through OpenCL vector types.
 *
 * @param i Particle index.
 * @param dt Time step.
 * @param dx Space step.
 * @param invdx 1/dx.
 * @param gvelo_loc Local velocity cache.
 * @return Particle position.
 *
 * @remark <code>NB_I</code>, <code>NB_II</code>, <code>NB_III</code> : points number in directions from 1st varying index to last.
 * @remark <code>__N__</code> is expanded at compilation time by vector width.
 * @remark <code>__NN__</code> is expanded at compilation time by a sequence of integer for each vector component.
 * @see parmepy.gpu.tools.parse_file
 */
float__N__ advection(uint i, float dt, float dx, float invdx, __local float* gvelo_loc)
{
  float__N__ v,        		/* Velocity at point */
    p,				/* Intermediary position */
    vp,				/* Velocity at right point */
    y,				/* Normalized distance to left point */
    coord;			/* Grid point coordinate */
  int__N__ ind,        		/* Integer coordinate (cell id) */
    i_ind,			/* Interpolation left point */
    i_ind_p;			/* Interpolation right point */
  v = (float__N__)(gvelo_loc[noBC_id(i+__NN__)],
		   );
  coord = (float__N__)((i+__NN__)*dx,
		       );
  p = fma((float__N__)(0.5*dt),v,coord);
  ind = convert_int__N___rtn(p * invdx);
  y = (fma(- convert_float__N__(ind),dx,p))* invdx ;
  i_ind = ((ind + NB_I) % NB_I);
  i_ind_p = ((i_ind + 1) % NB_I);
  v = (float__N__)(gvelo_loc[noBC_id(i_ind.s__NN__)],
		   );
  vp = (float__N__)(gvelo_loc[noBC_id(i_ind_p.s__NN__)],
		    );
  return fma(mix(v,vp,y),dt,coord);
}
