/**
 * @file advection_and_remeshing.cl
 * Advection and remeshing kernel.
 */

/**
 * Performs advection and then remeshing of the particles' scalar.
 * A work-group is handling a 1D problem. Thus, gidY and gidZ are constants among work-items of a work-group.
 * Each work-item computes NB_I/WI_NB particles positions. To avoid concurrent witings, in case of strong velocity gradients, work-items computes contiguous particles.
 * Particle are computed through OpenCL vector types of lenght 2, 4 or 8.
 * Scalar results are stored in a local buffer as a cache and then copied to global memory buffer.
 *
 * @param gvelo Velocity field
 * @param pscal Particle scalar
 * @param gscal Grid scalar
 * @param dt Time step
 * @param min_position Domain lower coordinate
 * @param dx Space step
 *
 * @remark NB_I, NB_II, NB_III : points number in directions from 1st varying index to last.
 * @remark WI_NB corresponds to the work-item number.
 * @remark \__N__ is expanded at compilation time by vector width.
 * @remark \__NN__ is expanded at compilation time by a sequence of integer for each vector component.
 * @see parmepy.gpu.tools.parse_file
 */
__kernel void advection_and_remeshing(__global const float* gvelo,
				      __global const float* pscal,
				      __global float* gscal,
				      float dt,float min_position, float dx)
{
  uint gidX = get_global_id(0);	/* OpenCL work-itme global index (X) */
  uint gidY = get_global_id(1); /* OpenCL work-itme global index (Y) */
  uint gidZ = get_global_id(2); /* OpenCL work-itme global index (Z) */
  float invdx = 1.0/dx;		/* Space step inverse */
  uint i;			/* Particle index in 1D problem */
  float p,			/* Particle position */
    s,				/* Particle scalar */
    v;				/* Particle velocity */
  uint line_index = gidY*NB_I+ gidZ*NB_I*NB_II; /* Current 1D problem index */
  __local float gscal_loc[NB_I]; /* Local buffer for result */

  for(i=gidX; i<NB_I; i+=(WI_NB))
    {
      /* Read velocity */
      v = gvelo[i+line_index];
      /* Fill velocity cache */
      gvelo_loc[noBC_id(i)] = v;
      /* Initialize result buffer */
      gscal_loc[noBC_id(i)] = 0.0;
    }

  /* Synchronize work-group */
  barrier(CLK_LOCAL_MEM_FENCE);

  for(i=gidX*PART_NB_PER_WI; i<(gidX + 1)*PART_NB_PER_WI; i+=1)
    {
      /* Read Particle scalar */
      s = pscal[i + line_index];
      /* Compute particle position */
      p = advection(i, dt, dx, invdx, gvelo);
      /* Remesh particle */
      remesh(i, dx, invdx, s, p, gscal_loc);
    }

  /* Synchronize work-group */
  barrier(CLK_LOCAL_MEM_FENCE);

  for(i=gidX; i<NB_I; i+=(WI_NB))
    {
      /* Store result */
      gscal[i + line_index] = gscal_loc[noBC_id(i)];
    }
}
