"""
@file gpu_discrete.py

Contains class for discrete fields on GPU.
"""
from parmepy import __VERBOSE__
from parmepy.constants import ORDER, np, debug
from parmepy.fields.discrete import DiscreteField
from parmepy.gpu import cl, PARMES_REAL_GPU, CL_PROFILE
from parmepy.tools.timers import timed_function
from parmepy.gpu.gpu_kernel import KernelLauncher, KernelListLauncher


class GPUDiscreteField(DiscreteField):
    """
    GPU Discrete vector field implementation.
    Allocates OpenCL device memory for the field.
    """
    def __init__(self, queue, topology=None, isVector=False, name="?",
                 precision=PARMES_REAL_GPU, layout=True):
        """
        Constructor.
        @param queue : OpenCL queue
        @param precision : Floating point precision
        @param parent : Continuous field.
        @param topology : Topology informations
        @param name : Field name
        @param idFromParent : Index in the parent's discrete fields
        @param layout : Boolean indicating if components are arranged in memory
        @see parmepy.fields.vector.VectorField.__init__
        """
        super(GPUDiscreteField, self).__init__(topology, isVector, name)
        ## OpenCL queue
        self.queue = queue
        ## Precision for the field
        self.precision = precision
        ## Memory used
        self.mem_size = 0
        ## Flag for updated data localization
        self.data_on_device = False
        ## Initialization OpenCL kernel as KernelLauncher
        self.init_kernel = None
        self._isReleased = False
        ## OpenCL Buffer pointer
        self.gpu_data = [None] * self.nbComponents
        ## Data layout is direction dependant
        self.layout = layout
        for d in xrange(self.nbComponents):
            self.data[d] = np.asarray(self.data[d],
                                      dtype=precision, order=ORDER)
            self.gpu_data[d] = cl.Buffer(self.queue.context,
                                         cl.mem_flags.READ_WRITE,
                                         size=self.data[d].nbytes)
            self.mem_size += self.gpu_data[d].size
        if __VERBOSE__:
            print self.name, self.mem_size, "Bytes (",
            print self.mem_size / (1024 ** 2), "MB)"

    @classmethod
    def fromField(cls, queue, vfield, precision=PARMES_REAL_GPU,
                  layout=True):
        """
        Contructor from a discrete vector field.
        Mutates the given VectorField to a GPUVectorField.
        @param cls : Class of the class method (GPUVectorField)
        @param queue : OpenCL queue
        @param vfield : VectorField
        @param precision : Floating point precision
        @param layout : Boolean indicating if components are arranged in memory
        """
        if not isinstance(vfield, GPUDiscreteField):
            vfield.__class__ = cls
            GPUDiscreteField.__init__(
                vfield, queue,
                vfield.topology, vfield.isVector, vfield.name,
                precision)
            vfield.gpu_data = [None] * vfield.nbComponents
            vfield.layout = layout
            for d in xrange(vfield.nbComponents):
                vfield.data[d] = np.asarray(vfield.data[d],
                                            dtype=precision, order=ORDER)
                vfield.gpu_data[d] = cl.Buffer(vfield.queue.context,
                                               cl.mem_flags.READ_WRITE,
                                               size=vfield.data[d].nbytes)
                vfield.mem_size += vfield.gpu_data[d].size
            if __VERBOSE__:
                print vfield.name, vfield.mem_size, "Bytes (",
                print vfield.mem_size / (1024 ** 2), "MB)"

    def setInitializationKernel(self, kernel):
        """
        Set the initialization kernel
        @param kernel : KernelLauncher to use for initialize field.
        """
        self.init_kernel = kernel

    @debug
    @timed_function
    def dump(self, filename, mode=None):
        self.toHost()
        DiscreteField.dump(self, filename, mode)

    @debug
    @timed_function
    def load(self, filename):
        DiscreteField.load(self, filename)
        self.toDevice()
        self.contains_data = False
        self.data_on_device = True

    @debug
    @timed_function
    def initialize(self, formula=None, *args):
        """
        GPU data initialization.
        Performs the initialization from different ways if device not already
        contains up-to-date data:
          - with an OpenCL kernel,
          - with a python formula (as VectorField) and the copy data to device.
        @param formula : Formula to use.
        @param args : formula extra parameters
        """
        isGPUKernel = isinstance(formula, KernelLauncher) \
            or isinstance(formula, KernelListLauncher)
        if not isGPUKernel and self.init_kernel is None:
            DiscreteField.initialize(self, formula, *args)
            if self.contains_data:
                for d in xrange(self.nbComponents):
                    self.data[d] = np.asarray(
                        self.data[d],
                        dtype=self.precision, order=ORDER)
                self.toDevice()
        else:
            if isGPUKernel:
                self.init_kernel = formula
            coord_min = np.ones(4, dtype=self.precision)
            mesh_size = np.ones(4, dtype=self.precision)
            coord_min[0:self.dimension] = np.asarray(
                self.topology.mesh.origin,
                dtype=self.precision)
            mesh_size[0:self.dimension] = np.asarray(
                self.topology.mesh.space_step,
                dtype=self.precision)
            if self.nbComponents == 2:
                self.init_kernel(self.gpu_data[0],
                                 self.gpu_data[1],
                                 coord_min, mesh_size, *args)
            elif self.nbComponents == 3:
                self.init_kernel(self.gpu_data[0],
                                 self.gpu_data[1],
                                 self.gpu_data[2],
                                 coord_min, mesh_size, *args)
            else:
                self.init_kernel(self.gpu_data[0],
                                 coord_min, mesh_size, *args)
        self.contains_data = False
        self.data_on_device = True

    def release(self):
        if not self._isReleased:
            if __VERBOSE__:
                print "deallocate :", self.name,
                print " (" + str(self.mem_size / (1024. ** 2)) + " MBytes)"
            for d in xrange(self.nbComponents):
                self.gpu_data[d].release()
            if self.init_kernel is not None:
                for f_timer in self.init_kernel.f_timer:
                    self.timer.addFunctionTimer(f_timer)
            print self.timer
            self._isReleased = True

    @timed_function
    def toDevice(self):
        """
        Host to device method.

        Performs a direct OpenCL copy from numpy arrays
        to OpenCL Buffers.\n
        Arrange memory on device so that vector components are
        contiguous in the direction of the component, if layout flag is True.\n
        Example : A 3D vector field F(x,y,z) is made up of 3
        OpenCL Buffers Fx, Fy, Fz. The memory layout is :
        - Fx : x-major ordering. On device,
        Fx[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fx(i,j,k)
        - Fy : y-major ordering. On device,
        Fy[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fy(j,i,k)
        - Fz : z-major ordering. On device,
        Fz[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fz(k,i,j)
        """
        if __VERBOSE__:
            print "host->device :", self.name,
        time = 0.
        evt = [None] * 3

        if self.layout:
            evt[0] = cl.enqueue_copy(
                self.queue, self.gpu_data[0],
                self.data[0].ravel(order=ORDER))
            if self.nbComponents > 1:
                evt[1] = cl.enqueue_copy(
                    self.queue, self.gpu_data[1],
                    self.data[1].swapaxes(0, 1).ravel(order=ORDER))
            if self.nbComponents > 2:
                evt[2] = cl.enqueue_copy(
                    self.queue, self.gpu_data[2],
                    self.data[2].swapaxes(0, 1).swapaxes(0, 2).ravel(
                        order=ORDER))
        else:
            for d in xrange(self.nbComponents):
                evt[d] = cl.enqueue_copy(
                    self.queue, self.gpu_data[d],
                    self.data[d].ravel(order=ORDER))
        if CL_PROFILE:
            self.queue.finish()
            for e in evt:
                if e is not None:
                    time += (e.profile.end - e.profile.start) * 1e-9
        if __VERBOSE__ and CL_PROFILE:
            print self.mem_size, "Bytes transfered at ",
            print "{0:.3f} GBytes/sec".format(
                self.mem_size / (time * 1024 ** 3))
            print "Transfer host->device"

    @timed_function
    def toHost(self):
        """
        Device to host method.

        Performs a direct OpenCL copy from OpenCL Buffers
        to numpy arrays.\n
        As memory layout, if set, is arranged on device, not only a
        copy is performed but also transpositions to have numpy
        arrays consistent to each other.
        """
        if not self.contains_data:
            if __VERBOSE__:
                print "device->host :", self.name,
            self.queue.finish()
            shape = self.data[0].shape
            time = 0.
            evt = [None] * 3
            if self.layout:
                # Get XDIR component
                evt[0] = cl.enqueue_copy(self.queue, self.data[0],
                                         self.gpu_data[0])
                if self.nbComponents > 1:
                    # Get YDIR component
                    temp = np.empty_like(self.data[1])
                    if self.dimension == 3:
                        temp.resize((shape[1], shape[0], shape[2]))
                    else:
                        temp.resize((shape[1], shape[0]))
                    evt[1] = cl.enqueue_copy(self.queue, temp,
                                             self.gpu_data[1])
                    self.data[1] = temp.swapaxes(0, 1)
                if self.nbComponents > 2:
                    # Get ZDIR component
                    temp = np.empty_like(self.data[2])
                    temp.resize((shape[2], shape[0], shape[1]))
                    evt[2] = cl.enqueue_copy(self.queue, temp,
                                             self.gpu_data[2])
                    self.data[2] = temp.swapaxes(0, 2).swapaxes(0, 1)
            else:
                for d in xrange(self.nbComponents):
                    # Get d component
                    evt[d] = cl.enqueue_copy(self.queue, self.data[d],
                                             self.gpu_data[d])
            if CL_PROFILE:
                self.queue.finish()
                for e in evt:
                    if e is not None:
                        time += (e.profile.end - e.profile.start) * 1e-9
            if __VERBOSE__ and CL_PROFILE:
                print self.mem_size, "Bytes transfered at ",
                print "{0:.3f} GBytes/sec".format(
                    self.mem_size / (time * 1024 ** 3))
                print 'Transfer device->host'
