"""
@file parmepy.gpu.tests.test_advection_randomVelocity
Testing advection kernels with a random velocity field.
"""
from parmepy.domain.box import Box
from parmepy.fields.continuous import Field
from parmepy.operator.advection import Advection
from parmepy.gpu import PARMES_REAL_GPU
from parmepy.constants import ORDER, np
from parmepy.problem.simulation import Simulation


def setup_2D():
    box = Box(2, length=[1., 1.], origin=[0., 0.])
    scal = Field(domain=box, name='Scalar')
    velo = Field(domain=box, name='Velocity', isVector=True)
    return scal, velo


def setup_3D():
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar')
    velo = Field(domain=box, name='Velocity', isVector=True)
    return scal, velo


def assertion_2D_withPython(scal, velo, advec, advec_py):
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL_GPU, order=ORDER)
    velo_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[0])
    velo_d.data[1][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[1])
    scal_d.toDevice()
    velo_d.toDevice()

    advec_py.apply(Simulation(0., 0.01, 1))
    advec.apply(Simulation(0., 0.01, 1))

    py_res = scal_d.data[0].copy()
    scal_d.toHost()

    advec.finalize()
    return np.allclose(py_res, scal_d.data[0], rtol=5e-02, atol=5e-05)


def assertion_3D_withPython(scal, velo, advec, advec_py):
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL_GPU, order=ORDER)
    velo_d.data[0][...] = np.zeros_like(
        scal_d.data[0],
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[0])
    velo_d.data[1][...] = np.zeros_like(
        scal_d.data[0],
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[1])
    velo_d.data[2][...] = np.zeros_like(
        scal_d.data[0],
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[2])
    scal_d.toDevice()
    velo_d.toDevice()

    advec_py.apply(Simulation(0., 0.01, 1))
    advec.apply(Simulation(0., 0.01, 1))

    py_res = scal_d.data[0].copy()
    scal_d.toHost()

    advec.finalize()
    return np.allclose(py_res, scal_d.data[0], rtol=5e-02, atol=5e-05)


# M6 testing
def test_2D_m6_1k():
    """
    Testing M6 remeshing formula in 2D, 1 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method='gpu_1k_m6prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method='m6prime',
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m6_2k():
    """
    Testing M6 remeshing formula in 2D, 2 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method='gpu_2k_m6prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method='m6prime',
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m6_1k_sFH():
    """
    Testing M6 remeshing formula in 2D, 1 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method='gpu_1k_m6prime',
                      precision=PARMES_REAL_GPU,
                      splittingConfig="o2_FullHalf"
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method='m6prime',
                         splittingConfig="o2_FullHalf"
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m6_2k_sFH():
    """
    Testing M6 remeshing formula in 2D, 2 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method='gpu_2k_m6prime',
                      precision=PARMES_REAL_GPU,
                      splittingConfig="o2_FullHalf"
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method='m6prime',
                         splittingConfig="o2_FullHalf"
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_3D_m6_1k():
    """
    Testing M6 remeshing formula in 3D, 1 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='gpu_1k_m6prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method='m6prime',
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m6_2k():
    """
    Testing M6 remeshing formula in 3D, 2 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='gpu_2k_m6prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method='m6prime',
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m6_1k_sFH():
    """
    Testing M6 remeshing formula in 3D, 1 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='gpu_1k_m6prime',
                      precision=PARMES_REAL_GPU,
                      splittingConfig="o2_FullHalf"
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method='m6prime',
                         splittingConfig="o2_FullHalf"
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m6_2k_sFH():
    """
    Testing M6 remeshing formula in 3D, 2 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='gpu_2k_m6prime',
                      precision=PARMES_REAL_GPU,
                      splittingConfig="o2_FullHalf"
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method='m6prime',
                         splittingConfig="o2_FullHalf"
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


# M4 testing
def test_2D_m4_1k():
    """
    Testing M4 remeshing formula in 2D, 1 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method='gpu_1k_m4prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method='m4prime',
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m4_2k():
    """
    Testing M4 remeshing formula in 2D, 2 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method='gpu_2k_m4prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method='m4prime',
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m4_1k_sFH():
    """
    Testing M4 remeshing formula in 2D, 1 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method='gpu_1k_m4prime',
                      precision=PARMES_REAL_GPU,
                      splittingConfig="o2_FullHalf"
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method='m4prime',
                         splittingConfig="o2_FullHalf"
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m4_2k_sFH():
    """
    Testing M4 remeshing formula in 2D, 2 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method='gpu_2k_m4prime',
                      precision=PARMES_REAL_GPU,
                      splittingConfig="o2_FullHalf"
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method='m4prime',
                         splittingConfig="o2_FullHalf"
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_3D_m4_1k():
    """
    Testing M4 remeshing formula in 3D, 1 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='gpu_1k_m4prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method='m4prime',
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m4_2k():
    """
    Testing M4 remeshing formula in 3D, 2 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='gpu_2k_m4prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method='m4prime',
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m4_1k_sFH():
    """
    Testing M4 remeshing formula in 3D, 1 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='gpu_1k_m4prime',
                      precision=PARMES_REAL_GPU,
                      splittingConfig="o2_FullHalf"
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method='m4prime',
                         splittingConfig="o2_FullHalf"
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m4_2k_sFH():
    """
    Testing M4 remeshing formula in 3D, 2 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='gpu_2k_m4prime',
                      precision=PARMES_REAL_GPU,
                      splittingConfig="o2_FullHalf"
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method='m4prime',
                         splittingConfig="o2_FullHalf"
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


# M8 testing
def test_2D_m8_1k():
    """
    Testing M8 remeshing formula in 2D, 1 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method='gpu_1k_m8prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method='m8prime',
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m8_2k():
    """
    Testing M8 remeshing formula in 2D, 2 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method='gpu_2k_m8prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method='m8prime',
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m8_1k_sFH():
    """
    Testing M8 remeshing formula in 2D, 1 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method='gpu_1k_m8prime',
                      precision=PARMES_REAL_GPU,
                      splittingConfig="o2_FullHalf"
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method='m8prime',
                         splittingConfig="o2_FullHalf"
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_2D_m8_2k_sFH():
    """
    Testing M8 remeshing formula in 2D, 2 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [33, 33],
                                   scal: [33, 33]},
                      method='gpu_2k_m8prime',
                      precision=PARMES_REAL_GPU,
                      splittingConfig="o2_FullHalf"
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [33, 33],
                                      scal: [33, 33]},
                         method='m8prime',
                         splittingConfig="o2_FullHalf"
                         )
    assert assertion_2D_withPython(scal, velo, advec, advec_py)


def test_3D_m8_1k():
    """
    Testing M8 remeshing formula in 3D, 1 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='gpu_1k_m8prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method='m8prime',
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m8_2k():
    """
    Testing M8 remeshing formula in 3D, 2 kernel, simple precision,
    o2 splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='gpu_2k_m8prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method='m8prime',
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m8_1k_sFH():
    """
    Testing M8 remeshing formula in 3D, 1 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='gpu_1k_m8prime',
                      precision=PARMES_REAL_GPU,
                      splittingConfig="o2_FullHalf"
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method='m8prime',
                         splittingConfig="o2_FullHalf"
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_3D_m8_2k_sFH():
    """
    Testing M8 remeshing formula in 3D, 2 kernel, simple precision,
    o2_FullHalf splitting.
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='gpu_2k_m8prime',
                      precision=PARMES_REAL_GPU,
                      splittingConfig="o2_FullHalf"
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method='m8prime',
                         splittingConfig="o2_FullHalf"
                         )
    assert assertion_3D_withPython(scal, velo, advec, advec_py)


def test_rectangular_domain2D():
    box = Box(2, length=[1., 1.], origin=[0., 0.])
    scal = Field(domain=box, name='Scalar')
    velo = Field(domain=box, name='Velocity', isVector=True)

    advec = Advection(velo, scal,
                      resolutions={velo: [65, 33],
                                   scal: [65, 33]},
                      method='gpu_1k_m6prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [65, 33],
                                      scal: [65, 33]},
                         method='m6prime',
                         )
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                  dtype=PARMES_REAL_GPU, order=ORDER)
    velo_d.data[0][...] = np.asarray(
        np.random.random(velo_d.data[0].shape),
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[0])
    velo_d.data[1][...] = np.asarray(
        np.random.random(velo_d.data[1].shape),
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[1])
    scal_d.toDevice()
    velo_d.toDevice()

    advec_py.apply(Simulation(0., 0.01, 1))
    advec.apply(Simulation(0., 0.01, 1))

    py_res = scal_d.data[0].copy()
    scal_d.toHost()

    assert np.allclose(py_res, scal_d.data[0], rtol=5e-02, atol=5e-05)
    advec.finalize()


def test_rectangular_domain3D():
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar')
    velo = Field(domain=box, name='Velocity', isVector=True)

    advec = Advection(velo, scal,
                      resolutions={velo: [65, 33, 17],
                                   scal: [65, 33, 17]},
                      method='gpu_1k_m6prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [65, 33, 17],
                                      scal: [65, 33, 17]},
                         method='m6prime',
                         )
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                  dtype=PARMES_REAL_GPU, order=ORDER)
    velo_d.data[0][...] = np.asarray(
        np.random.random(velo_d.data[0].shape),
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[0])
    velo_d.data[1][...] = np.asarray(
        np.random.random(velo_d.data[1].shape),
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[1])
    velo_d.data[2][...] = np.asarray(
        np.random.random(velo_d.data[2].shape),
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[2])
    scal_d.toDevice()
    velo_d.toDevice()

    advec_py.apply(Simulation(0., 0.01, 1))
    advec.apply(Simulation(0., 0.01, 1))

    py_res = scal_d.data[0].copy()
    scal_d.toHost()

    assert np.allclose(py_res, scal_d.data[0], rtol=5e-02, atol=5e-05)
    advec.finalize()


def test_vector_2D():
    box = Box(2, length=[1., 1.], origin=[0., 0.])
    scal = Field(domain=box, name='Scalar', isVector=True)
    velo = Field(domain=box, name='Velocity', isVector=True)

    advec = Advection(velo, scal,
                      resolutions={velo: [129, 129],
                                   scal: [129, 129]},
                      method='gpu_1k_m6prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [129, 129],
                                      scal: [129, 129]},
                         method='m6prime',
                         )
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL_GPU, order=ORDER)
    scal_d.data[1][...] = np.asarray(np.random.random(scal_d.data[1].shape),
                                     dtype=PARMES_REAL_GPU, order=ORDER)
    velo_d.data[0][...] = np.asarray(
        np.random.random(velo_d.data[0].shape),
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[0])
    velo_d.data[1][...] = np.asarray(
        np.random.random(velo_d.data[1].shape),
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[1])
    scal_d.toDevice()
    velo_d.toDevice()

    advec_py.apply(Simulation(0., 0.01, 1))
    advec.apply(Simulation(0., 0.01, 1))

    py_res_X = scal_d.data[0].copy()
    py_res_Y = scal_d.data[1].copy()
    scal_d.toHost()

    print np.max(py_res_X-  scal_d.data[0]), np.max(py_res_Y- scal_d.data[1])
    assert np.allclose(py_res_X, scal_d.data[0], rtol=5e-02, atol=5e-05)
    assert np.allclose(py_res_Y, scal_d.data[1], rtol=5e-02, atol=5e-05)
    advec.finalize()


def test_vector_3D():
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar', isVector=True)
    velo = Field(domain=box, name='Velocity', isVector=True)

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='gpu_1k_m6prime',
                      precision=PARMES_REAL_GPU,
                      )
    advec_py = Advection(velo, scal,
                         resolutions={velo: [17, 17, 17],
                                      scal: [17, 17, 17]},
                         method='m6prime',
                         )
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL_GPU, order=ORDER)
    scal_d.data[1][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL_GPU, order=ORDER)
    scal_d.data[2][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL_GPU, order=ORDER)
    velo_d.data[0][...] = np.asarray(
        np.random.random(velo_d.data[0].shape),
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[0])
    velo_d.data[1][...] = np.asarray(
        np.random.random(velo_d.data[1].shape),
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[1])
    velo_d.data[2][...] = np.asarray(
        np.random.random(velo_d.data[2].shape),
        dtype=PARMES_REAL_GPU, order=ORDER) / (2. * scal_d.resolution[2])
    scal_d.toDevice()
    velo_d.toDevice()

    advec_py.apply(Simulation(0., 0.01, 1))
    advec.apply(Simulation(0., 0.01, 1))

    py_res_X = scal_d.data[0].copy()
    py_res_Y = scal_d.data[1].copy()
    py_res_Z = scal_d.data[2].copy()
    scal_d.toHost()

    assert np.allclose(py_res_X, scal_d.data[0], rtol=5e-02, atol=5e-05)
    assert np.allclose(py_res_Y, scal_d.data[1], rtol=5e-02, atol=5e-05)
    assert np.allclose(py_res_Z, scal_d.data[2], rtol=5e-02, atol=5e-05)
    advec.finalize()


if __name__ == '__main__':
    test_vector_3D()
