"""
@file tools.py

Tools for gpu management.
"""
from parmepy import __VERBOSE__
from parmepy.constants import S_DIR
from parmepy.gpu import cl, PARMES_REAL_GPU, GPU_SRC, CL_PROFILE

## Global variable handling an OpenCL Environment instance
__cl_env = None


class OpenCLEnvironment(object):
    """OpenCL environment informations and useful functions.
    """

    def __init__(self, platform_id, device_id, device_type,
                 precision, resolution, gl_sharing=False):
        """Create environment.
        @param platform_id : OpenCL platform id
        @param device_id : OpenCL device id
        @param device_type : OpenCL device type
        @param precision : Recquired precision
        @param resolution : Global resolution
        @param gl_sharing : Flag to build a OpenGL shared OpenCL context
        """
        self._platform_id = platform_id
        self._device_id = device_id
        self._device_type = device_type
        self._gl_sharing = gl_sharing
        ## OpenCL platform
        self.platform = self._get_platform(platform_id)
        ## OpenCL device
        self.device = self._get_device(self.platform, device_id, device_type)
        ## OpenCL context
        self.ctx = self._get_context(self.device, gl_sharing)
        ## OpenCL queue
        self.queue = self._get_queue(self.ctx)
        ## Floating point precision
        self.precision = precision
        self.macros = {}
        self.default_build_opts = ""
        if not self.precision is None:
            self.default_build_opts += self._get_precision_opts()
        if resolution is not None:
            for i in xrange(len(resolution)):
                self.default_build_opts += " -D NB" + S_DIR[i]
                self.default_build_opts += "=" + str(resolution[i])
            if len(resolution) == 3:
                self.default_sizes_constants = [
                    " -D NB_I=NB_X -D NB_II=NB_Y -D NB_III=NB_Z",
                    " -D NB_I=NB_Y -D NB_II=NB_X -D NB_III=NB_Z",
                    " -D NB_I=NB_Z -D NB_II=NB_X -D NB_III=NB_Y",
                    ]
            else:
                self.default_sizes_constants = [
                    " -D NB_I=NB_X -D NB_II=NB_Y",
                    " -D NB_I=NB_Y -D NB_II=NB_X",
                    ]
        else:
            self.default_sizes_constants = None

    def modify(self, platform_id, device_id, device_type,
               precision, resolution, gl_sharing=False):
        """
        Modify OpenCL environment parameters.
        @param platform_id : OpenCL platform id
        @param device_id : OpenCL device id
        @param device_type : OpenCL device type
        @param precision : Recquired precision
        @param resolution : Global resolution
        @param gl_sharing : Flag to build a OpenGL shared OpenCL context
        """
        platform_changed, device_changed = False, False
        if not platform_id == self._platform_id:
            print "platform changed"
            self._platform_id = platform_id
            self.platform = self._get_platform(platform_id)
            platform_changed = True
        if platform_changed or not (device_id is self._device_id
                                    and device_type == self._device_type):
            print "device changed"
            self._device_id = device_id
            self._device_type = device_type
            self.device = self._get_device(self.platform,
                                           device_id, device_type)
            device_changed = True
        if platform_changed or device_changed or \
                (not self._gl_sharing and not gl_sharing is self._gl_sharing):
            if self._gl_sharing and not gl_sharing:
                print "Warning: Loosing Gl shared context."
            self._gl_sharing = gl_sharing
            self.ctx = self._get_context(self.device, gl_sharing)
            self.queue = self._get_queue(self.ctx)
        if not self.precision is precision and not precision is None:
            if not self.precision is None:
                print "Warning, GPU precision is overrided from",
                print self.precision, 'to', precision
            self.precision = precision
            self.default_build_opts = self._get_precision_opts()

        if resolution is not None:
            self.default_build_opts = ""
            if not self.precision is None:
                self.default_build_opts += self._get_precision_opts()
            for i in xrange(len(resolution)):
                self.default_build_opts += " -D NB" + S_DIR[i]
                self.default_build_opts += "=" + str(resolution[i])
            if len(resolution) == 3:
                self.default_sizes_constants = [
                    " -D NB_I=NB_X -D NB_II=NB_Y -D NB_III=NB_Z",
                    " -D NB_I=NB_Y -D NB_II=NB_X -D NB_III=NB_Z",
                    " -D NB_I=NB_Z -D NB_II=NB_X -D NB_III=NB_Y",
                    ]
            else:
                self.default_sizes_constants = [
                    " -D NB_I=NB_X -D NB_II=NB_Y",
                    " -D NB_I=NB_Y -D NB_II=NB_X",
                    ]

    def _get_platform(self, platform_id):
        """
        Get an OpenCL platform.
        @param platform_id : OpenCL platform id
        @return OpenCL platform

        \code
        >>> _get_platform(0) == cl.get_platforms()[0]
        True
        >>> _get_platform(-1) == cl.get_platforms()[0]
        True
        >>>
        \endcode
        """
        try:
            # OpenCL platform
            platform = cl.get_platforms()[platform_id]
        except IndexError:
            print "  Incorrect platform_id :", platform_id, ".",
            print " Only ", len(cl.get_platforms()), " available.",
            print " Getting defalut platform. "
            platform = cl.get_platforms()[0]
        if __VERBOSE__:
            print "  Platform   "
            print "  - Name       :", platform.name
            print "  - Version    :", platform.version
        return platform

    def _get_device(self, platform, device_id, device_type):
        """
        Get an OpenCL device.
        @param platform : OpenCL platform
        @param device_id : OpenCL device id
        @param device_type : OpenCL device type
        @return OpenCL device

        Try to use given parameters and in case of fails, use pyopencl context
        creation function.
        """
        try:
            device = platform.get_devices(
                eval("cl.device_type." + device_type.upper()))[device_id]
        except cl.RuntimeError as e:
            print "RuntimeError:", e
            device = cl.create_some_context().devices[0]
        except AttributeError as e:
            print "AttributeError:", e
            device = cl.create_some_context().devices[0]
        if __VERBOSE__:
            print "  Device"
            print "  - Name                :",
            print device.name
            print "  - Type                :",
            print cl.device_type.to_string(device.type)
            print "  - C Version           :",
            print device.opencl_c_version
            print "  - Global mem size     :",
            print device.global_mem_size / (1024 ** 3), "GB"
        return device

    def _get_context(self, device, gl_sharing):
        """
        Get an OpenCL context.
        @param device : OpenCL device
        @param gl_sharing : Flag to build a OpenGL shared OpenCL context
        @return OpenCL context

        """
        props = None
        if gl_sharing:
            from pyopencl.tools import get_gl_sharing_context_properties
            import sys
            if sys.platform == "darwin":
                props = get_gl_sharing_context_properties()
            else:
                # Some OSs prefer clCreateContextFromType, some prefer
                # clCreateContext. Try both.
                props = \
                    [(cl.context_properties.PLATFORM, self.platform)] \
                    + get_gl_sharing_context_properties()
            ctx = cl.Context(properties=props, devices=[device])
        else:
            ctx = cl.Context([device])
        if __VERBOSE__:
            print " Context:"
            if not props is None:
                print "  - properties           :", props
        return ctx

    def _get_queue(self, ctx):
        """
        Get OpenCL queue from context
        @param ctx : OpenCL context
        @return OpenCL queue
        """
        props = None
        if CL_PROFILE:
            props = cl.command_queue_properties.PROFILING_ENABLE
            queue = cl.CommandQueue(ctx, properties=props)
        else:
            queue = cl.CommandQueue(ctx)
        if __VERBOSE__:
            print " Queue"
            if not props is None:
                print "  - properties           :", props
            print "==="
        return queue

    def get_WorkItems(self, resolution, vector_width=1):
        """
        Set the optimal work-item number and OpenCL space index.
        @param resolution : Problem resolution
        @param vector_width : OpenCL vector types width
        @return work-item number, global space index and local space index

        Use 64 work-items in 3D and 256 in 2D.
        \todo Use Both the number from device capability
        The problem must be a multiple of and greater
        than work-item number * vector_width
        """
        # Optimal work item number
        if len(resolution) == 3:
            workItemNumber = 64 if min(resolution) >= 64 \
                else min(resolution)
        else:
            workItemNumber = 256 if min(resolution) >= 256 \
                else min(resolution)
        # Change work-item regarding problem size
        if resolution[0] % workItemNumber > 0:
            if len(resolution) == 3:
                print "Warning : GPU best performances obtained for",
                print "problem sizes multiples of 64"
            else:
                print "Warning : GPU best performances obtained for",
                print "problem sizes multiples of 256"
        while(resolution[0] % workItemNumber > 0):
            workItemNumber = workItemNumber / 2
        # Change work-item regarding vector_width
        if workItemNumber * vector_width > resolution[0]:
            if resolution[0] % vector_width > 0:
                raise ValueError(
                    "Resolution ({0}) must be a multiple of {1}".format(
                        resolution[0], vector_width))
            workItemNumber = resolution[0] // vector_width
        if len(resolution) == 3:
            gwi = (int(workItemNumber),
                   int(resolution[1]), int(resolution[2]))
            lwi = (int(workItemNumber), 1, 1)
        else:
            gwi = (int(workItemNumber),
                   int(resolution[1]))
            lwi = (int(workItemNumber), 1)
        return workItemNumber, gwi, lwi

    def _get_precision_opts(self):
        """Check if device is capable regarding given precision
        @return build options regarding precision recquired
        """
        opts = ""
        # Precision supported
        if __VERBOSE__:
            print " Precision capability  ",
        capable = True
        if self.precision is PARMES_REAL_GPU:
            opts += " -cl-single-precision-constant"
            if __VERBOSE__:
                print "for Single Precision: "
            for v in ['DENORM', 'INF_NAN',
                      'ROUND_TO_NEAREST', 'ROUND_TO_ZERO', 'ROUND_TO_INF',
                      'FMA', 'CORRECTLY_ROUNDED_DIVIDE_SQRT', 'SOFT_FLOAT']:
                try:
                    if eval('(self.device.single_fp_config &' +
                            ' cl.device_fp_config.' +
                            v + ') == cl.device_fp_config.' + v):
                        if __VERBOSE__:
                            print v
                except AttributeError as ae:
                    if v is 'CORRECTLY_ROUNDED_DIVIDE_SQRT':
                        capable = False
                    if __VERBOSE__:
                        print v, 'is not supported in OpenCL C 1.2.\n',
                        print '   Exception catched : ', ae
        else:
            if __VERBOSE__:
                print "for Double Precision: "
            if self.device.double_fp_config > 0:
                for v in ['DENORM', 'INF_NAN', 'FMA',
                          'ROUND_TO_NEAREST', 'ROUND_TO_ZERO', 'ROUND_TO_INF',
                          'CORRECTLY_ROUNDED_DIVIDE_SQRT', 'SOFT_FLOAT']:
                    try:
                        if eval('(self.device.double_fp_config &' +
                                ' cl.device_fp_config.' +
                                v + ') == cl.device_fp_config.' + v):
                            if __VERBOSE__:
                                print v
                    except AttributeError as ae:
                        if v is 'CORRECTLY_ROUNDED_DIVIDE_SQRT':
                            capable = False
                        if __VERBOSE__:
                            print  v, 'is supported in OpenCL C 1.2.\n',
                            print '   Exception catched : ', ae
            else:
                raise ValueError("Double Precision is not supported by device")
        if capable:
            opts += " -cl-fp32-correctly-rounded-divide-sqrt"
        return opts

    def build_src(self, files, options="", vector_width=4):
        """
        Build OpenCL sources
        @param files: Source files
        @param options : Compiler options to use for buildind
        @param vector_width : OpenCL vector type width
        @return OpenCL binaries

        Parse the sources to handle single and double precision.
        """
        gpu_src = ""
        if isinstance(files, list):
            file_list = files
        else:
            file_list = [files]
        if __VERBOSE__:
            print "=== Kernel sources compiling ==="
            for sf in file_list:
                print "   - ", sf
        for sf in file_list:
            try:
                f = open(sf, 'r')
            except IOError as ioe:
                if ioe.errno == 2:
                    f = open(GPU_SRC + sf, 'r')
                else:
                    raise ioe
            gpu_src += "".join(self.parse_file(f, n=vector_width))
            f.close()
        for k in self.macros:
            gpu_src = gpu_src.replace(k, str(self.macros[k]))
        if self.precision is PARMES_REAL_GPU:
            prg = cl.Program(self.ctx, gpu_src.replace('.0', '.0f'))
        else:
            prg = cl.Program(self.ctx, gpu_src.replace('float', 'double'))
        # OpenCL program
        try:
            build = prg.build(self.default_build_opts + options)
        except Exception, e:
            print "Build files : "
            for sf in file_list:
                print "   - ", sf
            print "Build options : ", self.default_build_opts + options
            raise e
        if __VERBOSE__:
            #print options
            print "Build options : ",
            print build.get_build_info(
                self.device, cl.program_build_info.OPTIONS)
            print "Compiler status : ",
            print build.get_build_info(
                self.device, cl.program_build_info.STATUS)
            print "Compiler log : ",
            print build.get_build_info(self.device,
                                       cl.program_build_info.LOG)
            print "===\n"
        return build

    def parse_file(self, f, n=8):
        """
        Parse a file containing OpenCL sources.
        @param f : source file
        @param n : vector width
        @return parsed sources as string.

        - <code>__N__</code> is expanded as an integer corresponding to a
        vector with
        - <code>__NN__</code>, instruction is duplicated to operate on each
        vector component:
          - if line ends with '<code>;</code>', the whole instruciton is
          duplicated.
          - if line ends with '<code>,</code>' and contains
          '<code>(float__N__)(</code>', the float element is duplicated

        Examples with a 4-width vector:\n
        \code
        float__N__ x;           ->  float4 x;
        x.s__NN__ = 1.0f;       ->  x.s0 = 1.0f;
                                    x.s1 = 1.0f;
                                    x.s2 = 1.0f;
                                    x.s3 = 1.0f;
        x = (int__N__)(\__NN__,  ->  x = (int4)(0,
                       );                      1,
                                               2,
                                               3,
                                               );
        \endcode
        """
        src = ""
        for l in f.readlines():
            if l.find('(float__N__)(') >= 0 and l.find('__NN__') >= 0 and \
                    l[-2] == ',':
                sl = l.split("(float__N__)(")
                l = sl[0] + "(float" + str(n) + ")("
                el = sl[1].rsplit(',', 1)[0]
                for i in xrange(n):
                    l += el.replace('__NN__', str(i)) + ','
                l = l[:-1] + '\n'
            elif l.find('__NN__') >= 0 and l[-2] == ';':
                el = ""
                for i in xrange(n):
                    el += l.replace('__NN__', str(i))
                l = el
            src += l.replace('__N__', str(n))
        return src


def get_opengl_shared_environment(platform_id, device_id, device_type,
                                  resolution, precision):
    """
    Get an OpenCL environment with OpenGL shared enable.

    @param platform_id :OpenCL platform id
    @param device_id : OpenCL device id
    @param device_type : OpenCL device type
    @param precision : Required precision
    @return OpenCL platform, device, context and queue

    The context is obtained with gl-shared properties depending on the OS.
    """
    global __cl_env
    if __cl_env is None:
        __cl_env = OpenCLEnvironment(platform_id, device_id, device_type,
                                     precision, resolution, gl_sharing=True)
    else:
        __cl_env.modify(platform_id, device_id, device_type,
                        precision, resolution, gl_sharing=True)
    return __cl_env


def get_opencl_environment(platform_id, device_id, device_type, precision,
                           resolution):
    """
    Get an OpenCL environment.

    @param platform_id :OpenCL platform id
    @param device_id : OpenCL device id
    @param device_type : OpenCL device type
    @param precision : Required precision
    @return OpenCL platform, device, context and queue

    """
    global __cl_env
    if __cl_env is None:
        __cl_env = OpenCLEnvironment(platform_id, device_id, device_type,
                                     precision, resolution)
    else:
        __cl_env.modify(platform_id, device_id, device_type,
                        precision, resolution)
    return __cl_env


def explore():
    """Print environment details"""
    print "OpenCL exploration : "
    platforms = cl.get_platforms()
    platforms_info = ["name", "version", "vendor", "profile", "extensions"]
    devices_info = ["name",
                    "version",
                    "vendor",
                    "profile",
                    "extensions",
                    "available",
                    "type",
                    "compiler_available",
                    "double_fp_config",
                    "single_fp_config",
                    "global_mem_size",
                    "global_mem_cache_type",
                    "global_mem_cache_size",
                    "global_mem_cacheline_size",
                    "local_mem_size",
                    "local_mem_type",
                    "max_clock_frequency",
                    "max_compute_units",
                    "max_constant_buffer_size",
                    "max_mem_alloc_size",
                    "max_work_group_size",
                    "max_work_item_dimensions",
                    "max_work_item_sizes",
                    "preferred_vector_width_double",
                    "preferred_vector_width_float",
                    "preferred_vector_width_int"]
    for pltfm in platforms:
        print "Platform:", pltfm.name
        for pltfm_info in platforms_info:
            print "  |-", pltfm_info, ':', eval("pltfm." + pltfm_info)
        devices = pltfm.get_devices()
        for dvc in devices:
            print "  |- Device:", dvc.name
            for dvc_info in devices_info:
                print "    |-", dvc_info, ':', eval("dvc." + dvc_info)

