"""
@file mesh.py
Cartesian mesh class for local mesh.

"""
from parmepy.constants import np, PARMES_INTEGER, debug


class SubMesh(object):
    """
    Description of a local mesh, on one process of
    the topology.
    """

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, topo, g_start, resolution):

        ## Topology that creates (and owns) this mesh
        self._topology = topo
        ## Local resolution of this mesh, including ghost points
        self.resolution = resolution
        ## Dimension of the mesh
        self.dim = self.resolution.size
        ## index of the lowest point of this mesh
        ## (in each dir) in the global mesh
        self.global_start = g_start
        ## index of the upper point (in each dir), global mesh
        self.global_end = self.global_start + self.resolution - 1\
            - 2 * topo.ghosts
        ## Mesh step size
        self.space_step = topo.domain.length / \
            (topo.globalMeshResolution - 1)
        ## Mesh local indices, only for "computed" points
        ## (i.e. excluding ghosts)
        self.local_start = topo.ghosts.copy()
        self.local_end = self.resolution.copy() - topo.ghosts[:] - 1
        ## List of indices for computational points (i.e. without ghosts)
        ## usage : field[iCompute] returns the array field for all indices
        ## excluding ghost points.
        self.iCompute = []
        for d in range(self.dim):
            self.iCompute.append(slice(self.local_start[d],
                                       self.local_end[d] + 1))

        ## Coordinates of the "lowest" point of this mesh (including ghost)
        ## Warning FP : this strongly depends on where is the origin
        ## of the domain, of the type of boundary conditions
        ## and if origin is on ghosts or "real" points.
        self.origin = topo.domain.origin.copy()
        self.origin[:] += self.space_step[:] * \
            (self.global_start[:].astype(PARMES_INTEGER) - topo.ghosts[:])

        self.end = self.origin + self.space_step * (self.resolution - 1)
        if self.dim == 1:
            cx = np.linspace(self.origin, self.end, self.resolution)
            self.coords = tuple([cx, ])
        elif self.dim == 2:
            cx = np.linspace(self.origin[0], self.end[0],
                             self.resolution[0])[:, np.newaxis]
            cy = np.linspace(self.origin[1], self.end[1],
                             self.resolution[1])[np.newaxis, :]
            self.coords = tuple([cx, cy])
        elif self.dim == 3:
            cx = np.linspace(self.origin[0], self.end[0],
                             self.resolution[0])[:, np.newaxis, np.newaxis]
            cy = np.linspace(self.origin[1], self.end[1],
                             self.resolution[1])[np.newaxis, :, np.newaxis]
            cz = np.linspace(self.origin[2], self.end[2],
                             self.resolution[2])[np.newaxis, np.newaxis, :]
            self.coords = tuple([cx, cy, cz])

    def __str__(self):
        """ Sub mesh display """
        s = 'Coords:' + str(self._topology.coords[:])
        s += ' Sub mesh resolution:' + str(self.resolution) + '\n'
        s += 'Starting point global indices :' + str(self.global_start) + '\n'
        s += 'Local indices for "computed" points (excluding ghosts) :\n'
        s += '   start = ' + str(self.local_start)
        s += ', end = ' + str(self.local_end) + '\n'
        return s
