# -*- coding: utf-8 -*-
"""
@file differential_operations.py
Library of functions used to perform classical vector calculus
(diff operations like grad, curl ...)
"""
from parmepy.constants import debug, XDIR, YDIR, ZDIR
from abc import ABCMeta, abstractmethod
from finite_differences import FD_C_4
import numpy as np


class DifferentialOperation(object):
    __metaclass__ = ABCMeta

    ## @debug
    ## def __new__(cls, *args, **kw):
    ##     return object.__new__(cls, *args, **kw)

    @debug
    @abstractmethod
    def __init__(self, topo, method=None):
        pass


class Curl(DifferentialOperation):
    """
    Computes \f$ \nabla \ times (f) \f$, f being a vector field.
    """
    def __init__(self, topo, method=None):

        DifferentialOperation.__init__(self, topo)
        if method is not None and method.find('FD_order2') >= 0:
                raise ValueError("2nd order scheme Not yet implemented")
        else:
            assert (topo.ghosts >= 2).all(),\
                'you need a ghost layer for FD4 scheme.'
            self.fcall = self.FDCentral4
            self.fd_scheme = FD_C_4(topo.mesh.space_step)
            self.indices = topo.mesh.iCompute
            self.fd_scheme.computeIndices(self.indices)
        self.nbComponents = topo.domain.dimension

    def __call__(self, variable, data=None):
        return self.fcall(variable, data)

    def FDCentral4(self, variable, data=None):
        """
        @param variable : a list of numpy arrays
        """
        # Get slices for subarrays ...
       # init (if required) return data
        if data is None:
            data = []
            for i in xrange(self.nbComponents):
                data.append(variable[i].copy())
                #data[i][...] = 0.0

        #data[0][...] = 0.0
        #data[1][...] = 0.0
        #data[2][...] = 0.0

        #--  d/dy vz -- in data[XDIR]
        self.fd_scheme.inplace(variable[ZDIR], YDIR, data[XDIR])
        # -- -d/dz vy -- in data[YDIR]
        work = data[YDIR]
        self.fd_scheme.inplace(variable[YDIR], ZDIR, work)
        # data_x = d/dy vz - d/dz vy
        data[XDIR][self.indices] -= work[self.indices]

        #--  d/dz vx -- in data[YDIR]
        self.fd_scheme.inplace(variable[XDIR], ZDIR, data[YDIR])
        # -- -d/dx vz -- in data[ZDIR]
        work = data[ZDIR]
        self.fd_scheme.inplace(variable[ZDIR], XDIR, work)

        # data_y = d/dz vx - d/dx vz
        data[YDIR][self.indices] -= work[self.indices]

        # - d/dy vx in data[ZDIR]
        self.fd_scheme.inplace(variable[XDIR], YDIR, data[ZDIR])
        data[ZDIR][self.indices] *= -1.
        # add d/dx vy, add to data[ZDIR]
        self.fd_scheme.accumulate(variable[YDIR], XDIR, data[ZDIR])

        return data


class DivV(DifferentialOperation):
    """
    Computes \f$ \nabla.(\rho V) \f$, \f$ \rho\f$ a scalar field and
    V a vector field.
    """
    def __init__(self, topo, method=None):
        if method is not None and method.find('FD_order2') >= 0:
                raise ValueError("2nd order scheme Not yet implemented")
        elif method.find('FD_C4_nowork') >= 0:
            self.fcall = self.FDCentral4
            self.fd_scheme = FD_C_4((topo.mesh.space_step))
        else:
            self.fcall = self.FDCentral4Work
            self.fd_scheme = FD_C_4((topo.mesh.space_step))

        self.indices = topo.mesh.iCompute
        self.fd_scheme.computeIndices(self.indices)
        assert (topo.ghosts >= 2).all(),\
            'you need a ghost layer for FD4 scheme.'
        self.nbComponents = topo.domain.dimension

    def __call__(self, var1, scal, data=None, work=None):
        return self.fcall(var1, scal, data)

    def FDCentral4Work(self, var1, scal, data=None, work=None):
        """
        Compute central finite difference scheme, order 4.
        """
        if data is None:
            data = np.zeros_like(var1[0])
        if work is None:
            work = np.zeros_like(var1[0])

        #data[...] = 0.0

        #  d/dx (scal * v1x), saved in data
        self.fd_scheme.inplace(scal * var1[XDIR], XDIR, data)
        #  d/dy (scal * v1y), saved in work
        self.fd_scheme.inplace(scal * var1[YDIR], YDIR, work)
        data[self.indices] += work[self.indices]
        #  d/dz(scal * v1z), saved in work
        self.fd_scheme.inplace(scal * var1[ZDIR], ZDIR, work)
        data[self.indices] += work[self.indices]
        return data

    def FDCentral4(self, var1, scal, data=None):
        """
        """
        if data is None:
            data = np.zeros_like(var1[0])
        else:
            data[...] = 0.0

        #  d/dx (scal * v1x)
        self.fd_scheme.accumulate(var1[XDIR] * scal, XDIR, data)
        #  d/dy (scal * v1y)
        self.fd_scheme.accumulate(var1[YDIR] * scal, YDIR, data)
        #  d/dz(scal * v1z)
        self.fd_scheme.accumulate(var1[ZDIR] * scal, ZDIR, data)

        return data


class DivT(DifferentialOperation):
    """
    Computes \f$ \nabla.(wx.V,wy.V,wz.V) \f$, \f$w\f$ and V some vector fields.
    """
    def __init__(self, topo, method=None):
        if method is not None and method.find('FD_order2') >= 0:
                raise ValueError("2nd order scheme Not yet implemented")
        else:
            self.fcall = self.FDCentral4
            self.div = DivV(topo, method='FD_C4')
            self.divZ = DivV(topo, method='FD_C4_nowork')
        self.nbComponents = topo.domain.dimension

    def __call__(self, var1, var2, data=None):
        return self.fcall(var1, var2, data)

    def FDCentral4(self, var1, var2, data=None):
        """
        """
        # init (if required) return data
        if data is None:
            data = []
            for i in xrange(self.nbComponents):
                data.append(np.zeros_like(var1[i]))

        # we use data[Y] and data[Z] as work ...
        data[XDIR] = self.div(var1, var2[XDIR],
                              data=data[XDIR], work=data[YDIR])
        data[YDIR] = self.div(var1, var2[YDIR],
                              data=data[YDIR], work=data[ZDIR])
        # No work vector for zdir
        data[ZDIR] = self.divZ(var1, var2[ZDIR], data=data[ZDIR])
        return data


class GradS(DifferentialOperation):
    """
    Computes \f$ \nabla(\rho) \f$, \f$ \rho\f$ a scalar field
    """
    def __init__(self, topo, method=None):
        if method is not None and method.find('FD_order2') >= 0:
                raise ValueError("2nd order scheme Not yet implemented")
        elif method.find('FD_C4') >= 0:
            self.fcall = self.FDCentral4
            self.fd_scheme = FD_C_4((topo.mesh.space_step))
        else:
            raise ValueError("FD scheme Not yet implemented")

        self.indices = topo.mesh.iCompute
        self.fd_scheme.computeIndices(self.indices)
        assert (topo.ghosts >= 2).all(),\
            'you need a ghost layer for FD4 scheme.'
        ## number of components for vector fields
        self.nbComponents = topo.domain.dimension

    def __call__(self, scal, data=None, work=None):
        return self.fcall(scal, data)

    def FDCentral4(self, scal, data=None):
        """
        Compute central finite difference scheme, order 4.
        @param scal a numpy array
        """
        if data is None:
            data = []
            for i in xrange(self.nbComponents):
                data.append(np.zeros_like(scal))
        #data[...] = 0.0

        #  d/dx (scal), saved in data
        self.fd_scheme.inplace(scal, XDIR, data[XDIR])
        #  d/dy (scal), saved in data
        self.fd_scheme.inplace(scal, YDIR, data[YDIR])
        #  d/dz(scal), saved in data
        self.fd_scheme.inplace(scal, ZDIR, data[ZDIR])
        return data


class GradVxW(DifferentialOperation):
    """
    Computes \f$ [\nabla(V)][W] \f$, with
    \f$V\f$ and \f$W\f$ some vector fields.
    """
    def __init__(self, topo, method=None):
        if method is not None and method.find('FD_order2') >= 0:
                raise ValueError("2nd order scheme Not yet implemented")
        else:
            self.fcall = self.FDCentral4
            self.grad = GradS(topo, method='FD_C4')

            self.divZ = DivV(topo, method='FD_C4_nowork')
        self.nbComponents = topo.domain.dimension

        if self.nbComponents == 1:
            self.worklength = 1
        elif self.nbComponents == 2:
            self.worklength = 3
        else:
            self.worklength = 5
        self.prod = BlockProd()

    def __call__(self, var1, var2, diagnostics=None, data=None):
        return self.fcall(var1, var2, diagnostics, data)

    def FDCentral4(self, var1, var2, diagnostics=None, data=None):
        """
        """
        # init (if required) return data
        if data is None:
            data = []
            for i in xrange(self.worklength):
                data.append(np.zeros_like(var1[0]))
        else:
            assert len(data) == self.worklength
        if diagnostics is not None:
            diagnostics[:] = 0.0

        for direc in xrange(self.nbComponents):
            data[direc:self.nbComponents + direc] = \
                self.grad(var1[direc],
                          data=data[direc:self.nbComponents + direc])

            if diagnostics is not None:
                diagnostics[0] = max(diagnostics[0],
                                     sum([abs(data[i])
                                          for i in xrange(direc,
                                                          self.nbComponents
                                                          + direc)]))

                # compute grad(Vx).var2
                data[direc] = self.prod(data[direc:self.nbComponents + direc],
                                        var2, data[direc])
        return data


class BlockProd(object):
    """
    Block product of two lists of numpy arrays
    """
    def __init__(self):
        pass

    def __call__(self, list1, list2, data=None):
        assert len(list1) == len(list2)

        if data is not None:
            assert data.shape == list1[0].shape
            data[...] = 0.0
        else:
            data = np.zeros_like(list1[0])

        for i in xrange(len(list1)):
            data += list1[i] * list2[i]
        return data
