# -*- coding: utf-8 -*-
"""
@file fct2op.py

Convert differential operator for stretching into a function
to Discrete time integration
"""
from parmepy.numerics.differentialOperator import DifferentialOperator
import numpy as np


class Fct2Op(object):
    """
    Fct2Op
    make the link between differentialOperator and
    the integration time scheme (Euler,RK,...)
    """

    def __init__(self, field1, field2, method='', choice='', topology=None):
        """
        Constructor.
        store field2 to work with it in the apply function

        @param field1 : vorticity field
        @param field2 : velocity field or grad(U)
            depending on 'choice' parameter
        @param choice : can be 'gradV' for GradU.w or 'divWU' for div(wu)
        """
        self.field1 = field1
        self.field2 = field2
        self.method = method
        self.choice = choice
        self.topology = topology

    def apply(self, t, y):
        """
        Apply operator.
        Fct2Op.apply is used like a def function
        """
#        self.synchro.apply()
        if self.choice == 'gradV':  # field2 = gradU
            temp0 = y[0][...] * 0.
            temp1 = y[0][...] * 0.
            temp2 = y[0][...] * 0.

            ind0a = self.topology.mesh.local_start[0]
            ind0b = self.topology.mesh.local_end[0] + 1
            ind1a = self.topology.mesh.local_start[1]
            ind1b = self.topology.mesh.local_end[1] + 1
            ind2a = self.topology.mesh.local_start[2]
            ind2b = self.topology.mesh.local_end[2] + 1
            ind0 = np.arange(ind0a, ind0b)
            ind1 = np.arange(ind1a, ind1b)
            ind2 = np.arange(ind2a, ind2b)

            ## Perform Matrix/Scalar product on local mesh
            temp0[ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] =\
                self.field2[0][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] *\
                y[0][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] +\
                self.field2[1][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] *\
                y[1][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] +\
                self.field2[2][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] *\
                y[2][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b]
            temp1[ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] =\
                self.field2[3][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] *\
                y[0][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] +\
                self.field2[4][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] *\
                y[1][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] +\
                self.field2[5][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] *\
                y[2][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b]
            temp2[ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] =\
                self.field2[6][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] *\
                y[0][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] +\
                self.field2[7][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] *\
                y[1][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] +\
                self.field2[8][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] *\
                y[2][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b]
#            result = np.concatenate((
#                np.array([temp0]),np.array([temp1]),np.array([temp2])))
            return temp0, temp1, temp2

        if self.choice == 'divWU':  # field2 = velocity field
            result = DifferentialOperator(
                y, self.field2, self.method, self.choice).__call__()
            return result[0], result[1], result[2]

        if self.choice == 'hessianU':  # field2 = velocity field
            gradU, maxgersh = DifferentialOperator(
                self.field1, self.field2,
                self.method, choice='gradV').__call__()
            GradUline1 = np.concatenate(([gradU[0]], [gradU[1]], [gradU[2]]))
            GradUline2 = np.concatenate(([gradU[3]], [gradU[4]], [gradU[5]]))
            GradUline3 = np.concatenate(([gradU[6]], [gradU[7]], [gradU[8]]))
            result1, maxgersh = DifferentialOperator(
                self.field1, GradUline1,
                self.method, choice='gradV').__call__()
            result2, maxgersh = DifferentialOperator(
                self.field1, GradUline2,
                self.method, choice='gradV').__call__()
            result3, maxgersh = DifferentialOperator(
                self.field1, GradUline3,
                self.method, choice='gradV').__call__()
            return result1, result2, result3
        else:
            print 'choice', choice, 'in Fct2Op is not defined'

    def __str__(self):
        s = "fct2Op(DiscreteOperator). " + DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Fct2Op"
    print Fct2Op.__doc__
