"""
@file runge_kutta2.py

RK2 method interface.
"""
from parmepy.numerics.integrators.odesolver import ODESolver


class RK2(ODESolver):
    """
    ODESolver implementation for solving an equation system with RK2 method.
    y'(t) = f(t,y)

    y(t_n+1)= y(t_n) + dt*y'[y(t_n)+dt/2*y'(y(t_n))].

    """
    def __init__(self, dim, f=None):
        """
        @param dim, vector field (f) dimension (number of component)
        @param f function f(t,y) : Right hand side of the equation to solve.
        """
        ODESolver.__init__(self, dim, f)
        self.name = 'RK2'

    def __call__(self, y, dt, t):
        # k1 = f(t,y)
        # k2 = f(t + dt/2, y + dt/2 * k1)
        # result = y + dt * k2

        assert len(y) == self._dim
        # k1 ... in k2
        k2 = self.f(t, y)
        # k2
        k2 = [y[i] + dt / 2 * k2[i] for i in xrange(self._dim)]
        k2 = self.f(t + dt / 2., k2)

        return [y[i] + dt * k2[i] for i in xrange(self._dim)]


## class RK2_scalar(ODESolver):
##     """
##     ODESolver implementation for solving an equation system with RK2 method
##     for scalar quantities.
##     y'(t) = f(t,y)

##     y(t_n+1)= y(t_n) + dt*y'[y(t_n)+dt/2*y'(y(t_n))].

##     """
##     def __init__(self, f=None):
##         """
##         Constructor.

##         @param f function f(t,y) : Right hand side of the equation to solve.
##         @param dim : dimensions
##         """
##         ODESolver.__init__(self, f)
##         self.name = 'RK2'

##     def __call__(self, f, t, dt, y, res=None, work=None, yprime=None,
##                  **f_kwargs):

##         """
##         Integration step for RK2 Method.
##          up = f[t, Y(t,space)].

##         @param f : right-hand side.
##         @param t : current time.
##         @param dt : time step.
##         @param y : function Y(t,space).
##         @param res : result (may be allocated when not None).
##         @param work : allocated working array.
##         @param yprime : y'(t) already computed (avoid 1st call to f)
##         @param f_kwargs : Parameters for right-hand side computing.
##         """
##         if res is None:
##             res = np.empty_like(y)
##         if work is None:
##             work = np.empty_like(y)
##         if yprime is None:
##             work[...] = f(t, y, res=work, **f_kwargs)
##         else:
##             work[...] = yprime
##         work[...] *= 0.5 * dt
##         res[...] = y
##         res[...] += dt * f(t + 0.5 * dt, y + work, **f_kwargs)

##     def __str__(self):
##         """ToString method"""
##         return "RK2 Method"


if __name__ == "__main__":
    print __doc__
    print "- Provided class : RK2"
    print RK2.__doc__
