# -*- coding: utf-8 -*-
"""
@file runge_kutta4.py

RK4 method interface.
"""
from parmepy.numerics.integrators.odesolver import ODESolver


class RK4(ODESolver):
    """
    ODESolver implementation for solving an equation system with RK4 method.
    """
    def __init__(self, dim, f=None):
        """
        @param f function f(t,y) : Right hand side of the equation to solve.
        @param dim : dimensions
        """
        ODESolver.__init__(self, dim, f)
        self.name = 'RK4'

    def __call__(self, y, dt, t):

        # k1 = f(t,y)
        # k2 = f(t + dt/2, y + dt/2 * k1)
        # k3 = f(t + dt/2, y + dt/2 * k2)
        # k4 = f(t + dt, y + dt * k3)
        # result = y + dt/6( k1 + 2 * k2 + 2 * k3 + k4)

        # k1 in result
        result = self.f(t, y)
        # k2 in work
        work = [y[i] + dt / 2 * result[i] for i in xrange(self._dim)]
        work = self.f(t + dt / 2, work)

        # k1 + 2 * k2
        for i in xrange(self._dim):
            result[i] += 2 * work[i]
        # k3 in work
        work = [y[i] + dt / 2 * work[i] for i in xrange(self._dim)]
        work = self.f(t + dt / 2, work)
        # k1 + 2 * k2 + 2 * k3
        for i in xrange(self._dim):
            result[i] += 2 * work[i]
        # k4 in work
        work = [y[i] + dt * work[i] for i in xrange(self._dim)]
        work = self.f(t + dt, work)
        # k1 + 2 * k2 + 2 * k3 + k4
        for i in xrange(self._dim):
            result[i] += work[i]
            result[i] *= dt / 6
            result[i] += y[i]

        return result

    def __str__(self):
        """ToString method"""
        return "RK4 Method"

class RK4_scalar(ODESolver):
    """
    ODESolver implementation for solving an equation system with RK4 method
    for scalar quantities.
    y'(t) = f(t,y)

    y(t_n+1)= y(t_n) + dt*y'[y(t_n)+dt/2*y'(y(t_n))].

    """
    def __init__(self, f=None):
        """
        Constructor.

        @param f function f(t,y) : Right hand side of the equation to solve.
        @param dim : dimensions
        """
        ODESolver.__init__(self, f)
        self.name = 'RK4'

    def __call__(self, t, dt, y, res=None, work=None, y_temp=None,
                 yprime=None, **f_kwargs):

        """
        Integration step for RK4 Method.
         up = f[t, Y(t,space)].

        @param t : current time.
        @param dt : time step.
        @param y : function Y(t,space).
        @param res : result (may be allocated when not None).
        @param work : allocated working array (list of 3 like y arrays).
        @param yprime : y'(t) already computed (avoid 1st call to f)
        @param f_kwargs : Parameters for right-hand side computing.
        """
        if res is None:
            res = np.empty_like(y)
        if work is None:
            work = [np.empty_like(y) for i in xrange(3)]
        if y_temp is None:
            y_temp = np.empty_like(y)
        if yprime is None:
            yprime = np.empty_like(y)
            yprime[...] = self.f(t, y, res=yprime, **f_kwargs)

        y_temp[...] = y + 0.5 * dt * yprime
        work[0][...] = self.f(t + 0.5 * dt, y_temp, res=work[0], **f_kwargs)

        y_temp[...] = y + 0.5 * dt * work[0]
        work[1][...] = self.f(t + 0.5 * dt, y_temp, res=work[1], **f_kwargs)

        y_temp[...] = y + dt * work[1]
        work[2][...] = self.f(t + dt, y_temp, res=work[2], **f_kwargs)
        res[...] = y
        res[...] += (yprime + 2. * work[0] + 2. * work[1] + work[2]) * dt / 6.

    def __str__(self):
        """ToString method"""
        return "RK4 Method"

if __name__ == "__main__":
    print __doc__
    print "- Provided class : RK4"
    print RK4.__doc__
