"""
@file interpolation.py
"""
from parmepy.constants import np
from parmepy.numerics.method import NumMethod


class Linear(NumMethod):
    """Remshing with M6prime function"""

    def __init__(self, f, d, dx, origin):
        NumMethod.__init__(self)
        self.name = 'LinearInterpolation'
        self.field = f
        self.d = d
        self.dx = dx
        self.origin = origin

    def __call__(self, t, x, res=None,
                 i_row=None, i_col=None, i_dep=None, i_y=None):
        """
        Remesh particles at position ppos with scalar
        pscal along direction d.

        @param x : position
        @param t : time
        @param res : Result variable
        @param i_row : allocated working array (rows indices array).
        @param i_col : allocated working array (columns indices array).
        @param i_dep : allocated working array (depth indices array).
        @param i_y : allocated working array (distance to left grid point).
        """
        if res is None:
            res = np.empty_like(x)
        if i_y is None:
            i_y = np.empty_like(x)
        if len(self.field.data[self.d].shape) == 2:
            if i_row is None or i_col is None:
                i_row, i_col = np.indices(x.shape)
            if self.d == 0:
                floor = res
                floor[...] = (x - self.origin[self.d]) / self.dx[self.d]
                i_y[...] = floor
                floor[...] = np.floor(floor)
                i_y[...] -= floor
                i_row[...] = floor.astype(np.int32) % x.shape[self.d]
                i_col[...] = np.indices((x.shape[1],))[0][np.newaxis, :]
                res[...] = self.field.data[self.d][i_row, i_col] * (1. - i_y)
                i_row[...] = (i_row + 1) % x.shape[self.d]
                res[...] += self.field.data[self.d][i_row, i_col] * i_y
            else:
                floor = res
                floor[...] = (x - self.origin[self.d]) / self.dx[self.d]
                i_y[...] = floor
                floor[...] = np.floor(floor)
                i_y[...] -= floor
                i_col[...] = floor.astype(np.int32) % x.shape[self.d]
                i_row[...] = np.indices((x.shape[0],))[0][:, np.newaxis]
                res[...] = self.field.data[self.d][i_row, i_col] * (1. - i_y)
                i_col[...] = (i_col + 1) % x.shape[self.d]
                res[...] += self.field.data[self.d][i_row, i_col] * i_y

        elif len(self.field.data[self.d].shape) == 3:
            if i_row is None or i_col is None or i_dep is None:
                i_row, i_col, i_dep = np.indices(x.shape)
            if self.d == 0:
                floor = res
                floor[...] = (x - self.origin[self.d]) / self.dx[self.d]
                i_y[...] = floor
                floor[...] = np.floor(floor)
                i_y[...] -= floor
                i_row[...] = floor.astype(np.int32) % x.shape[self.d]
                i_col[...] = \
                    np.indices((x.shape[1],))[0][np.newaxis, :, np.newaxis]
                i_dep[...] = \
                    np.indices((x.shape[2],))[0][np.newaxis, np.newaxis, :]
                res[...] = \
                    self.field.data[self.d][i_row, i_col, i_dep] * (1. - i_y)
                i_row[...] = (i_row + 1) % x.shape[self.d]
                res[...] += \
                    self.field.data[self.d][i_row, i_col, i_dep] * i_y
            elif self.d == 1:
                floor = res
                floor[...] = (x - self.origin[self.d]) / self.dx[self.d]
                i_y[...] = floor
                floor[...] = np.floor(floor)
                i_y[...] -= floor
                i_col[...] = floor.astype(np.int32) % x.shape[self.d]
                i_row[...] = \
                    np.indices((x.shape[0],))[0][:, np.newaxis, np.newaxis]
                i_dep[...] = \
                    np.indices((x.shape[2],))[0][np.newaxis, np.newaxis, :]
                res[...] = \
                    self.field.data[self.d][i_row, i_col, i_dep] * (1. - i_y)
                i_col[...] = (i_col + 1) % x.shape[self.d]
                res[...] += \
                    self.field.data[self.d][i_row, i_col, i_dep] * i_y
            else:
                floor = res
                floor[...] = (x - self.origin[self.d]) / self.dx[self.d]
                i_y[...] = floor
                floor[...] = np.floor(floor)
                i_y[...] -= floor
                i_dep[...] = floor.astype(np.int32) % x.shape[self.d]
                i_row[...] = \
                    np.indices((x.shape[0],))[0][:, np.newaxis, np.newaxis]
                i_col[...] = \
                    np.indices((x.shape[1],))[0][np.newaxis, :, np.newaxis]
                res[...] = \
                    self.field.data[self.d][i_row, i_col, i_dep] * (1. - i_y)
                i_dep[...] = (i_dep + 1) % x.shape[self.d]
                res[...] += \
                    self.field.data[self.d][i_row, i_col, i_dep] * i_y

        else:
            if i_row is None:
                i_row = np.indices(x.shape)[0]
            floor = res
            floor[...] = (x - self.origin[self.d]) / self.dx[self.d]
            i_y[...] = floor
            floor[...] = np.floor(floor)
            i_y[...] -= floor
            i_row[...] = floor.astype(np.int32) % x.shape[self.d]
            res[...] = self.field.data[i_row] * (1. - i_y)
            i_row[...] = (i_row + 1) % x.shape[self.d]
            res[...] += self.field.data[i_row] * i_y

        return res
