"""
@file advection.py

Advection of a field.
"""
from parmepy.constants import debug, np, PARMES_INDEX
from parmepy.operator.continuous import Operator
from parmepy.tools.timers import Timer
from parmepy.fields.continuous import Field
from parmepy.mpi import main_size
from parmepy.operator.redistribute import Redistribute


class Advection(Operator):
    """
    Advection of a field,
    \f{eqnarray*}
    X = Op(X,velocity)
    \f} for :
    \f{eqnarray*}
    \frac{\partial{X}}{\partial{t}} + velocity.\nabla X = 0
    \f}
    Note : we assume incompressible flow.

    Computations are performed within a dimensional splitting as folows:
      - 2nd order:
        - X-dir, half time step
        - Y-dir, half time step
        - Z-dir, full time step
        - Y-dir, half time step
        - X-dir, half time step
      - 2nd order full half-steps:
        - X-dir, half time step
        - Y-dir, half time step
        - Z-dir, half time step
        - Z-dir, half time step
        - Y-dir, half time step
        - X-dir, half time step
      - 1st order g:
        - X-dir, half time step
        - Y-dir, half time step
        - Z-dir, half time step

    """

    @debug
    def __init__(self, velocity, advectedFields, resolutions, method='',
                 topo=None, ghosts=None, splittingConfig='o2', **other_config):
        """
        Create a Transport operator from given variables velocity and scalar.

        @param velocity : velocity variable.
        @param advectedFields : Advected fields (may be a list of Fields).
        @param resolutions : list of resolutions (one per variable)
        @param method : Method used
        @param splittingConfig : Dimensional splitting configuration
        (default 'o2')
        @param other_config : Method specific parameters
        @param topo : a predefined topology to discretize variables
        """
        v = [velocity]
        if isinstance(advectedFields, list):
            [v.append(f) for f in advectedFields]
            self.output = advectedFields
        else:
            v.append(advectedFields)
            self.output = [advectedFields]
        Operator.__init__(self, v, method, topo=topo,
                          ghosts=ghosts, name="Advection")
        ## Transport velocity
        self.velocity = velocity
        ## Transported fields
        self.advectedFields = advectedFields
        ## Resolutions of the fields. Dictionnary with
        ## fields as keys.
        ## Example :
        ## resolutions[velocity] may return [32,32,32].
        self.resolutions = resolutions
        ## Extra parameters (depend on the method)
        self.config = other_config
        self.input = [self.velocity]
        self._isSCALES = False
        self._old_dir = 0
        dim = self.velocity.dimension
        self.splitting = []
        if splittingConfig == 'o2_FullHalf':
            ## Half timestep in all directions
            [self.splitting.append((i, 0.5)) for i in xrange(dim)]
            [self.splitting.append((dim - 1 - i, 0.5)) for i in xrange(dim)]
        elif splittingConfig == 'o1':
            [self.splitting.append((i, 1.)) for i in xrange(dim)]
        elif splittingConfig == 'x_only':
            self.splitting.append((0, 1.))
        elif splittingConfig == 'y_only':
            self.splitting.append((1, 1.))
        elif splittingConfig == 'z_only':
            self.splitting.append((2, 1.))
        else:
            ## Half timestep in all directions but last
            [self.splitting.append((i, 0.5)) for i in xrange(dim - 1)]
            self.splitting.append((dim - 1, 1.))
            [self.splitting.append((dim - 2 - i, 0.5))
             for i in xrange(dim - 1)]

    @debug
    def setUp(self):
        """
        Discretisation according to the chosen method.
        Available methods :
        - 'scales' : SCALES fortran routines (3d only, list of vector
        and/or scalar)
          - 'p_O2' : order 4 method, corrected to allow large CFL number,
          untagged particles
          - 'p_O4' : order 4 method, corrected to allow large CFL number,
          untagged particles
          - 'p_L2' : limited and corrected lambda 2
          - 'p_M4' : M4prime formula
          - 'p_M6' (default) : M6prime formula
          - 'p_M8' : M8prime formula
        - 'gpu' : OpenCL kernels (2d and 3d, single field, scalar or vector)
          - '1k' : Single OpenCL kernel for advection and remeshing
          - '2k' : Separate kernels
          - 'm4prime' : M4prime formula
          - 'm6prime' : M6prime formula
          - 'm8prime' : M8prime formula
          - 'l6star' : Lambda6star formula
        - other : Pure python (2d and 3d, list of vector and/or scalar)
          - 'rk2' : Runge Kutta 2nd order advection
          - 'm4prime' : M4prime formula
          - 'm6prime' : M6prime formula
          - 'm8prime' : M8prime formula
          - 'l6star' : Lambda6star formula
        """

        # --- Advection solver from SCALES ---
        if self.method.find('scales') >= 0:
            if not self.advectedFields.dimension == 3:
                    raise ValueError("Scales Advection not implemented in 2D.")
            if self.ghosts is not None:
                if (self.ghosts > 0).any():
                    raise ValueError("Scales Advection not\
                    yet implemented with ghosts points.")
            # - Scales imports -
            from parmepy.operator.discrete.scales_advection import \
                ScalesAdvection
            from parmepy.f2py import scales2py as scales
            self._isSCALES = True

            # - Extract order form self.method (default p_M6) -
            order = 'p_M6'
            for o in ['p_O2', 'p_O4', 'p_L2', 'p_M4', 'p_M6', 'p_M8']:
                if self.method.find(o) >= 0:
                    order = o

            # - Create the topologies (get param from scales) -
            # Scales nbcells equals resolutions - 1
            nbcells = np.asarray(self.resolutions[self.advectedFields],
                                 dtype=PARMES_INDEX) - 1
            topodims = [1, 1, main_size]
            scalesres, scalesoffset, stab_coeff = \
                scales.init_advection_solver(nbcells,
                                             self.domain.length,
                                             topodims, order=order)
            # Use same topodims as scales to create Cartesian topology
            # in order to discretize our fields
            if self._predefinedTopo is not None:
                topo = self._predefinedTopo
                assert (topo.shape == topodims).all(), 'input topology is\
                    not scales compliant.'
                for v in self.variables:
                    self.discreteFields[v] = v.discretize(topo)
            else:  # Note Franck : is it reasonable to allow different
                # topologies in the same
                # advection operator???
                for v in self.variables:
                    # the topology for v ...
                    topo = \
                        self.domain.getOrCreateTopology(self.domain.dimension,
                                                        self.resolutions[v],
                                                        topodims,
                                                        ghosts=self.ghosts)
                    # ... and the corresponding discrete field
                    self.discreteFields[v] = v.discretize(topo)

            advectedFields = [self.discreteFields[f]
                              for f in self.discreteFields.keys()
                              if f is not self.velocity]

            # - Create the discreteOperator from the list of discrete fields -
            self.discreteOperator = ScalesAdvection(
                self.discreteFields[self.velocity],
                advectedFields, method=self.method,
                **self.config)

            # -- Final set up --
            self.discreteOperator.setUp()
            self._isUpToDate = True

        # --- GPU or pure-python advection ---
        else:
            if main_size == 1:
                # - topologies creation and variables discretization -
                if self._predefinedTopo is not None:
                    topo = self._predefinedTopo
                    for v in self.variables:
                        self.discreteFields[v] = v.discretize(topo)
                else:
                    for v in self.variables:
                        topo = self.domain.getOrCreateTopology(
                            self.domain.dimension, self.resolutions[v])
                        self.discreteFields[v] = v.discretize(topo)

                advectedFields = [self.discreteFields[f]
                                  for f in self.discreteFields.keys()
                                  if f is not self.velocity]
                particles_advectedFields = [
                    Field(adF.topology.domain, "Particle_AdvectedFields",
                          isVector=adF.isVector).discretize(adF.topology)
                    for adF in advectedFields]

                if self.method.find('gpu_2k') >= 0:
                    from parmepy.gpu.gpu_particle_advection_2k \
                        import GPUParticleAdvection2k as ParticleAdvection
                    particles_positions = \
                        Field(advectedFields[0].topology.domain,
                              "Particle_Position",  isVector=False
                              ).discretize(advectedFields[0].topology)
                elif self.method.find('gpu_1k') >= 0:
                    from parmepy.gpu.gpu_particle_advection_1k \
                        import GPUParticleAdvection1k as ParticleAdvection
                    particles_positions = None
                else:
                    from parmepy.operator.discrete.particle_advection \
                        import ParticleAdvection
                    particles_positions = \
                        Field(advectedFields[0].topology.domain,
                              "Particle_Position",  isVector=False
                              ).discretize(advectedFields[0].topology)

                self.discreteOperator = [
                    ParticleAdvection(
                        self.discreteFields[self.velocity],
                        advectedFields, d, method=self.method,
                        part_position=particles_positions,
                        part_advectedFields=particles_advectedFields,
                        **self.config)
                    for d in xrange(self.velocity.dimension)]
                # -- Final set up --
                for dop in self.discreteOperator:
                    dop.setUp()
                self._isUpToDate = True
            else:
                if self._predefinedTopo is not None:
                    raise ValueError("User-defined topology is not\
                    (yet) allowed for advection.")
                self.discreteOperator = self.domain.dimension * [None]
                for d in xrange(self.domain.dimension):
                    topodims = np.ones((self.domain.dimension))
                    if d == self.domain.dimension - 1:
                        topodims[0] = main_size
                    else:
                        topodims[-1] = main_size
                    for v in self.variables:
                        topo = self.domain.getOrCreateTopology(
                            self.domain.dimension, self.resolutions[v],
                            topoResolution=topodims)
                        self.discreteFields[v] = v.discretize(topo)

                    advectedFields = [self.discreteFields[f]
                                      for f in self.discreteFields.keys()
                                      if f is not self.velocity]
                    particles_advectedFields = [
                        Field(adF.topology.domain, "Particle_AdvectedFields",
                              isVector=adF.isVector).discretize(adF.topology)
                        for adF in advectedFields]

                    if self.method.find('gpu_2k') >= 0:
                        from parmepy.gpu.gpu_particle_advection_2k \
                            import GPUParticleAdvection2k as ParticleAdvection
                        particles_positions = \
                            Field(advectedFields[0].topology.domain,
                                  "Particle_Position",  isVector=False
                                  ).discretize(advectedFields[0].topology)
                    elif self.method.find('gpu_1k') >= 0:
                        from parmepy.gpu.gpu_particle_advection_1k \
                            import GPUParticleAdvection1k as ParticleAdvection
                        particles_positions = None
                    else:
                        print "Using default advection operator"
                        from parmepy.operator.discrete.particle_advection \
                            import ParticleAdvection
                        particles_positions = \
                            Field(advectedFields[0].topology.domain,
                                  "Particle_Position",  isVector=False
                                  ).discretize(advectedFields[0].topology)

                    self.discreteOperator[d] = ParticleAdvection(
                        self.discreteFields[self.velocity],
                        advectedFields, d, method=self.method,
                        part_position=particles_positions,
                        part_advectedFields=particles_advectedFields,
                        **self.config)

                # -- Final set up --
                for dop in self.discreteOperator:
                    dop.setUp()

                # # Build bridges
                self.bridges = self.domain.dimension * [None]
                for dfrom in xrange(self.domain.dimension):
                    self.bridges[dfrom] = self.domain.dimension * [None]
                    for dto in xrange(self.domain.dimension):
                        if dfrom == dto:
                            self.bridges[dfrom][dto] = None
                        else:
                            self.bridges[dfrom][dto] = Redistribute(
                                self.variables,
                                self.discreteOperator[dfrom],
                                self.discreteOperator[dto],
                                method='FromDiscreteOperators')
                            self.bridges[dfrom][dto].setUp()
                self._isUpToDate = True

    @debug
    def apply(self, simulation=None):
        """
        Apply this operator to its variables.
        @param simulation : object that describes the simulation
        parameters (time, time step, iteration number ...), see
        parmepy.problem.simulation.Simulation for details.

        Redefinition for advection. Applying a dimensional splitting.
        """
        if self._isSCALES:
            self.discreteOperator.apply(simulation)
        else:
            if main_size == 1:  # Sequential
                for split_id, split in enumerate(self.splitting):
                    self.discreteOperator[split[0]].apply(
                        simulation, split[1], split_id, self._old_dir)
                    self._old_dir = split[0]
            else:  # Parallel, must use bridges to exchange data
                for split_id, split in enumerate(self.splitting):
                    if not self._old_dir == split[0]:
                        self.bridges[self._old_dir][split[0]].apply()
                        self.bridges[self._old_dir][split[0]].wait()
                    self.discreteOperator[split[0]].apply(
                        simulation, split[1], split_id, self._old_dir)
                    self._old_dir = split[0]

    @debug
    def finalize(self):
        """
        Memory cleaning.
        """
        self.timer = Timer(self)
        for dop in self.discreteOperator:
            dop.finalize()
            self.timer = self.timer + dop.timer
        print self.timer


if __name__ == "__main__":
    print __doc__
    print "- Provided class : Advection"
    print Advection.__doc__
