"""
@file operator/analytic.py

Analytic operator representation.
"""
from parmepy.constants import debug
from parmepy.operator.continuous import Operator
from parmepy.operator.discrete.analytic import Analytic_D


class Analytic(Operator):
    """
    Applies an analytic formula, given by user, on its fields.
    """

    @debug
    def __init__(self, variables, resolutions, formula=None, method=None,
                 ghosts=None, topo=None, **other_config):
        """
        Create an operator using an analytic formula to compute field(s).
        @param variables : list of fields on which this operator will apply.
        @param formula : the formula to be applied
        @param resolutions : list of resolutions (one per variable)
        @param method : Method used
        @param other_config : Method specific parameters
        @param topo : a predefined topology to discretize variables
        @param ghosts : number of points in the ghost layer
        """

        Operator.__init__(self, variables, method, ghosts=ghosts,
                          topo=topo, name="Analytic")

        if formula is not None:
            ## A formula applied to all variables of this operator
            self.formula = formula
            for v in self.variables:
                v.setFormula(formula)
        else:
            self.formula = self.variables[0]._formula
            for v in self.variables:
                assert v._formula is self.formula

        if self.formula is None:
            if self.method is None or self.method.find('gpu') == -1:
                raise ValueError("A formula must be set.")

        ## Dict of resolutions (one per variable)
        self.resolutions = resolutions
        self.config = other_config
        self.output = self.variables
 
    @debug
    def setUp(self):
        """
        Set up of Analytic operator.
        """

        if self._predefinedTopo is not None:
            topo = self._predefinedTopo
            for v in self.variables:
                self.discreteFields[v] = v.discretize(topo)

        else:
            # Variables discretization
            for v in self.variables:
                # the topology for v ...
                topo = self.domain.getOrCreateTopology(self.domain.dimension,
                                                       self.resolutions[v],
                                                       ghosts=self.ghosts)
                # ... and the corresponding discrete field
                self.discreteFields[v] = v.discretize(topo)

        # Select and create discrete operator
        if self.method is not None and self.method.find('gpu') >= 0:
            from parmepy.gpu.gpu_analytic import GPUAnalytic
            self.discreteOperator = GPUAnalytic(self.discreteFields.values(),
                                                method=self.method,
                                                **self.config)
        else:
            self.discreteOperator = Analytic_D(self.discreteFields.values(),
                                               self.formula,
                                               method=self.method,
                                               **self.config)
        self.discreteOperator.setUp()
        self._isUpToDate = True

    def __str__(self):
        """ToString method"""
        s = super(Analytic, self).__str__()
        if self.discreteOperator is None:
            s += "Formula = " + str(self.formula)
        return s + "\n"

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Analytic"
    print Analytic.__doc__
