"""
@file operator/curl.py

Curl of a field.
"""
from parmepy.constants import debug, np
from parmepy.operator.continuous import Operator
from parmepy.operator.discrete.curl import CurlFD
from parmepy.operator.updateGhosts import UpdateGhosts


class Curl(Operator):
    """
    Computes \f$ outVar = \nabla \times inVar \f$
    """

    @debug
    def __init__(self, invar, outvar, resolution, method=None, ghosts=None,
                 topo=None):
        """
        Create an operator to compute the curl of a field.
        @param invar : the input vector field
        @param outvar : curl of input vector field
        @param resolution : resolution (global) of the fields
        @param method : method used (default = finite difference, 4th order)
        @param topo : a predefined topology to discretize invar/outvar
        """

        Operator.__init__(self, [invar, outvar], method, ghosts=ghosts,
                          topo=topo)

        ## Dict of resolutions (one per variable)
        self.resolution = resolution
        self.output = [outvar]
        self.input = [invar]
        self.invar = invar
        self.outvar = outvar
        if self.method is None:
            self.method = 'FD_C4'
        ## An operator to update ghost points of input var.
        self.ghostsOperator = None

    @debug
    def setUp(self):
        if self.method.find('FD_order2') >= 0:
            nbGhosts = 1
        elif self.method.find('FD_C4') >= 0:
            nbGhosts = 2
        else:
            nbGhosts = 0

        # get (or create) the topology
        if self._predefinedTopo is not None:
            assert (self._predefinedTopo.ghosts >= nbGhosts).all()
            topo = self._predefinedTopo

        else:
            # same topo for all variables
            if self.ghosts is not None:
                self.ghosts[self.ghosts < nbGhosts] = nbGhosts
            else:
                self.ghosts = np.ones(self.domain.dimension) * nbGhosts

            # default topology constructor, with global resolution
            # and domain
            topo = self.domain.getOrCreateTopology(self.domain.dimension,
                                                   self.resolution,
                                                   ghosts=self.ghosts)
        # Now, discretisation of the variables
        for v in self.variables:
            self.discreteFields[v] = v.discretize(topo)

        # Create a discrete operator, according to the chosen method
        # (only finite differences at the time).
        self.discreteOperator = CurlFD(self.discreteFields[self.invar],
                                       self.discreteFields[self.outvar],
                                       method=self.method)
        self.discreteOperator.setUp()
        # ghost op
        self.ghostsOperator = UpdateGhosts(self.input, topo=topo)

        self._isUpToDate = True

    def apply(self, simulation=None):
        # update ghost points, if required
        if self.ghostsOperator is not None:
            self.ghostsOperator.apply()
        # computation ...
        self.discreteOperator.apply(simulation)

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Curl"
    print Curl.__doc__
