# -*- coding: utf-8 -*-
"""
@file diffusion.py

Operator for diffusion problem.

"""
from parmepy.operator.continuous import Operator
from parmepy.f2py import fftw2py
from parmepy.operator.discrete.diffusion_fft import DiffusionFFT
from parmepy.constants import debug


class Diffusion(Operator):
    """
    Diffusion operator
    \f{eqnarray*}
    \omega = Op(\omega)
    \f} with :
    \f{eqnarray*}
    \frac{\partial \omega}{\partial t} &=& \nu\Delta\omega
    \f}
    """

    @debug
    def __init__(self, velocity=None, vorticity=None,
                 resolutions=None, method='',
                 **other_config):
        """
        Constructor.
        Create a Diffusion operator using FFT.

        @param velocity ContinuousVectorField : velocity variable.
        @param vorticity ContinuousVectorField : vorticity variable.
        @param viscosity : viscosity of the considered medium.
        """
        Operator.__init__(self, [velocity, vorticity], method,
                          name="Curl and Diffusion")
        self.velocity = velocity
        self.vorticity = vorticity
        self.resolutions = resolutions
        self.config = other_config

    @debug
    def setUp(self):
        """
        Diffusion operator discretization method.
        Create a discrete Diffusion operator from given specifications.
        """
        localres, localoffset = fftw2py.init_fftw_solver(
            self.resolutions[self.vorticity],
            self.domain.length)

        topodims = self.resolutions[self.vorticity] / localres
        print 'topodims DIFFUSION', topodims
        # variables discretization

        for v in self.variables:
            topo = self.domain.getOrCreateTopology(self.domain.dimension,
                                                   self.resolutions[v],
                                                   topodims)
            vd = v.discretize(topo)
            self.discreteFields[v] = vd

        self.discreteOperator =\
            DiffusionFFT(self.discreteFields[self.velocity],
                         self.discreteFields[self.vorticity],
                         self.method, **self.config)

        self.discreteOperator.setUp()
        self._isUpToDate = True

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Diffusion"
    print Diffusion.__doc__
