# -*- coding: utf-8 -*-
"""
@file diffusion.py

Operator for diffusion problem.

"""
from parmepy.operator.continuous import Operator
from parmepy.f2py import fftw2py
from parmepy.operator.discrete.diffusion_fft import DiffusionFFT
from parmepy.constants import debug
from parmepy.mpi import main_size
import numpy as np


class Diffusion(Operator):
    """
    Diffusion operator
    \f{eqnarray*}
    \omega = Op(\omega)
    \f} with :
    \f{eqnarray*}
    \frac{\partial \omega}{\partial t} &=& \nu\Delta\omega
    \f}
    """

    @debug
    def __init__(self, vorticity, resolution, viscosity, method=None,
                 ghosts=None, **other_config):
        """
        Constructor for the diffusion operator.
        @param[in,out] vorticity : field \f$ \omega \f$
        @param[in] resolution :  \f$ \omega \f$ global resolution.
        @param[in] viscosity : \f$\nu\f$, viscosity of the considered medium.
        """
        ## input/output field, solution of the problem
        self.vorticity = vorticity
        ## Global resolution for vorticity
        self.resolution = resolution
        ## viscosity
        self.viscosity = viscosity
        Operator.__init__(self, [vorticity], method, ghosts=ghosts,
                          name="Diffusion")
        self.config = other_config
        # The only available method at the time is fftw
        if method is None:
            self.method = 'fftw'
        self.input = [self.vorticity]
        self.output = [self.vorticity]

    @debug
    def setUp(self):
        """
        Diffusion operator setup : discretization of fields and operator.

        """
        # The only available solver is fftw
        if self.method is not 'fftw':
            print self.method
            raise AttributeError("Method not yet implemented.")

        # Compute local resolution/distribution of data
        # according to fftw requirements.
        localres, localoffset = fftw2py.init_fftw_solver(
            self.resolution, self.domain.length)

        topodims = np.ones((self.domain.dimension))
        topodims[-1] = main_size
        #variables discretization
        if self.ghosts is not None:
            raise AttributeError("Ghosts points not yet\
            implemented for diffusion operator.")
        for v in self.variables:
            topo = self.domain.getOrCreateTopology(self.domain.dimension,
                                                   self.resolution, topodims,
                                                   precomputed=True,
                                                   offset=localoffset,
                                                   localres=localres,
                                                   ghosts=self.ghosts)

            self.discreteFields[v] = v.discretize(topo)

        self.discreteOperator = DiffusionFFT(
            self.discreteFields[self.vorticity], self.viscosity,
            method=self.method, **self.config)

        self.discreteOperator.setUp()
        self._isUpToDate = True

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Diffusion"
    print Diffusion.__doc__
