# -*- coding: utf-8 -*-
"""
@file diffusion_fft.py
Discrete Diffusion operator using FFTW (fortran)
"""
from parmepy.f2py import fftw2py
from parmepy.operator.discrete.discrete import DiscreteOperator
from parmepy.constants import debug
from parmepy.tools.timers import timed_function


class DiffusionFFT(DiscreteOperator):
    """
    Discretized Poisson operator based on FFTW.
    See details in parmepy.operator.diffusion.

    """
    @debug
    def __init__(self, vorticity, viscosity, method=None):
        """
        Constructor.
        @param[in,out] vorticity :  discretisation of the field \f$ \omega \f$.
        @param[in] viscosity : \f$\nu\f$, viscosity of the considered medium.
        """
        ## Discretisation of the solution field
        self.vorticity = vorticity
        ## Viscosity.
        self.viscosity = viscosity

        if self.vorticity.dimension == 1:
            raise AttributeError("Wrong problem dimension: only 2D \
                and 3D cases are implemented.")
        # Base class initialisation
        DiscreteOperator.__init__(self, [self.vorticity], method,
                                  name="Diffusion FFT")
        self.input = [self.vorticity]
        self.output = [self.vorticity]

    @debug
    @timed_function
    def apply(self, simulation):
        if simulation is None:
            raise ValueError("Missing dt value for diffusion computation.")
        dt = simulation.timeStep
        if (self.vorticity.dimension == 2):
            self.vorticity.data = fftw2py.solve_diffusion_2d(
                self.viscosity * dt, self.vorticity.data)

        elif (self.vorticity.dimension == 3):
            self.vorticity.data[0], self.vorticity.data[1],\
                self.vorticity.data[2] = \
                fftw2py.solve_diffusion_3d(self.viscosity * dt,
                                           self.vorticity.data[0],
                                           self.vorticity.data[1],
                                           self.vorticity.data[2])

        else:
            raise ValueError("invalid problem dimension")

    def finalize(self):
        """
        Clean memory (fftw plans and so on)
        """
        fftw2py.clean_fftw_solver(self.vorticity.dimension)

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Diffusion_d"
    print DiffusionFFT.__doc__
