"""
@file operator/discrete/discrete.py
Abstract interface for discrete operators.
"""
from abc import ABCMeta, abstractmethod
from parmepy.constants import debug
from parmepy.tools.timers import Timer


class DiscreteOperator(object):
    """
    Abstract description of a discretized operator.
    It provides a common interface to all discrete
    operators.

    """

    __metaclass__ = ABCMeta

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    @abstractmethod
    def __init__(self, variables, method=None, name=None):
        """
        Create an empty discrete operator.
        """
        if isinstance(variables, list):
            ## variables
            self.variables = variables
        else:
            self.variables = [variables]
        ## Input variables
        self.input = []
        ## Output variables
        self.output = []
        ## Operator numerical method.
        self.method = method
        ##
        self.numMethod = None
        ## Operator name
        self.name = name
        ## Object to store computational times of lower level functions
        self.timer = Timer(self)
        ## bool to check if the setup function has been called for
        ## this operator
        self._isUpToDate = False

    def setUp(self):
        """
        Update the operator --> after calling this function,
        the operator is ready to use.
        """
        self._isUpToDate = True

    def isUp(self):
        """
        True if ready to be applied (--> setup function has
        been called succesfully)
        """
        return self._isUpToDate

    @debug
    @abstractmethod
    def apply(self, simulation=None):
        """
        Apply this operator to its variables.
        @param simulation : object that describes the simulation
        parameters (time, time step, iteration number ...), see
        parmepy.problem.simulation.Simulation for details.
        """

    @debug
    def finalize(self):
        """
        Cleaning, if required.
        """
        print self.timer

    def __str__(self):
        """Common printings for discrete operators."""
        shortName = str(self.__class__).rpartition('.')[-1][0:-2]
        s = shortName + " discrete operator. \n"
        if self.input is not None:
            s += "Input fields : \n"
            for f in self.input:
                s += str(f) + "\n"
        if self.output is not None:
            s += "Output fields : \n"
            for f in self.output:
                s += str(f) + "\n"
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : DiscreteOperator"
    print DiscreteOperator.__doc__
