# -*- coding: utf-8 -*-
"""
@file operator/discrete/multiphase.py
Discrete MultiPhase Rot Grad P
"""
from discrete import DiscreteOperator
from parmepy.numerics.differentialOperator import DifferentialOperator
from parmepy.numerics.fct2op import Fct2Op
from parmepy.constants import np, debug
from parmepy.tools.timers import timed_function


class Baroclinic_d(DiscreteOperator):
    """

    """
    @debug
    def __init__(self, velocity, vorticity, density, viscosity, method=None):
        """
        Constructor.
        Create the old velocity field to compute the -Dp/rho
        in N.S equation
        @param operator.
        """
        DiscreteOperator.__init__(self, [velocity, vorticity, density, viscosity], method,
                                  name="Baroclinic_d")

        self.velocity = velocity
        self.velocity_old = np.zeros_like(self.velocity.data)
        self.vorticity = vorticity
        self.density = density
        self.viscosity = viscosity
        self.input = [self.velocity, self.vorticity, self.density, self.viscosity]
        self.output = [self.vorticity]

    @debug
    @timed_function
    def apply(self, simulation=None):
        if simulation is None:
            raise ValueError("Missing simulation value for computation.")

        t = simulation.time
        dt = simulation.timeStep
        iCompute = self.velocity.topology.mesh.iCompute

        if t > dt :
            result = np.zeros_like(self.velocity.data)

            # result = du/dt
            for d in xrange(self.velocity.dimension):
                result[d][iCompute] = (self.velocity[d][iCompute] - self.velocity_old[d][iCompute]) / dt

            # result = result + (u. grad)u
            UgradU = DifferentialOperator(self.velocity,
                                          self.velocity,
                                          method = self.method,
                                          choice='divWU').__call__()

            for d in xrange(self.velocity.dimension):
             result[d][iCompute] = result[d][iCompute] + UgradU[d][iCompute]

            ## result = result -nu*\Laplacian u = - gradP/rho
            viscousTerm = DifferentialOperator(self.velocity,
                                               self.velocity,
                                               method = self.method,
                                               choice='laplacianV').__call__()
            for d in xrange(self.velocity.dimension):
                viscousTerm[d][iCompute] = self.viscosity[iCompute] * viscousTerm[d][iCompute]
                result[d][iCompute] = result[d][iCompute] - viscousTerm[d][iCompute]
            
            ## barotrop = -(gradP/rho) x (gradRho/rho)
            gradRho_rho = DifferentialOperator(self.density,
                                               self.density,
                                               method = self.method,
                                               choice='gradS').__call__()
            for d in xrange(self.velocity.dimension):
                gradRho_rho[d][iCompute] = gradRho_rho[d][iCompute] / self.density[iCompute]
            barotrop = np.zeros_like(result)
            barotrop[0][iCompute] = result[1][iCompute] * gradRho_rho[2][iCompute] - \
                                    result[2][iCompute] * gradRho_rho[1][iCompute]
            barotrop[1][iCompute] = result[2][iCompute] * gradRho_rho[0][iCompute] - \
                                    result[0][iCompute] * gradRho_rho[2][iCompute]
            barotrop[2][iCompute] = result[0][iCompute] * gradRho_rho[1][iCompute] - \
                                    result[1][iCompute] * gradRho_rho[0][iCompute]

            ## vorti(n+1) = vorti(n) + dt * barotrop
            for d in xrange(self.vorticity.dimension):
                self.vorticity[d][iCompute] = self.vorticity[d][iCompute] + dt * barotrop[d][iCompute]

        for d in xrange(self.velocity.dimension):
            self.velocity_old[d][iCompute] = self.velocity.data[d][iCompute]


if __name__ == "__main__":
    print __doc__
    print "- Provided class : Baroclinic_d"
    print Baroclinic_d.__doc__

