"""
@file particle_advection.py

Advection solver, particular method, pure-python version.

"""
from parmepy.constants import debug, np, S_DIR
from discrete import DiscreteOperator
from parmepy.fields.continuous import Field
from parmepy.numerics.interpolation import Linear
from parmepy.tools.timers import timed_function


class ParticleAdvection(DiscreteOperator):
    """
    Particular method for advection of a list of fields in a given direction.
    """

    @debug
    def __init__(self, velocity, advectedFields, d,
                 part_position=None, part_advectedFields=None, method=''):
        """
        Constructor.
        @param velocity discrete field
        @param advectedFields : list of discrete fields to be advected
        @param d : Direction to advect
        @param method : Method used
          - 'rk2' : Runge Kutta 2nd order advection
          - 'm4prime' : M4prime formula
          - 'l4star' : Labmda4star formula
          - 'l4star_c3' : Labmda4star C3 formula
          - 'l4star_c4' : Labmda4star C4 formula
          - 'm8prime' : M8prime formula
          - 'l6star' : Lambda6star formula
        """
        ## Advection velocity
        self.velocity = velocity
        variables = [self.velocity]
        if not isinstance(advectedFields, list):
            self.advectedFields = [advectedFields]
        else:
            self.advectedFields = advectedFields
        # update the list of discrete variables for this operator
        [variables.append(advecF) for advecF in self.advectedFields]

        DiscreteOperator.__init__(self, variables, method,
                                  name="Particle Advection" + S_DIR[d])

        self.input = [self.velocity]
        self.output = [df for df in self.variables if not df is self.velocity]
        self.dir = d
        self.part_position = part_position
        self.part_advectedFields = part_advectedFields
        self.num_method = None
        self.available_methods = [
            'l2star_c4', 'l2star_c3', 'l2star_c2', 'm4prime',
            'l4star_c4', 'l4star_c3', 'l4star',
            'l6star_c6', 'l6star_c5', 'l6star_c4', 'l6star', 'm8prime',
            'l8star']

    @debug
    def setUp(self):
        ## Result position
        if self.part_position is None:
            self.part_position = \
                Field(self.advectedFields[0].topology.domain,
                      "Particle_Position",
                      isVector=False
                      ).discretize(self.advectedFields[0].topology)
        ## Result advectedFields
        if self.part_advectedFields is None:
            self.part_advectedFields = [
                Field(adF.topology.domain, "Particle_AdvectedFields",
                      isVector=adF.isVector).discretize(adF.topology)
                for adF in self.advectedFields]

        self.interpolation = \
            Linear(self.velocity, self.dir,
                   self.advectedFields[0].topology.mesh.space_step,
                   self.advectedFields[0].topology.mesh.origin)

        if self.method.find('rk4') >= 0:
            self.name += '_rk4'
            from parmepy.numerics.integrators.runge_kutta4 \
                import RK4_scalar as RKn
        else:
            self.name += '_rk2'
            from parmepy.numerics.integrators.runge_kutta2 \
                import RK2_scalar as RKn
        self.num_advec = RKn(self.interpolation)

        remesh = 'l4star'
        for m in self.available_methods:
            if self.method.find(m) >= 0:
                remesh = m
                break
        self.name += '_' + remesh
        from parmepy.numerics.remeshing import Remeshing
        self.num_remesh = Remeshing(
            self.velocity.dimension,
            self.advectedFields[0].topology.mesh.space_step,
            self.advectedFields[0].topology.mesh.origin,
            remesh)

        # integration work variable (interpolated velocities)
        if self.method.find('rk4') >= 0:
            self.work = [np.empty_like(self.part_position.data[0])
                         for i in xrange(3)]
        else:
            self.work = np.empty_like(self.part_position.data[0])
        # integration work variable (temporary positions)
        self.y_temp = np.empty_like(self.part_position.data[0])
        # interpolation work variable (distance to nearest grid point)
        self.i_y = np.empty_like(self.part_position.data[0])
        # interpolation work variables (indices arrays)
        if len(self.part_position.data[0].shape) == 2:
            self.i_row, self.i_col = \
                np.indices(self.part_position.data[0].shape)
            self.i_dep = None
        elif len(self.part_position.data[0].shape) == 3:
            self.i_row, self.i_col, self.i_dep = \
                np.indices(self.part_position.data[0].shape)
        else:
            self.i_row = np.indices(self.part_position.data[0].shape)[0]
            self.i_col, self.i_dep = None, None

        self._isUpToDate = True

    @debug
    @timed_function
    def apply(self, simulation, dtCoeff, split_id, old_dir=None):
        """
        Apply advection operator.

        @param t : current time.
        @param dt : time step.
        @param d : Direction of splitting.

        Advection algorithm:
        @li 1. Particle initialization : \n
                 - by copy advectedFields from grid to particles
                 if previous splitting direction equals current
                 splitting direction.\n
                 - by transposition of advectedFields from grid to particle.
        @li 2. Particle advection :\n
                 - compute particle position in splitting direction as a
                 advectedFields. Performs a RK2 resolution of dx_p/dt = a_p.
        """
        t, dt = simulation.time, simulation.timeStep * dtCoeff
        # copy for particle advected initialization
        for p_adF, adF in zip(self.part_advectedFields, self.advectedFields):
            for dim in xrange(adF.nbComponents):
                p_adF[dim][...] = adF[dim][...]

        # Get particle initial position (grid points)
        #for p_pos, adF in zip(self.part_position, self.advectedFields):
        self.part_position[0][...] = \
            self.part_position.topology.mesh.coords[self.dir]

        # Advect particles
        velocity = self.velocity.data[self.dir] if self.velocity.isVector \
            else self.velocity.data
        self.num_advec(
            t, dt, self.part_position.data[0], res=self.part_position.data[0],
            work=self.work, y_temp=self.y_temp, yprime=velocity,
            i_row=self.i_row, i_col=self.i_col, i_dep=self.i_dep,
            i_y=self.i_y)

        # Remesh particles
        for adF, p_adF in zip(self.advectedFields,
                              self.part_advectedFields):
            for dim in xrange(adF.nbComponents):
                self.num_remesh(
                    self.part_position.data[0],
                    adF.topology.mesh.origin[self.dir],
                    p_adF.data[dim], self.dir, res=adF[dim],
                    i_row=self.i_row, i_col=self.i_col, i_dep=self.i_dep,
                    i_y=self.i_y, work=self.work)

    @debug
    def finalize(self):
        pass

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Transport_d"
    print ParticleAdvection.__doc__
