# -*- coding: utf-8 -*-
"""
@file poisson_fft.py
Discrete operator for Poisson problem (fftw based)
"""
import numpy as np
from parmepy.f2py import fftw2py
from parmepy.constants import PARMES_REAL, ORDER
from discrete import DiscreteOperator
from parmepy.constants import debug, prof
from parmepy.tools.timers import timed_function


class PoissonFFT(DiscreteOperator):
    """
    Discretized Poisson operator based on FFTW.
    See details in parmepy.operator.poisson
    """

    @debug
    def __init__(self, velocity, vorticity, method=None, projection=False,
                 multires=False, filterSize=None):
        """
        Constructor.
        @param[out] velocity : discretisation of the solution field
        @param[in] vorticity : discretisation of the RHS (mind the minus rhs!)
        """
        ## Solution field
        self.velocity = velocity
        ## RHS field
        self.vorticity = vorticity
        ## Solenoidal projection on vorticity ?
        self.projection = projection
        ## Multiresolution ?
        self.multires = multires
        ## Filter size array = domainLength/(CoarseRes-1)
        self.filterSize = filterSize
        # Base class initialisation
        DiscreteOperator.__init__(self, [velocity, vorticity], method,
                                  name="Poisson FFT")

        # If 2D problem, vorticity must be a scalar
        if self.velocity.dimension == 2:
            assert self.vorticity.dimension == 1
            ## fftw solver case (2 or 3D)
            self.case = '2D'
        elif self.velocity.dimension == 3:
            self.case = '3D'
        else:
            raise AttributeError("Wrong problem dimension: only 2D \
                                        and 3D cases are implemented.")

        self.input = [self.vorticity]
        self.output = [self.velocity]

    @debug
    @prof
    @timed_function
    def apply(self, simulation=None):
        ite = simulation.currentIteration

        if self.case is '2D':
            self.velocity.data[0], self.velocity.data[1] =\
                fftw2py.solve_poisson_2d(self.vorticity.data,
                                         self.velocity.data[0],
                                         self.velocity.data[1])

        elif self.case is '3D':
            #=== SOLENOIDAL PROJECTION OF VORTICITY FIELD ===
            if (self.projection and ite % 1 == 0):
                self.vorticity.data[0], self.vorticity.data[1], \
                    self.vorticity.data[2] = \
                    fftw2py.projection_om_3d(self.vorticity.data[0],
                                             self.vorticity.data[1],
                                             self.vorticity.data[2])

            if (self.multires):
                # Projects vorticity values from fine to coarse grid 
                # in frequencies space by nullifying the smallest modes
                vortFilter = np.copy(self.vorticity.data)
                vortFilter[0], vortFilter[1], \
                    vortFilter[2] = \
                    fftw2py.multires_om_3d(self.filterSize[0],
                                           self.filterSize[1],
                                           self.filterSize[2],
                                           self.vorticity.data[0],
                                           self.vorticity.data[1],
                                           self.vorticity.data[2])

                # Solves Poisson equation using filter vorticity
                self.velocity.data[0], self.velocity.data[1], \
                    self.velocity.data[2] = \
                    fftw2py.solve_poisson_3d(vortFilter[0],
                                             vortFilter[1],
                                             vortFilter[2],
                                             self.velocity.data[0],
                                             self.velocity.data[1],
                                             self.velocity.data[2])

            else :
                # Solves Poisson equation using usual vorticity
                self.velocity.data[0], self.velocity.data[1], \
                    self.velocity.data[2] = \
                    fftw2py.solve_poisson_3d(self.vorticity.data[0],
                                             self.vorticity.data[1],
                                             self.vorticity.data[2],
                                             self.velocity.data[0],
                                             self.velocity.data[1],
                                             self.velocity.data[2])

        else:
            raise ValueError("invalid problem dimension")

    def finalize(self):
        """
        Clean memory (fftw plans and so on)
        """
        fftw2py.clean_fftw_solver(self.velocity.dimension)


if __name__ == "__main__":
    print __doc__
    print "- Provided class : Poisson_d"
    print PoissonFFT.__doc__
