# -*- coding: utf-8 -*-
"""
@file operator/discrete/stretching.py

Discretisation of the stretching operator for two different formulations.
"""

from parmepy.constants import debug
from discrete import DiscreteOperator
from parmepy.numerics.integrators.euler import Euler
from parmepy.numerics.integrators.runge_kutta2 import RK2
from parmepy.numerics.integrators.runge_kutta3 import RK3
from parmepy.numerics.integrators.runge_kutta4 import RK4
from parmepy.tools.timers import timed_function
from parmepy.numerics.differential_operations import DivT, GradVxW
import numpy as np
import math
ceil = math.ceil


class ConservativeStretching(DiscreteOperator):
    """
    Discretisation of the following problem :
    \f{eqnarray*} \frac{\partial\omega}{\partial t} = \nabla.(\omega:v) \f}
    """

    @debug
    def __init__(self, velocity, vorticity, method=None):
        """
        @param velocity : discrete field
        @param vorticity : discrete field
        @param method : numerical method for space/time discretizations
        """
        ## velocity discrete field
        self.velocity = velocity
        ## vorticity discrete field
        self.vorticity = vorticity

        DiscreteOperator.__init__(self, [self.velocity, self.vorticity],
                                  method,
                                  name="ConservativeStretching")

        self.input = [self.velocity, self.vorticity]
        self.output = [self.vorticity]

        # \todo multiresolution case
        assert self.velocity.topology == self.vorticity.topology,\
            'Multiresolution case not yet implemented.'

        # We must have a topo for this operator ...
        # A function to compute the gradient of a vector field
        self.divFunc = DivT(self.velocity.topology,
                            method=method)

        # Time integration scheme.
        def rhs(t, y):
            return self.divFunc(self.velocity.data, y)

        # TOdo : check fd call with numpy and/or parmes fields.
        # stability constant
        self.cststretch = 0.
        # set time integration method
        if self.method.find('Euler') >= 0:
            self.num_method = Euler
            self.cststretch = 2.0
        elif self.method.find('RK2') >= 0:
            self.num_method = RK2
            self.cststretch = 2.0
        elif self.method.find('RK3') >= 0:
            self.num_method = RK3
            self.cststretch = 2.5127
        elif self.method.find('RK4') >= 0:
            self.num_method = RK4
            self.cststretch = 2.7853
        else:
            self.num_method = RK3
            self.cststretch = 2.5127

        self.timeIntegrator = self.num_method(self.vorticity.nbComponents, rhs)

    @timed_function
    def apply(self, simulation):
        if simulation is None:
            raise ValueError("Missing simulation value for computation.")

        # time step
        dt = simulation.timeStep
        # current time
        t = simulation.time
        # Call time integrator
        self.vorticity.data[0][...], self.vorticity.data[1][...], \
            self.vorticity.data[2][...] = \
            self.timeIntegrator(self.vorticity.data, dt, t)


class GradStretching(DiscreteOperator):
    """
    Discretisation of the following problem :
    \f{eqnarray*} \frac{\partial\omega}{\partial t}=[\nabla(v)][\omega]\f}
    """

    @debug
    def __init__(self, velocity, vorticity, method=None):
        """
        @param velocity : discrete field
        @param vorticity : discrete field
        @param method : numerical method for space/time discretizations
        """
        ## velocity discrete field
        self.velocity = velocity
        ## vorticity discrete field
        self.vorticity = vorticity

        DiscreteOperator.__init__(self, [self.velocity, self.vorticity],
                                  method,
                                  name="ConservativeStretching")

        self.input = [self.velocity, self.vorticity]
        self.output = [self.vorticity]

        # \todo multiresolution case
        assert self.velocity.topology == self.vorticity.topology,\
            'Multiresolution case not yet implemented.'

        # We must have a topo for this operator ...
        # A function to compute the gradient of a vector field
        self.graduv = GradVxW(self.velocity.topology, method=method)

        # Time integration scheme.
        def rhs(t, y):
            return self.graduv(self.velocity.data, y)

        # TOdo : check fd call with numpy and/or parmes fields.
        # stability constant
        self.cststretch = 0.
        # set time integration method
        if self.method.find('Euler') >= 0:
            self.num_method = Euler
            self.cststretch = 2.0
        elif self.method.find('RK2') >= 0:
            self.num_method = RK2
            self.cststretch = 2.0
        elif self.method.find('RK3') >= 0:
            self.num_method = RK3
            self.cststretch = 2.5127
        elif self.method.find('RK4') >= 0:
            self.num_method = RK4
            self.cststretch = 2.7853
        else:
            self.num_method = RK3
            self.cststretch = 2.5127

        ## time integrator
        self.timeIntegrator = self.num_method(self.vorticity.nbComponents, rhs)
        ## a vector of float with diagnostics computed
        ## from GradVxW (max div ...)
        self.diagnostics = np.ones((5))

    @timed_function
    def apply(self, simulation):
        if simulation is None:
            raise ValueError("Missing simulation value for computation.")

        # time step
        dt = simulation.timeStep
        # current time
        t = simulation.time
        self.graduv(self.velocity, self.vorticity, self.diagnostics)
        # Compute the number of required subcycles
        ndt, subdt = self.checkStability(dt)
        # Call time integrator
        print "time steps ...", dt, subdt
        assert sum(subdt) == dt
        for i in xrange(ndt):
            self.vorticity.data[0], self.vorticity.data[1],\
                self.vorticity.data[2] =\
                self.timeIntegrator(self.vorticity.data, subdt[i], t)

    def checkStability(self, dt):
        """
        Computes a stability condition depending on some
        diagnostics (from GradVxW)
        @param current time step
        @return int,float : the number of required subcycles and
        the subcycle time-step.
        """
        dt_stab = min(dt, self.cststretch / self.diagnostics[0])
        nb_cycles = int(ceil(dt / dt_stab))
        subdt = np.zeros((nb_cycles))
        subdt[:] = dt_stab
        subdt[-1] = dt - (nb_cycles - 1) * dt_stab
        return nb_cycles, subdt

if __name__ == "__main__":
    print __doc__
    print "- Provided class : ConservativeStretching, GradStretching"
    print ConservativeStretching.__doc__


