# -*- coding: utf-8 -*-
"""
@file energy_enstrophy.py
Compute Energy and Enstrophy
"""
import numpy as np
from parmepy.constants import debug, PARMES_MPI_REAL
from parmepy.operator.monitors.monitoring import Monitoring
from parmepy.mpi import main_rank, main_size, MPI
from parmepy.tools.timers import timed_function


class Energy_enstrophy(Monitoring):
    """
    Computes enstrophy and the kinetic energy
    \f{eqnarray*}
    enstrophy = \frac{1}{\Omega}\int_\Omega \omega^2 d\Omega
    \f} with \f$\Omega\f$ the volume or surface of the physical domain
    \f$\omega\f$ the vorticity and
    \f{eqnarray*}
    energy = \frac{1}{2\Omega}\int_\Omega v^2 d\Omega
    \f}
    """

    def __init__(self, velocity, vorticity,
                 topo, viscosity, frequency, prefix=None):
        """
        Constructor.
        @param fields : List of fields
        @param resolutions :  list of resolutions (one per variable)
        @param viscosity :  kinematic viscosity
        @param topo : the topology on which we want to monitor the fields
        @param frequency : output file producing frequency.
        @param prefix : file name prefix, contains relative path.
        """
        Monitoring.__init__(self, [velocity, vorticity], frequency, topo=topo,
                            name="Energy Enstrophy")
        if prefix is None:
            self.prefix = './Energy'
        else:
            self.prefix = prefix
        ## velocity field
        self.velocity = velocity
        ## vorticity field
        self.vorticity = vorticity
        ## viscosity (scalar)
        self.viscosity = viscosity
        self._buffer_1 = 0.
        self._buffer_2 = 0.
        if (main_rank == 0):
            self.f = open(self.prefix, 'w')
        self.input = [velocity, vorticity]
        self.output = []
        self.topo = self._predefinedTopo
        self.topovel = self.topo
        self.topovort = self.topo
        # \todo : rewrite for multiresolution case

    def setUp(self):
        Monitoring.setUp(self)

        self.discreteFields[self.velocity] = \
            self.velocity.discretize(self.topovel)
        self.discreteFields[self.vorticity] =\
            self.vorticity.discretize(self.topovort)

        self._isUpToDate = True

        spaceStep = [self.topovel.mesh.space_step,
                     self.topovort.mesh.space_step]

        length = self.topo.domain.length
        self.coeffEnergy = 0.5 * (np.prod(spaceStep[0]) /
                                  np.prod(length))
        self.coeffEnstrophy = (np.prod(spaceStep[1]) /
                               np.prod(length))

    @debug
    @timed_function
    def apply(self, simulation=None):
        """
        Computation of kinetic energy, enstrophy &
        Checking energy and enstrophy decay
        """
        if simulation is None:
            raise ValueError("Missing simulation value for computation.")

        t = simulation.time
        dt = simulation.timeStep
        ite = simulation.currentIteration
        # Kinetic energy computation
        vd = self.discreteFields[self.velocity]
        # get the list of computation points (no ghosts)
        ind = self.topo.mesh.iCompute
        nbc = vd.nbComponents
        energy = self.coeffEnergy * np.sum([vd[i][ind] ** 2
                                            for i in xrange(nbc)])

        # Enstrophy computation
        vortd = self.discreteFields[self.vorticity]
        nbc = vortd.nbComponents
        enstrophy = self.coeffEnstrophy * np.sum([vortd[i][ind] ** 2
                                                  for i in xrange(nbc)])
        # Collective communications

        energy = self.topovel.comm.reduce(energy, PARMES_MPI_REAL,
                                          op=MPI.SUM, root=0)
        energyBuff1 = self.topovel.comm.reduce(self._buffer_1, PARMES_MPI_REAL,
                                               op=MPI.SUM, root=0)
        energyBuff2 = self.topovel.comm.reduce(self._buffer_2, PARMES_MPI_REAL,
                                               op=MPI.SUM, root=0)
        enstrophy = self.topovort.comm.reduce(enstrophy, PARMES_MPI_REAL,
                                              op=MPI.SUM, root=0)

        # Update buffers

        self._buffer_2 = self._buffer_1
        self._buffer_1 = energy

        # Effective viscosity computation

        if main_rank == 0:

            # 1st order
    #        effectiveViscosity = - (energy - energyBuff1) / (dt * enstrophy)
    #        timeDerivativeEnergy = - (energy -  energyBuff1) / dt

            # 2nd order
            effectiveViscosity = - (3.0 * energy - 4.0 *
                                    energyBuff1 + 1.0 * energyBuff2) / \
                (2.0 * dt * enstrophy)

            ratio = effectiveViscosity / self.viscosity

            # Energy dissipation computation

            timeDerivativeEnergy = - (3.0 * energy - 4.0 * energyBuff1 + 1.0 *
                                      energyBuff2) / (2.0 * dt)

            nuEnstrophy = self.viscosity * enstrophy

            nu_effEnstrophy = effectiveViscosity * enstrophy

    #        dissipE = main_comm.reduce(nuEnstrophy,
    #  PARMES_MPI_REAL, op=MPI.SUM, root=0)

        if ((main_rank == 0) and (ite % self.freq == 0) and (t > dt)):
            self.f = open(self.prefix, 'a')
            self.f.write("%s   %s \
                %s   %s   %s   %s   %s\n" % (t, energy,
                                             enstrophy,
                                             ratio,
                                             timeDerivativeEnergy,
                                             nuEnstrophy,
                                             nu_effEnstrophy))
            self.f.close()

    def __str__(self):
        s = "Energy_enstrophy. "
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Energy_enstrophy"
    print Energy_enstrophy.__doc__
