"""
@file printer.py

Classes for handling ouputs.
"""
from parmepy.constants import np, S_DIR, PARMES_REAL, debug
from parmepy.operator.monitors.monitoring import Monitoring
import evtk.hl as evtk
from parmepy.mpi import main_rank
from parmepy.tools.timers import timed_function


class Printer(Monitoring):
    """
    Basic printer.

    Performs outputs in VTK images.
    """

    def __init__(self, frequency=0, fields=[], prefix=None, ext=None):
        """
        Create a results printer for given fields, filename
        prefix (relative path) and an output frequency.

        @param fields : list of variables to export.
        @param frequency : output rate (output every freq iteration)
        @param prefix : output file name prefix.
        @param ext : output file extension, default=.vtk.
        """
        Monitoring.__init__(self, fields, frequency, name="Printer")
        ## output file name prefix
        if prefix is None:
            self.prefix = './out_'
        else:
            self.prefix = prefix
        ## Extension for filename
        if ext is None:
            self.ext = '.vtk'
        else:
            self.ext = ext
        self.input = fields
        self.output = []
        ## Method to collect data in case of distributed data
        self.get_data_method = None
        if self.freq > 0:
            ## Printer core method
            self.step = self._printStep
        else:
            self.step = self._passStep
        ## Internal counter
        self._call_number = 0

    def setUp(self):
        """
        Print initial state
        """
        self.step(0)

    @debug
    @timed_function
    def apply(self, simulation=None):
        if simulation is None:
            raise ValueError("Missing simulation value for monitoring.")

        self._call_number += 1
        if self.freq > 0 and (simulation.currentIteration % self.freq) == 0:
            self.step(simulation.currentIteration)

    def finalize(self):
        if self.freq == -1:
            self._printStep(self._call_number)
        Monitoring.finalize(self)

    def _build_vtk_dict(self):
        """Build a dictionary from fields to VTK image."""
        res = {}

        for f in self.variables:
            for df in f.discreteFields.values():
                #ind = df.topology.mesh.iCompute
                for d in xrange(df.nbComponents):
                    if f.isVector:
                        name = df.name + S_DIR[d]
                    else:
                        name = df.name
                    if len(df.data[d].shape) == 2:
                        res[name] = np.expand_dims(
                            df.data[d], axis=2).astype(PARMES_REAL)
                    else:
                        res[name] = \
                            df.data[d].astype(PARMES_REAL)
        return res

    def set_get_data_method(self, method):
        """Set data collect method."""
        if callable(method):
            self.get_data_method = method
        else:
            raise ValueError("Cannot set non callable method to get data " +
                             "to print. Given method : " + str(method))

    def _passStep(self, ite=None):
        """A code method that do nothing."""
        pass

    def _printStep(self, ite):
        """
        Try to write data into VTK files.
        If fails, turn to classical ascii output.
        @param iter : current iteration number
        """
        # Transfer from GPU to CPU if required
        for f in self.variables:
            for df in f.discreteFields.values():
                try:
                    df.toHost()
                except AttributeError:
                    pass
        # Set output file name
        filename = self.prefix + str(main_rank)
        filename += '_' + "iter_{0:03d}".format(ite) + self.ext

        ## VTK output \todo: Need fix in 2D, getting an IOError.
        if self.ext == '.vtk':
            topo = self.variables[0].discreteFields.keys()[0]
            orig = tuple([topo.mesh.origin[i]
                          for i in xrange(topo.mesh.dim)])
            coords = topo.mesh.coords
            ind = topo.mesh.local_start
            orig = tuple([topo.mesh.coords[i].flatten()[ind[i]]
                          for i in xrange(topo.mesh.dim)])
            spacing = tuple(topo.mesh.space_step)
            evtk.imageToVTK(filename, origin=orig, spacing=spacing,
                            pointData=self._build_vtk_dict())
        elif self.ext == '.dat':
            f = open(filename, 'w')
            if self.variables[0].domain.dimension == 2:
                ## Standard output
                df = self.variables[0].discreteFields.values()[0]
                shape = df.topology.mesh.resolution
                coords = df.topology.mesh.coords
                if len(shape) == 2:
                    for i in xrange(shape[0] - 1):
                        for j in xrange(shape[1] - 1):
                            f.write("{0:8.12} {1:8.12} ".format(
                                    coords[0][i, 0], coords[1][0, j]))
                            for field in self.variables:
                                df = field.discreteFields.values()[0]
                                if field.isVector:
                                    f.write("{0:8.12} {1:8.12} ".format(
                                            df[0][i, j],
                                            df[1][i, j]))
                                else:
                                    f.write("{0:8.12} ".format(
                                            df[i, j]))
                            f.write("\n")
            elif self.variables[0].domain.dimension == 3:
                df = self.variables[0].discreteFields.values()[0]
                shape = df.topology.mesh.resolution
                coords = df.topology.mesh.coords
                for i in xrange(shape[0] - 1):
                    for j in xrange(shape[1] - 1):
                        for k in xrange(shape[2] - 1):
                            f.write(
                                "{0:8.12} {1:8.12} {2:8.12} ".format(
                                    coords[0][i, 0, 0],
                                    coords[1][0, j, 0],
                                    coords[2][0, 0, k]))
                            for field in self.variables:
                                df = field.discreteFields.values()[0]
                                if field.isVector:
                                    f.write(
                                        "{0:8.12} {1:8.12} " +
                                        "{2:8.12} ".format(
                                            df[0][i, j, k],
                                            df[1][i, j, k],
                                            df[2][i, j, k]))
                                else:
                                    f.write("{0:8.12} ".format(
                                            df[i, j, k]))
                            f.write("\n")
            else:
                df = self.variables[0].discreteFields.values()[0]
                shape = df.topology.mesh.resolution
                coords = df.topology.mesh.coords
                for i in xrange(shape[0] - 1):
                    f.write("{0:8.12} ".format(coords[0][i]))
                    for field in self.variables:
                        df = field.discreteFields.values()[0]
                        if field.isVector:
                            f.write(
                                "{0:8.12} ".format(df[0][i]))
                        else:
                            f.write("{0:8.12} ".format(df[i]))
                            f.write("\n")
            f.close()

if __name__ == "__main__":
    print __doc__
    print "- Provided class : " + Printer
    print Printer.__doc__
