# -*- coding: utf-8 -*-
"""
@file poisson.py
Poisson problem.

"""
from parmepy.operator.continuous import Operator
from parmepy.f2py import fftw2py
from parmepy.operator.discrete.poisson_fft import PoissonFFT
from parmepy.constants import debug
import numpy as np
from parmepy.mpi.main_var import main_size


class Poisson(Operator):
    """
    Poisson problem,
    \f{eqnarray*}
    v = Op(\omega)
    \f} with :
    \f{eqnarray*}
    \Delta \phi &=& -\omega \\
    v &=& \nabla \times \phi
    \f}
    """

    @debug
    def __init__(self, velocity, vorticity, resolutions, ghosts=None,
                 method=None,
                 **other_config):
        """
        Constructor for the Poisson problem.

        @param[out] velocity : solution field
        @param[in] vorticity : rhs field
        @param[in] resolutions :  list of resolutions (one per variable)
        """
        ## solution of the problem
        self.velocity = velocity
        ## -(right-hand side)
        self.vorticity = vorticity
        ## resolutions list. Warning : for fftw all variables must have
        ## the same resolution.
        self.resolutions = resolutions

        Operator.__init__(self, [velocity, vorticity], method, ghosts=ghosts,
                          name="Poisson")
        self.config = other_config
        if method is None:
            self.method = 'fftw'
        if self.method is 'fftw':
            self.resolution = self.resolutions[self.velocity]
            assert self.resolution == self.resolutions[self.vorticity],\
                'Poisson error : for fftw, all variables must have\
                the same global resolution.'
        self.input = [self.vorticity]
        self.output = [self.velocity]

    @debug
    def setUp(self):
        """
        """

        # Poisson solver based on fftw
        if self.method is not 'fftw':
            print self.method
            raise AttributeError("Method not yet implemented.")
        if self.ghosts is not None:
            raise AttributeError("Ghosts points not yet\
            implemented for poisson operator.")

        # Compute local resolution/distribution of data
        # according to fftw requirements.
        localres, localoffset = fftw2py.init_fftw_solver(
            self.resolution, self.domain.length)

        topodims = np.ones((self.domain.dimension))
        topodims[-1] = main_size
        #variables discretization
        for v in self.variables:
            topo = self.domain.getOrCreateTopology(self.domain.dimension,
                                                   self.resolution, topodims,
                                                   precomputed=True,
                                                   offset=localoffset,
                                                   localres=localres,
                                                   ghosts=self.ghosts)

            self.discreteFields[v] = v.discretize(topo)

        # Build and setup for the discrete operator
        self.discreteOperator = PoissonFFT(self.discreteFields[self.velocity],
                                           self.discreteFields[self.vorticity],
                                           method=self.method,
                                           **self.config)
        self.discreteOperator.setUp()
        self._isUpToDate = True

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Poisson"
    print Poisson.__doc__
