"""
@file redistribute.py
Setup for data transfer/redistribution between two parmes topologies.

This operator is an inter operator which is supposed to define the process to
transfer variables and data from one operator to another.
This mainly concerns data redistribution if the two operators work on
different mpi topologies.

When is it required to define an operator between op1 and op2?
If:
- the intersection between op1.output-variables and op2.input-variables
  is not empty
- AND if the topology on which the variables of op1 are defined is different
from the one of op2 variables.

Note Franck: this kind of operator may also be useful
to define the interpolation/filter process for data transfer for
a variable defined on several meshes.

"""

from parmepy.constants import debug, PARMES_MPI_REAL, ORDERMPI
from parmepy.operator.continuous import Operator
from parmepy.mpi.topology import Bridge
from parmepy.mpi import main_rank, main_comm


class Redistribute(Operator):
    """
    Interconnection between two operators.
    SetUp will compute (or get if it already exists) a Bridge between two
    topologies.
    Apply redistributes data from opFrom topology to opTo topology.

    """
    @debug
    def __init__(self, variables, opFrom, opTo, method=None):

        """
        Create an operator to distribute data between two mpi topologies for a
        list of variables belonging to two operators.

        @param variables : the set of variables to be redistributed
        @param opFrom : source operator
        @param opTo : target (i.e.) the operator that handles the topology on
        which data must be redistributed.
        """

        Operator.__init__(self, variables, method, name="Redistribute")

        ## Source Operator
        self.opFrom = opFrom
        ## Targeted operator
        self.opTo = opTo

        # Then check if variables belong to both operators
        for v in self.variables:
            assert v in opFrom.variables and v in opTo.variables, \
                'Redistribute error : one of the variable is not present\
                in both source and target operator.'
        self.input = self.output = self.variables

    @debug
    def setUp(self):
        """
        Computes intersection of two topologies.

        """
        assert self.opFrom.isUp and self.opTo.isUp(), \
            'You should setup both opFrom and opTo operators\
            before any attempt to setup a redistribute operator.'

        self.bridges = {}
        for v in self.variables:
            vd_from = self.opFrom.discreteFields[v]
            vd_to = self.opTo.discreteFields[v]
            self.bridges[v] = Bridge(vd_from.topology, vd_to.topology)

        self._isUpToDate = True

    def apply(self, simulation=None):
        """
        Proceed with data redistribution from opFrom to opTo
        """
        # TODO :
        # - save a set of bridges in the domain and access them from operator
        # - process all variables in one shot if they have the same topo
        # (use buffers for mpi send/recv? )
        # - move MPI datatypes into the bridge? --> and free MPI type properly
        dim = self.domain.dimension
        for v in self.variables:
            br = self.bridges[v]
            self.r_request = {}
            self.s_request = {}
            # Apply for each component of the data
            for d in range(v.nbComponents):
                vTo = self.opTo.discreteFields[v].data[d]
                vFrom = self.opFrom.discreteFields[v].data[d]
                if br.hasLocalInter:
                    vTo[br.ito] = vFrom[br.ifrom]

                for rk in br.recvFrom.keys():
                    ind = rk * (d + 1)
                    subvshape = tuple([br.recvFrom[rk][i].stop -
                                       br.recvFrom[rk][i].start
                                       for i in range(dim)])
                    substart = tuple([br.recvFrom[rk][i].start
                                      for i in range(dim)])
                    recvTYPE = PARMES_MPI_REAL.Create_subarray(vTo.shape,
                                                               subvshape,
                                                               substart,
                                                               order=
                                                               ORDERMPI)
                    recvTYPE.Commit()
                    self.r_request[ind] = main_comm.Irecv([vTo, 1,
                                                           recvTYPE],
                                                          source=rk,
                                                          tag=rk)
                for rk in br.sendTo.keys():
                    ind = rk * (d + 1)
                    subvshape = tuple([br.sendTo[rk][i].stop -
                                       br.sendTo[rk][i].start
                                       for i in range(dim)])
                    substart = tuple([br.sendTo[rk][i].start
                                      for i in range(dim)])
                    sendTYPE = PARMES_MPI_REAL.Create_subarray(vFrom.shape,
                                                               subvshape,
                                                               substart,
                                                               order=
                                                               ORDERMPI)
                    sendTYPE.Commit()
                    self.s_request[ind] = main_comm.Issend([vFrom, 1,
                                                            sendTYPE],
                                                           dest=rk,
                                                           tag=main_rank)

    def wait(self):
        """
        MPI Barrier to wait for the end
        of all communication requests.
        """
        for rk in self.r_request.keys():
            #            print main_rank, "wait for (rec)...", rk
            self.r_request[rk].Wait()
        for rk in self.s_request.keys():
            #print main_rank, "wait for (send)...", rk
            self.s_request[rk].Wait()

    def test(self, rsend=None, rrecv=None):
        """
        if neither rsend or rrecv is given return
        True if all communication request are complete
        else check for sending to rsend or
        receiving from rrecv. Process ranks
        should be given in main_comm.
        @param rsend : rank of the process
        to which a message has been sent
        and for which we want to test
        message completion.
        @param  rrecv : rank of the process
        from which a message has been receive
        and for which we want to test
        message completion.
        """
        if(rsend is not None or rrecv is not None):
            send_res = True
            recv_res = True
            if rsend is not None:
                send_res = self.s_request[rsend].Test()
            if rrecv is not None:
                recv_res = self.r_request[rrecv].Test()
            res = send_res and recv_res
        else:
            res = True
            for rk in self.r_request.keys():
                res = self.r_request[rk].Test()
                if not res:
                    return res
            for rk in self.s_request.keys():
                res = self.s_request[rk].Test()
                if not res:
                    return res
        return res

    def __str__(self):
        """ToString method"""
        s = "Bridge from :\n " + str(self.opFrom)
        s += "\n to operator:\n " + str(self.opTo)
        s += "for variable(s) " + str([v.name for v in self.variables])
        return s + "\n"

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Redistribute"
    print Redistribute.__doc__
