# -*- coding: utf-8 -*-
"""
@file reprojection_criterion.py
Compute reprojection criterion and divergence maximum
"""
import numpy as np
from numpy import linalg as la
from parmepy.constants import np, debug, PARMES_REAL, PARMES_MPI_REAL
from parmepy.operator.monitors.monitoring import Monitoring
from parmepy.numerics.differentialOperator import DifferentialOperator
from parmepy.mpi import main_rank, main_comm, main_size, MPI
from parmepy.tools.timers import timed_function


class Reprojection_criterion(Monitoring):
    """
    Computes and prints reprojection criterion and divergence maximum
    """

    def __init__(self, velocity, vorticity,
                 resolutions=None,
                 method=None,
                 frequency=None, prefix=None):
        """
        Constructor.
        @param fields : List of fields
        @param resolutions :  list of resolutions (one per variable)
        @param frequency : output file producing frequency.
        @param prefix : file name prefix, contains relative path.
        """
        Monitoring.__init__(self, [velocity, vorticity], frequency,
                            name="Reprojection")
        self.prefix = prefix
        self.velocity = velocity
        self.vorticity = vorticity
        self.resolutions = resolutions
        self.method = method
        if (main_rank == 0):
            self.f = open(self.prefix, 'w')
        self.input = [velocity, vorticity]
        self.output = []

    def setUp(self):
        Monitoring.setUp(self)
        discreteFields = {}
        topodims = np.ones((self.domain.dimension))
        topodims[-1] = main_size
        nbGhosts = 0
        if self.method.find('FD_order2') >= 0:
            nbGhosts = 1
        elif self.method.find('FD_C4') >= 0:
            nbGhosts = 2
        else:
            nbGhosts = 2

        # Variables discretization
        ghosts = np.zeros((self.domain.dimension))
        ghosts[:] = nbGhosts
        for v in self.variables:
            # the topology for v ...
            topo = self.domain.getOrCreateTopology(self.domain.dimension,
                                                   self.resolutions[v],
                                                   ghosts=ghosts)
            # ... and the corresponding discrete field
            discreteFields[v] = v.discretize(topo)
        self.velo = discreteFields[self.velocity]
        self.vort = discreteFields[self.vorticity]

        self._isUpToDate = True

    @debug
    @timed_function
    def apply(self, simulation=None):
        """
        Computes and prints reprojection criterion and divergence maximum
        """
        t = simulation.time
        dt = simulation.timeStep
        ite = simulation.currentIteration
        gradVort, maxArrayVort, divVort = \
            DifferentialOperator(self.vort, self.vort, self.method,
                                 choice='gradV').__call__()

        gradVelo, maxArrayVelo, divVelo = \
            DifferentialOperator(self.velo, self.velo, self.method,
                                 choice='gradV').__call__()

        maxDiv = np.max(abs(divVort))
        maxdwx_dx = np.max(abs(gradVort[0]))
        maxdwx_dy = np.max(abs(gradVort[1]))
        maxdwx_dz = np.max(abs(gradVort[2]))
        maxdwy_dx = np.max(abs(gradVort[3]))
        maxdwy_dy = np.max(abs(gradVort[4]))
        maxdwy_dz = np.max(abs(gradVort[5]))
        maxdwz_dx = np.max(abs(gradVort[6]))
        maxdwz_dy = np.max(abs(gradVort[7]))
        maxdwz_dz = np.max(abs(gradVort[8]))
        maxmax = max(maxdwx_dx, maxdwx_dy, maxdwx_dz,
                     maxdwy_dx, maxdwy_dy, maxdwy_dz,
                     maxdwz_dx, maxdwz_dy, maxdwz_dz)
        projCriterion = maxDiv / maxmax
        dtstabAdv = 0.5 / maxArrayVelo[1]
        dtstabStretch = 2.5127 / maxArrayVelo[0]

        if ((main_rank == 0) and (ite % self.freq == 0) and (t > dt)):
            self.f = open(self.prefix, 'a')
            self.f.write("%s   %s   %s   %s   %s   %s   %s   %s   %s   %s   %s   %s   %s   %s\n" % (t, projCriterion,
                                                                                                    maxDiv,
                                                                                                    maxdwx_dx,
                                                                                                    maxdwx_dy,
                                                                                                    maxdwx_dz,
                                                                                                    maxdwy_dx,
                                                                                                    maxdwy_dy,
                                                                                                    maxdwy_dz,
                                                                                                    maxdwz_dx,
                                                                                                    maxdwz_dy,
                                                                                                    maxdwz_dz,
                                                                                                    dtstabAdv,
                                                                                                    dtstabStretch))
            self.f.close()

    def __str__(self):
        s = "Reprojection_criterion. "
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Reprojection_criterion"
    print Reprojection_criterion.__doc__
