# -*- coding: utf-8 -*-
"""
@file operator/stretching.py

Computation of stretch. term in Navier-Stokes.

"""
from parmepy.constants import debug, GRADUW, CONSERVATIVE
from parmepy.operator.continuous import Operator
from parmepy.operator.discrete.stretching import ConservativeStretching
from parmepy.operator.discrete.stretching import GradStretching
from parmepy.operator.updateGhosts import UpdateGhosts
import numpy as np


class Stretching(Operator):
    """
    \todo write latex formulas
    """

    @debug
    def __init__(self, velocity, vorticity, resolutions,
                 formulation=None, method=None, topo=None, ghosts=None,
                 ghostUpdate=None):
        """
        Create a Stretching operator from given
        velocity and vorticity variables.
        'FD_C4 RK3',

        @param velocity field
        @param vorticity field
        @param resolutions : grid resolution of velocity and vorticity
        @param method : solving method
        (default = finite differences, 4th order, in space
        and Runge-Kutta 3 in time.)
        @param topo : a predefined topology to discretize velocity/vorticity
        @param ghosts : number of ghosts points. Default depends on the method.
        Autom. computed if not set.
        """
        Operator.__init__(self, [velocity, vorticity], method, topo=topo,
                          ghosts=ghosts, name="Stretching")
        ## velocity variable (vector)
        self.velocity = velocity
        ## vorticity variable (vector)
        self.vorticity = vorticity
        ## Grid resolution for each variable (dictionnary)
        self.resolutions = resolutions
        ## Numerical methods for time and space discretization
        if method is None:
            self.method = 'FD_C4_RK3'
        else:
            self.method = method
        ## Formulation used for the stretching equation.
        ## Default = conservative form.
        if formulation is not None:
            self.formulation = formulation
        else:
            self.formulation = CONSERVATIVE
        ## Do this operator deals with ghost points update?
        ## default = yes
        if ghostUpdate is None:
            self.ghostUpdate = True
        else:
            self.ghostUpdate = ghostUpdate
        self.input = [self.velocity, self.vorticity]
        self.output = [self.vorticity]
        ## operator to update ghost points of the input variables
        self.synchronize = None

    @debug
    def setUp(self):
        print '================ Stretching scheme ====================='
        print ' Space discretization and integr. scheme:', self.method
        print '========================================================'
        nbGhosts = 0
        if self.method.find('FD_order2') >= 0:
            nbGhosts = 1
        elif self.method.find('FD_C4') >= 0:
            nbGhosts = 2
        else:
            raise ValueError("Unknown method for stretching operator.")

        if self._predefinedTopo is not None:
            assert (self._predefinedTopo.ghosts >= nbGhosts).all()
            topo = self._predefinedTopo
            for v in self.variables:
                self.discreteFields[v] = v.discretize(topo)

        else:
            # Variables discretization
            if self.ghosts is not None:
                self.ghosts[self.ghosts < nbGhosts] = nbGhosts
            else:
                self.ghosts = np.ones((self.domain.dimension)) * nbGhosts

            for v in self.variables:
                topo = self.domain.getOrCreateTopology(self.domain.dimension,
                                                       self.resolutions[v],
                                                       ghosts=self.ghosts)
                self.discreteFields[v] = v.discretize(topo)

        if self.formulation is CONSERVATIVE:
            self.discreteOperator =\
                ConservativeStretching(self.discreteFields[self.velocity],
                                       self.discreteFields[self.vorticity],
                                       method=self.method)
            if self.ghostUpdate:
                topov = self.discreteFields[self.velocity].topology
                self.synchronize = UpdateGhosts(self.input, topo=topov)
                self.synchronize.setUp()
        elif self.formulation is GRADUW:
            self.discreteOperator =\
                GradStretching(self.discreteFields[self.velocity],
                               self.discreteFields[self.vorticity],
                               method=self.method)
            if self.ghostUpdate:
                topov = self.discreteFields[self.velocity].topology
                self.synchronize = UpdateGhosts(self.velocity, topo=topov)
                self.synchronize.setUp()
        self.discreteOperator.setUp()
        self._isUpToDate = True

    def apply(self, simulation=None):
        # update ghost points, if required
        if self.synchronize is not None:
            self.synchronize.apply()
        # computation ...
        self.discreteOperator.apply(simulation)


if __name__ == "__main__":
    print __doc__
    print "- Provided class : Stretching"
    print Stretching.__doc__
