"""
@file operator/synchronizeGhosts.py

Update ghost points for some fields defined on a specific topology.
"""
from parmepy.operator.continuous import Operator
from parmepy.operator.discrete.synchronizeGhosts import SynchronizeGhosts_d
from parmepy.constants import debug
from parmepy.fields.vector import VectorField
from parmepy.fields.scalar import ScalarField

class SynchronizeGhosts(Operator):
    """
    Ghost points synchronization.
    """

    @debug
    def __init__(self, fieldslist, resolution=None,
                 method='', topology=None, **other_config):
        """
        Defines a way to send/recv values at ghosts points for a list
        of fields, discretized on a given topology.

        @param  fieldslist : a list of fields
        """
        if isinstance(fieldslist, list):
            Operator.__init__(self, fieldslist, method,
                              name="Ghost Synchronization")
        else:
            Operator.__init__(self, [fieldslist], method,
                              name="Ghost Synchronization")
        self.fieldslist = fieldslist
        self.resolution = resolution
        self.method = method
        self.topology = topology
        self.config = other_config

    @debug
    def setUp(self, ghosts=None):
        """
        SynchroGhost operator discretization method.
        Create an discrete SynchroGhost operator from given specifications.

        """
        Operator.setUp(self)
        for v in self.variables:
        # the topology for v ...
            if self.topology is not None:
                topo = self.topology
            else:
                if ghosts is None:
                    topo = self.domain.getOrCreateTopology(
                        self.domain.dimension,
                        self.resolution[v])
                else:
                    topo = self.domain.getOrCreateTopology(
                        self.domain.dimension,
                        self.resolution[v],
                        ghosts=ghosts)
                topo.ghosts = ghosts
            # ... and the corresponding discrete field
                if v.__class__ is VectorField or v.__class__ is ScalarField:
                    self.discreteOperator = \
                        SynchronizeGhosts_d(v, topo, self.method)
                else:
                    self.discreteFields[v] = v.discretize(topo)
                    self.discreteOperator = \
                        SynchronizeGhosts_d(self.discreteFields[v],
                        topo, self.method)

        self.discreteOperator.setUp()
        self._isUpToDate = True


if (__name__ == "__main__"):
    print __doc__
    print "- Provided class : SynchronizeGhosts"
    print SynchronizeGhosts.__doc__
