"""
Testing Scales advection operator.
"""
import numpy as np
from parmepy.constants import PARMES_REAL, ORDER
from parmepy.domain.box import Box
from parmepy.fields.continuous import Field
from parmepy.operator.advection import Advection
from parmepy.problem.simulation import Simulation


def test_nullVelocity_m4():
    """Basic test with random velocity. Using M4prime"""
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar')
    scal_ref = Field(domain=box, name='Scalar_ref')
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y, z: (0., 0., 0.), isVector=True)
    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='scales_p_M4'
                      )
    advec_py = Advection(velo, scal_ref,
                         resolutions={velo: [17, 17, 17],
                                      scal_ref: [17, 17, 17]},
                         method='m4prime',
                         splittingConfig='o2_FullHalf'
                         )
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    scal_ref_d = scal_ref.discreteFields.values()[0]

    scal_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_ref_d.data[0][...] = scal_d.data[0][...]

    advec.apply(Simulation(0., 0.1, 1))
    advec_py.apply(Simulation(0., 0.1, 1))

    assert np.allclose(scal_ref_d.data[0], scal_d.data[0])


def test_nullVelocity_vec_m4():
    """Basic test with random velocity and vector field. Using M4prime"""
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar', isVector=True)
    scal_ref = Field(domain=box, name='Scalar_ref', isVector=True)
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y, z: (0., 0., 0.), isVector=True)
    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='scales_p_M4'
                      )
    advec_py = Advection(velo, scal_ref,
                         resolutions={velo: [17, 17, 17],
                                      scal_ref: [17, 17, 17]},
                         method='m4prime',
                         splittingConfig='o2_FullHalf'
                         )
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    scal_ref_d = scal_ref.discreteFields.values()[0]

    scal_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_d.data[1][...] = np.asarray(
        np.random.random(scal_d.data[1].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_d.data[2][...] = np.asarray(
        np.random.random(scal_d.data[2].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_ref_d.data[0][...] = scal_d.data[0][...]
    scal_ref_d.data[1][...] = scal_d.data[1][...]
    scal_ref_d.data[2][...] = scal_d.data[2][...]

    advec.apply(Simulation(0., 0.1, 1))
    advec_py.apply(Simulation(0., 0.1, 1))

    assert np.allclose(scal_ref_d.data[0], scal_d.data[0])
    assert np.allclose(scal_ref_d.data[1], scal_d.data[1])
    assert np.allclose(scal_ref_d.data[2], scal_d.data[2])


def test_nullVelocity_m6():
    """Basic test with null velocity. Using M6prime"""
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar')
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y, z: (0., 0., 0.), isVector=True)
    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='scales_p_M6'
                      )
    advec.setUp()

    scal_d = scal.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    scal_init = scal_d.data[0].copy()

    advec.apply(Simulation(0., 0.1, 1))

    assert np.allclose(scal_init, scal_d.data[0])


def test_nullVelocity_vec_m6():
    """Basic test with null velocity and vector field. Using M6prime"""
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar', isVector=True)
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y, z: (0., 0., 0.), isVector=True)
    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='scales_p_M6'
                      )
    advec.setUp()

    scal_d = scal.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_d.data[1][...] = np.asarray(
        np.random.random(scal_d.data[1].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_d.data[2][...] = np.asarray(
        np.random.random(scal_d.data[2].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_init0 = scal_d.data[0].copy()
    scal_init1 = scal_d.data[1].copy()
    scal_init2 = scal_d.data[2].copy()

    advec.apply(Simulation(0., 0.1, 1))

    assert np.allclose(scal_init0, scal_d.data[0])
    assert np.allclose(scal_init1, scal_d.data[1])
    assert np.allclose(scal_init2, scal_d.data[2])


def test_nullVelocity_m8():
    """Basic test with random velocity. Using M4prime"""
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar')
    scal_ref = Field(domain=box, name='Scalar_ref')
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y, z: (0., 0., 0.), isVector=True)
    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='scales_p_M8'
                      )
    advec_py = Advection(velo, scal_ref,
                         resolutions={velo: [17, 17, 17],
                                      scal_ref: [17, 17, 17]},
                         method='m8prime',
                         splittingConfig='o2_FullHalf'
                         )
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    scal_ref_d = scal_ref.discreteFields.values()[0]

    scal_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_ref_d.data[0][...] = scal_d.data[0][...]

    advec.apply(Simulation(0., 0.1, 1))
    advec_py.apply(Simulation(0., 0.1, 1))

    assert np.allclose(scal_ref_d.data[0], scal_d.data[0])


def test_nullVelocity_vec_m8():
    """Basic test with random velocity and vector field. Using M4prime"""
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar', isVector=True)
    scal_ref = Field(domain=box, name='Scalar_ref', isVector=True)
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y, z: (0., 0., 0.), isVector=True)
    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='scales_p_M8'
                      )
    advec_py = Advection(velo, scal_ref,
                         resolutions={velo: [17, 17, 17],
                                      scal_ref: [17, 17, 17]},
                         method='m8prime',
                         splittingConfig='o2_FullHalf'
                         )
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    scal_ref_d = scal_ref.discreteFields.values()[0]

    scal_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_d.data[1][...] = np.asarray(
        np.random.random(scal_d.data[1].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_d.data[2][...] = np.asarray(
        np.random.random(scal_d.data[2].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_ref_d.data[0][...] = scal_d.data[0][...]
    scal_ref_d.data[1][...] = scal_d.data[1][...]
    scal_ref_d.data[2][...] = scal_d.data[2][...]

    advec.apply(Simulation(0., 0.075, 1))
    advec_py.apply(Simulation(0., 0.075, 1))

    assert np.allclose(scal_ref_d.data[0], scal_d.data[0])
    assert np.allclose(scal_ref_d.data[1], scal_d.data[1])
    assert np.allclose(scal_ref_d.data[2], scal_d.data[2])


def _randomVelocity_m4():
    """Basic test with random velocity. Using M4prime"""
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar')
    scal_ref = Field(domain=box, name='Scalar_ref')
    velo = Field(domain=box, name='Velocity', isVector=True)
    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='scales_p_M4'
                      )
    advec_py = Advection(velo, scal_ref,
                         resolutions={velo: [17, 17, 17],
                                      scal_ref: [17, 17, 17]},
                         method='m4prime',
                         splittingConfig='o2_FullHalf'
                         )
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    scal_ref_d = scal_ref.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]

    scal_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_ref_d.data[0][...] = scal_d.data[0][...]
    velo_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[0])
    velo_d.data[1][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[1])
    velo_d.data[2][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[1])

    advec.apply(Simulation(0., 0.1, 1))
    advec_py.apply(Simulation(0., 0.1, 1))

    assert np.allclose(scal_ref_d.data[0], scal_d.data[0])


def _randomVelocity_vec_m4():
    """Basic test with random velocity vector Field. Using M4prime"""
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar', isVector=True)
    scal_ref = Field(domain=box, name='Scalar_ref', isVector=True)
    velo = Field(domain=box, name='Velocity', isVector=True)
    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='scales_p_M4'
                      )
    advec_py = Advection(velo, scal_ref,
                         resolutions={velo: [17, 17, 17],
                                      scal_ref: [17, 17, 17]},
                         method='m4prime',
                         splittingConfig='o2_FullHalf'
                         )
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    scal_ref_d = scal_ref.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]

    scal_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_d.data[1][...] = np.asarray(
        np.random.random(scal_d.data[1].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_d.data[2][...] = np.asarray(
        np.random.random(scal_d.data[2].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_ref_d.data[0][...] = scal_d.data[0][...]
    scal_ref_d.data[1][...] = scal_d.data[1][...]
    scal_ref_d.data[2][...] = scal_d.data[2][...]
    velo_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[0])
    velo_d.data[1][...] = np.asarray(
        np.random.random(scal_d.data[1].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[1])
    velo_d.data[2][...] = np.asarray(
        np.random.random(scal_d.data[2].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[1])

    advec.apply(Simulation(0., 0.1, 1))
    advec_py.apply(Simulation(0., 0.1, 1))

    assert np.allclose(scal_ref_d.data[0], scal_d.data[0])
    assert np.allclose(scal_ref_d.data[1], scal_d.data[1])
    assert np.allclose(scal_ref_d.data[2], scal_d.data[2])


def test_randomVelocity_m6():
    """Basic test with random velocity. Using M6prime"""
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar')
    scal_ref = Field(domain=box, name='Scalar_ref')
    velo = Field(domain=box, name='Velocity', isVector=True)
    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='scales_p_M6'
                      )
    advec_py = Advection(velo, scal_ref,
                         resolutions={velo: [17, 17, 17],
                                      scal_ref: [17, 17, 17]},
                         method='m6prime',
                         splittingConfig='o2_FullHalf'
                         )
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    scal_ref_d = scal_ref.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]

    scal_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_ref_d.data[0][...] = scal_d.data[0][...]
    velo_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[0])
    velo_d.data[1][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[1])
    velo_d.data[2][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[1])

    advec.apply(Simulation(0., 0.1, 1))
    advec_py.apply(Simulation(0., 0.1, 1))

    assert np.allclose(scal_ref_d.data[0], scal_d.data[0])


def test_randomVelocity_vec_m6():
    """Basic test with random velocity vector Field. Using M6prime"""
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar', isVector=True)
    scal_ref = Field(domain=box, name='Scalar_ref', isVector=True)
    velo = Field(domain=box, name='Velocity', isVector=True)
    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='scales_p_M6'
                      )
    advec_py = Advection(velo, scal_ref,
                         resolutions={velo: [17, 17, 17],
                                      scal_ref: [17, 17, 17]},
                         method='m6prime',
                         splittingConfig='o2_FullHalf'
                         )
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    scal_ref_d = scal_ref.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]

    scal_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_d.data[1][...] = np.asarray(
        np.random.random(scal_d.data[1].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_d.data[2][...] = np.asarray(
        np.random.random(scal_d.data[2].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_ref_d.data[0][...] = scal_d.data[0][...]
    scal_ref_d.data[1][...] = scal_d.data[1][...]
    scal_ref_d.data[2][...] = scal_d.data[2][...]
    velo_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[0])
    velo_d.data[1][...] = np.asarray(
        np.random.random(scal_d.data[1].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[1])
    velo_d.data[2][...] = np.asarray(
        np.random.random(scal_d.data[2].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[1])

    advec.apply(Simulation(0., 0.1, 1))
    advec_py.apply(Simulation(0., 0.1, 1))

    assert np.allclose(scal_ref_d.data[0], scal_d.data[0])
    assert np.allclose(scal_ref_d.data[1], scal_d.data[1])
    assert np.allclose(scal_ref_d.data[2], scal_d.data[2])


def test_randomVelocity_m8():
    """Basic test with random velocity. Using M8prime"""
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar')
    scal_ref = Field(domain=box, name='Scalar_ref')
    velo = Field(domain=box, name='Velocity', isVector=True)
    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='scales_p_M8'
                      )
    advec_py = Advection(velo, scal_ref,
                         resolutions={velo: [17, 17, 17],
                                      scal_ref: [17, 17, 17]},
                         method='m8prime',
                         splittingConfig='o2_FullHalf'
                         )
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    scal_ref_d = scal_ref.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]

    scal_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_ref_d.data[0][...] = scal_d.data[0][...]
    velo_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[0])
    velo_d.data[1][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[1])
    velo_d.data[2][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[1])

    advec.apply(Simulation(0., 0.1, 1))
    advec_py.apply(Simulation(0., 0.1, 1))

    assert np.allclose(scal_ref_d.data[0], scal_d.data[0], atol=1e-07)


def test_randomVelocity_vec_m8():
    """Basic test with random velocity vector Field. Using M8prime"""
    box = Box(3, length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar', isVector=True)
    scal_ref = Field(domain=box, name='Scalar_ref', isVector=True)
    velo = Field(domain=box, name='Velocity', isVector=True)
    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      method='scales_p_M8'
                      )
    advec_py = Advection(velo, scal_ref,
                         resolutions={velo: [17, 17, 17],
                                      scal_ref: [17, 17, 17]},
                         method='m8prime',
                         splittingConfig='o2_FullHalf'
                         )
    advec.setUp()
    advec_py.setUp()

    scal_d = scal.discreteFields.values()[0]
    scal_ref_d = scal_ref.discreteFields.values()[0]
    velo_d = velo.discreteFields.values()[0]

    scal_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_d.data[1][...] = np.asarray(
        np.random.random(scal_d.data[1].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_d.data[2][...] = np.asarray(
        np.random.random(scal_d.data[2].shape),
        dtype=PARMES_REAL, order=ORDER)
    scal_ref_d.data[0][...] = scal_d.data[0][...]
    scal_ref_d.data[1][...] = scal_d.data[1][...]
    scal_ref_d.data[2][...] = scal_d.data[2][...]
    velo_d.data[0][...] = np.asarray(
        np.random.random(scal_d.data[0].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[0])
    velo_d.data[1][...] = np.asarray(
        np.random.random(scal_d.data[1].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[1])
    velo_d.data[2][...] = np.asarray(
        np.random.random(scal_d.data[2].shape),
        dtype=PARMES_REAL, order=ORDER) / (2. * scal_d.resolution[1])

    advec.apply(Simulation(0., 0.1, 1))
    advec_py.apply(Simulation(0., 0.1, 1))

    assert np.allclose(scal_ref_d.data[0], scal_d.data[0], atol=1e-07)
    assert np.allclose(scal_ref_d.data[1], scal_d.data[1], atol=1e-07)
    assert np.allclose(scal_ref_d.data[2], scal_d.data[2], atol=1e-07)


# if __name__ == '__main__':
#     test_nullVelocity_m4()
#     test_nullVelocity_m4()
#     test_nullVelocity_m6()
#     test_nullVelocity_m8()
#     test_nullVelocity_vec_m4()
#     test_nullVelocity_vec_m6()
#     test_nullVelocity_vec_m8()
#     test_randomVelocity_m4()
#     test_randomVelocity_m6()
#     test_randomVelocity_m8()
#     test_randomVelocity_vec_m4()
#     test_randomVelocity_vec_m6()
#     test_randomVelocity_vec_m8()
