"""
@file problem.py

Complete problem description.
"""
from parmepy.constants import debug
from parmepy import __VERBOSE__
from parmepy.operator.monitors.monitoring import Monitoring
from parmepy.operator.redistribute import Redistribute
from parmepy.tools.timers import Timer, timed_function
from parmepy.mpi import main_rank
from scitools.NumPyDB import NumPyDB_cPickle


class Problem(object):
    """
    Problem representation.

    Contains several operators and monitors that apply on variables.
    Variables are defined on different domains.\n
    Each operator is set up and variables are initialized in a set up step.\n
    To solve the problem, a loop over time-steps is launched. A step consists
    in calling the apply method of each operators and monitors.\n
    To finish, a finalize method is called.\

    \remark: In case of GPU real-time rendering (i.e. use of an
    OpenGLRendering object), The loop over time-steps is passed to Qt4
    """

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, operators, simulation, monitors=[],
                 dumpFreq=100, name=None):
        """
        Create a transport problem instance.

        @param operators : list of operators.
        @param simulation : a parmepy.simulation.Simulation object
        to describe simulation parameters.
        @param monitors : list of monitors.
        @param name : an id for the problem
        @param dumpFreq : frequency of dump (i.e. saving to a file)
        for the problem; set dumpFreq = -1 for no dumps. Default = 100.
        """
        ## Problem name
        self.name = name
        ## Problem operators
        self.operators = operators
        for m in monitors:
            self.operators.append(m)
        ## Computes time step and manage iterations
        self.simulation = simulation
        self.domain = self.operators[0].variables[0].domain
        for op in self.operators:
            for v in op.variables:
                if not self.domain is v.domain:
                    raise ValueError("Problem must have only one " +
                                     "domain for variables.")
        ## A list of variables that must be initialized before
        ## any call to op.apply()
        self.input = []
        ## Object to store computational times of lower level functions
        self.timer = Timer(self)
        ## call to problem.dump frequency during apply.
        if dumpFreq >= 0:
            ## dump problem every self.dumpFreq iter
            self.dumpFreq = dumpFreq
            self._doDump = True
        else:
            self._doDump = False
            self.dumpFreq = 100000

        ## Id for the problem. Used for dump file name.
        if name is None:
            self.name = 'parmesPb'
        else:
            self.name = name
        ## Default file name prefix for dump.
        self.filename = str(self.name)
        self._filedump = self.filename + '_rk_' + str(main_rank)

    @debug
    @timed_function
    def setUp(self):
        """
        Set solver.
        """
        for op in self.operators:
            if not isinstance(op, Monitoring):
                if not isinstance(op, Redistribute):
                    op.setUp()

        if  __VERBOSE__ and main_rank == 0:
            print "==== Variables initialization ===="

        # Build variables list to initialize
        # These are operators input variables that are not output of
        # previous operators in the operator stack
        self.input += self.operators[-1].input
        for op in self.operators[:-1][::-1]:
            for v in op.output:
                if v in self.input:
                    self.input.remove(v)
            for v in op.input:
                if not v in self.input:
                    self.input.append(v)

        for v in self.input:
            v.initialize()

        for op in self.operators:
            if isinstance(op, Monitoring):
                op.setUp()
            if isinstance(op, Redistribute):
                op.setUp()

        if __VERBOSE__ and main_rank == 0:
            print "===="
        for op in self.operators:
            try:
                if op.isGLRender:
                    self.glRenderer = op
                    op.setMainLoop(self)
            except AttributeError:
                pass

    @debug
    @timed_function
    def solve(self):
        """
        Solve problem.

        Performs simulations iterations by calling each
        operators of the list until timer ends.\n
        At end of time step, call an io step.\n
        Displays timings at simulation end.
        """
        if main_rank == 0:
            print "\n\n Start solving ..."
        try:
        ## Pass the main loop to the OpenGL viewer
            if not self.glRenderer is None:
                self.glRenderer.startMainLoop()
        except AttributeError:
            while not self.simulation.isOver:
                self.simulation.printState()
                for op in self.operators:
                    op.apply(self.simulation)
                testdump = \
                    self.simulation.currentIteration % self.dumpFreq is 0
                self.simulation.advance()
                if self._doDump and testdump:
                    self.dump()

    @debug
    def finalize(self):
        """
        Finalize method
        """
        if main_rank == 0:
            print "\n\n==== End ===="
        ## We terminate monitors before operators.
        for op in self.operators:
            if isinstance(op, Monitoring):
                op.finalize()
        for op in self.operators:
            if not isinstance(op, Monitoring):
                op.finalize()
        print self.timer
        if main_rank == 0:
            print "===\n"

    def __str__(self):
        """ToString method"""
        s = "Problem based on\n"
        s += str(self.domain.topologies)
        s += "with following operators : \n"
        for op in self.operators:
            s += str(op)
        s += "working on fields : \n"
        for f in self.variables:
            s += str(f)
        s += "defined on domains : \n"
        for d in self.domains:
            s += str(d)
        s += "\nsolved by : \n" + str(self.solver)

        return s

    def dump(self, filename=None):
        """
        Serialize some data of the problem to file
        (only data required for a proper restart, namely fields in self.input
        and simulation).
        @param filename : prefix for output file. Real name = filename_rk_N,
        N being current process number. If None use default value from problem
        parameters (self.filename)
        """
        if filename is not None:
            self.filename = filename
            self._filedump = filename + '_rk_' + str(main_rank)
        db = NumPyDB_cPickle(self._filedump, mode='store')
        db.dump(self.simulation, 'simulation')
        for v in self.input:
            print 'dump v ...'
            print v
            v.dump(self.filename, mode='append')

    def restart(self, filename=None):
        """
        Load serialized data to restart from a previous state.
        self.input variables and simulation are loaded.
        @param  filename : prefix for downloaded file.
        Real name = filename_rk_N, N being current process number.
        If None use default value from problem
        parameters (self.filename)
        """
        if filename is not None:
            self.filename = filename
            self._filedump = filename + '_rk_' + str(main_rank)
        db = NumPyDB_cPickle(self._filedump, mode='load')
        self.simulation = db.load('simulation')[0]
        self.simulation.reset()
        for v in self.input:
            print "load ...", self.filename
            v.load(self.filename)

        for op in self.operators:
            if isinstance(op, Monitoring):
                op.setUp()
            if isinstance(op, Redistribute):
                op.setUp()

    def setDumpFreq(self, freq):
        """
        set rate of problem.dump call (every 'rate' iteration)
        @param freq : the frequency of output
        """
        self.dumpFreq = freq
        if freq < 0:
            self._doDump = False

if __name__ == "__main__":
    print __doc__
    print "- Provided class : problem"
    print Problem.__doc__
