!> @addtogroup part
!! @{
!------------------------------------------------------------------------------
!
! MODULE: advec
!
!
! DESCRIPTION:
!> The module advec provides all public interfaces to solve an advection equation
!! with a particle method.
!
!> @details
!!     This module contains the generic procedure to initialize and parametrise the
!! advection solver based on particles method. It also contains the subroutine
!! "advec_step" wich solves the equation for a given time step. It is the only one
!! module which is supposed to be included by a code using this library of
!! particle methods.
!
!> @author
!! Jean-Baptiste Lagaert, LEGI
!
!------------------------------------------------------------------------------
module advec_Vect

    use precision_tools
    use advec_abstract_proc
    implicit none

    ! ===== Private variables =====
    !> numerical method use to advect the scalar
    character(len=str_short), private   :: type_part_solv
    !> dimensionnal splitting (eg classical, Strang or particle)
    character(len=str_short), private   :: dim_splitting
    !> Group size along current direction
    integer, private, dimension(2)  :: gsX, gsY, gsZ


    ! ===== Public procedures =====
    ! Scheme used to advec the scalar (order 2 or 4 ?)
!    public                              :: type_part_solver

    ! Advection methods
!    public                              :: advec_init           ! initialize the scalar solver
    public                              :: advec_step_Vect      ! advec the scalar field during a time step.
!    procedure(advec_step_Torder2), pointer, public    :: advec_step => null()
!    public                              :: advec_step_Torder1   ! advec the scalar field during a time step.
!    public                              :: advec_step_Torder2   ! advec the scalar field during a time step.
!
    ! Remeshing formula
    procedure(AC_remesh), pointer, private :: advec_remesh_bis => null()

contains

! ===== Public methods =====

!> Return the name of the particle method used for the advection
!!    @return type_part_solver      = numerical method used for advection
function type_part_solver()
    character(len=str_short)    :: type_part_solver
    
    type_part_solver = type_part_solv
end function


!> Initialise the particle advection methods
!!    @param[in]    order       = to choose the remeshing method (and thus the order)
!!    @param[out]   stab_coeff  = stability coefficient (condition stability is
!!                                  dt< stab_coeff/norm_inf(V))
!!    @param[in]    dim_split   = dimensionnal splitting (eg classical,
!!                                    Strang splitting or particle splitting)
!!    @param[in]    verbosity   = to display info about chosen remeshing formula (optional)
subroutine advec_init_Vect(order, stab_coeff, verbosity, dim_split)

    use advec_variables       ! contains info about solver parameters and others.
    use cart_topology   ! Description of mesh and of mpi topology
    use advecX          ! solver for advection along X
    use advecY          ! solver for advection along Y
    use advecZ          ! solver for advection along Z
    use advec_common    ! some procedures common to advection along all directions

    ! Input/Output
    character(len=*), optional, intent(in)  ::  order, dim_split
    logical, optional, intent(in)           ::  verbosity
    real(WP), optional, intent(out)         ::  stab_coeff

    ! Use default solver if it is not chosen by the user.
    if(present(order)) then
        type_part_solv = order
    else
        type_part_solv = 'p_O2'
    end if

    ! Initialize the solver
    if (present(verbosity)) then
        call AC_solver_init(type_part_solv, verbosity)
    else
        call AC_solver_init(type_part_solv)
    end if

    if (present(stab_coeff)) stab_coeff = 1.0/(dble(bl_size))

    ! Call the right remeshing formula
    select case(type_part_solv)
        case('p_O2')
            advec_remesh_bis => AC_remesh_lambda_group
        case('p_O4')
            advec_remesh_bis => AC_remesh_lambda_group
        case('p_L2')
            advec_remesh_bis => AC_remesh_limit_lambda_group
        case('p_M6')
            advec_remesh_bis => AC_remesh_Mprime_group
        case('p_M8')
            advec_remesh_bis => AC_remesh_Mprime_group
        case default
            advec_remesh_bis => AC_remesh_lambda_group
    end select

    call AC_setup_init()

    ! Save group size
    gsX =group_size(1,:)
    gsY =group_size(2,:)
    gsZ =group_size(3,:)

end subroutine advec_init_Vect


!> Solve advection equation - order 2 in time (order 2 dimensional splitting)
!!    @param[in]        dt          = time step
!!    @param[in]        Vx          = velocity along x (could be discretised on a bigger mesh then the scalar)
!!    @param[in]        Vy          = velocity along y
!!    @param[in]        Vz          = velocity along z
!!    @param[in,out]    scal        = scalar field to advect
subroutine advec_step_Vect(dt, Vx, Vy, Vz, scal_Vector)

    use advecX          ! Method to advec along X
    use advecY          ! Method to advec along Y
    use advecZ          ! Method to advec along Z

    ! Input/Output
    real(WP), intent(in)                        :: dt
    real(WP), dimension(:,:,:), intent(in)      :: Vx, Vy, Vz
    real(WP), dimension(:,:,:,:), intent(inout) :: scal_Vector

    call advecX_calc_Vect(dt/2.0, Vx, scal_Vector)
    call advecY_calc_Vect(dt/2.0, Vy, scal_Vector)
    call advecZ_calc_Vect(dt/2.0, Vz, scal_Vector)
    call advecZ_calc_Vect(dt/2.0, Vz, scal_Vector)
    call advecY_calc_Vect(dt/2.0, Vy, scal_Vector)
    call advecX_calc_Vect(dt/2.0, Vx, scal_Vector)

end subroutine advec_step_Vect


!> Scalar advection (this procedure call the right solver, depending on the simulation setup)
!!    @param[in]        dt          = time step
!!    @param[in]        Vx          = velocity along X (could be discretised on a bigger mesh then the scalar)
!!    @param[in,out]    scal_vect   = scalar field to advect
subroutine advecX_calc_Vect(dt, Vx, scal_vect)

    use advecX          ! Procedure specific to advection along X
    use advec_common    ! Some procedures common to advection along all directions
    use advec_variables ! contains info about solver parameters and others.
    use cart_topology   ! Description of mesh and of mpi topology

    ! Input/Output
    real(WP), intent(in)                                                :: dt
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3)), intent(in)    :: Vx
    real(WP), dimension(:,:,:,:), intent(inout)                         :: scal_vect
    ! Other local variables
    integer, parameter                                  :: direction =1 ! current direction
    integer                                             :: j,k          ! indice of the currend mesh point
    integer                                             :: sca          ! indice of the currend scalar field
    integer, dimension(2)                               :: ind_group    ! indice of the currend group of line (=(i,k) by default)
    real(WP),dimension(N_proc(direction),gsX(1),gsX(2)) :: p_pos_adim   ! adimensionned particles position
    real(WP),dimension(N_proc(direction),gsX(1),gsX(2)) :: p_V          ! particles velocity

    ! Allocate send_ind_min/max
    if(allocated(send_group_min)) deallocate(send_group_min)
    allocate(send_group_min(group_size(direction,1),group_size(direction,2)))
    if(allocated(send_group_max)) deallocate(send_group_max)
    allocate(send_group_max(group_size(direction,1),group_size(direction,2)))

    ! Initialise the pointer for optimized remeshing
    call AC_setup_alongX()

    ind_group = 0

    do k = 1, N_proc(3), gsX(2)
        ind_group(2) = ind_group(2) + 1
        ind_group(1) = 0
        do j = 1, N_proc(2), gsX(1)
            ind_group(1) = ind_group(1) + 1

            ! ===== Init particles =====
            call advecX_init_group(Vx, j, k, gsX, p_pos_adim, p_V)

            ! ===== Advection =====
            ! -- Compute velocity (with a RK2 scheme) --
            call AC_velocity_interpol_group(dt, direction, gsX, ind_group, p_pos_adim, p_V)

            ! -- Advec particles --
            p_pos_adim = p_pos_adim + dt*p_V/d_sc(direction)

            ! ===== Remeshing =====
            do sca = 1, size(scal_vect,4)
                call advec_remesh_bis(direction, ind_group, gsX, &
                    & p_pos_adim, p_V, j, k, scal_Vect(:,:,:,sca), dt)
            end do

        end do
    end do

end subroutine advecX_calc_Vect


!> Scalar advection along Y (this procedure call the right solver, depending on the simulation setup)
!!    @param[in]        dt          = time step
!!    @param[in]        Vy          = velocity along y (could be discretised on a bigger mesh then the scalar)
!!    @param[in,out]    scal3D      = scalar field to advect
subroutine advecY_calc_Vect(dt, Vy, scal_vect)

    use advecY          ! Procedure specific to advection along Y
    use advec_common    ! Some procedures common to advection along all directions
    use advec_variables ! contains info about solver parameters and others.
    use cart_topology   ! Description of mesh and of mpi topology

    ! Input/Output
    real(WP), intent(in)                                                :: dt
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3)), intent(in)    :: Vy
    real(WP), dimension(:,:,:,:), intent(inout)                         :: scal_vect
    ! Other local variables
    integer, parameter                                  :: direction =2 ! current direction
    integer                                             :: i,k          ! indice of the currend mesh point
    integer                                             :: sca          ! indice of the currend scalar field
    integer, dimension(2)                               :: ind_group    ! indice of the currend group of line (=(i,k) by default)
    real(WP),dimension(N_proc(direction),gsY(1),gsY(2)) :: p_pos_adim   ! adimensionned particles position
    real(WP),dimension(N_proc(direction),gsY(1),gsY(2)) :: p_V          ! particles velocity

    ! Allocate send_ind_min/max
    if(allocated(send_group_min)) deallocate(send_group_min)
    allocate(send_group_min(group_size(direction,1),group_size(direction,2)))
    if(allocated(send_group_max)) deallocate(send_group_max)
    allocate(send_group_max(group_size(direction,1),group_size(direction,2)))

    ! Initialise the pointer for optimized remeshing
    call AC_setup_alongY()

    ind_group = 0

    do k = 1, N_proc(3), gsY(2)
        ind_group(2) = ind_group(2) + 1
        ind_group(1) = 0
        do i = 1, N_proc(1), gsY(1)
            ind_group(1) = ind_group(1) + 1

            ! ===== Init particles =====
            call advecY_init(Vy, i, k, gsY, p_pos_adim, p_V)

            ! ===== Advection =====
            ! -- Compute velocity (with a RK2 scheme) --
            call AC_velocity_interpol_group(dt, direction, gsY, ind_group, p_pos_adim, p_V)

            ! -- Advec particles --
            p_pos_adim = p_pos_adim + dt*p_V/d_sc(direction)

            ! ===== Remeshing =====
            do sca = 1, size(scal_vect,4)
                call advec_remesh_bis(direction, ind_group, gsY, p_pos_adim, p_V, i, k, scal_vect(:,:,:,sca), dt)
            end do

        end do
    end do

end subroutine advecY_calc_Vect


!> Scalar advection alongZ (this procedure call the right solver, depending on the simulation setup)
!!    @param[in]        dt          = time step
!!    @param[in]        Vz          = velocity along y (could be discretised on a bigger mesh then the scalar)
!!    @param[in,out]    scal3D      = scalar field to advect
subroutine advecZ_calc_Vect(dt, Vz, scal_vect)

    use advec_variables ! contains info about solver parameters and others.
    use cart_topology   ! Description of mesh and of mpi topology
    use advecZ          ! procdure devoted to advection along Z
    use advec_common    ! some procedures common to advection along all directions

    ! Input/Output
    real(WP), intent(in)                                                :: dt
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3)), intent(in)    :: Vz
    real(WP), dimension(:,:,:,:), intent(inout)                         :: scal_vect
    ! Other local variables
    integer, parameter                                  :: direction =3 ! current direction
    integer                                             :: i,j          ! indice of the currend mesh point
    integer                                             :: sca          ! indice of the currend scalar field
    integer, dimension(2)                               :: ind_group    ! indice of the currend group of line (=(i,k) by default)
    real(WP), dimension(N_proc(direction),gsZ(1),gsZ(2))  :: p_pos_adim ! adimensionned particles position
    real(WP), dimension(N_proc(direction),gsZ(1),gsZ(2))  :: p_V        ! particles velocity

    ! Allocate send_ind_min/max
    if(allocated(send_group_min)) deallocate(send_group_min)
    allocate(send_group_min(group_size(direction,1),group_size(direction,2)))
    if(allocated(send_group_max)) deallocate(send_group_max)
    allocate(send_group_max(group_size(direction,1),group_size(direction,2)))

    ! Initialise the pointer for optimized remeshing
    call AC_setup_alongZ()

    ind_group = 0

    do j = 1, N_proc(2), gsZ(2)
        ind_group(2) = ind_group(2) + 1
        ind_group(1) = 0
        do i = 1, N_proc(1), gsZ(1)
            ind_group(1) = ind_group(1) + 1

            ! ===== Init particles =====
            call advecZ_init_group(Vz, i, j, gsZ, p_pos_adim, p_V)

            ! ===== Advection =====
            ! -- Compute velocity (with a RK2 scheme) --
            call AC_velocity_interpol_group(dt, direction, gsZ, ind_group, p_pos_adim, p_V)
            ! -- Advec particles --
            p_pos_adim = p_pos_adim + dt*p_V/d_sc(direction)

            ! ===== Remeshing =====
            do sca = 1, size(scal_vect,4)
                call advec_remesh_bis(direction, ind_group, gsZ, p_pos_adim, p_V, i,j,scal_vect(:,:,:,sca), dt)
            end do

        end do
    end do

end subroutine advecZ_calc_Vect



!> ===== Private procedure =====
end module advec_Vect
!> @}
