!> @addtogroup part
!! @{
!------------------------------------------------------------------------------
!
! MODULE: advec_common_velo
!
!
! DESCRIPTION:
!> The module ``advec_common_velo'' gather function and subroutines used to interpolate
!! velocity at particle position which are not specific to a direction
!! @details
!! This module gathers functions and routines used to advec scalar which are not
!! specific to a direction. This is a parallel implementation using MPI and
!! the cartesien topology it provides. It also contains the variables common to
!! the solver along each direction and other generic variables used for the
!! advection based on the particle method.
!!
!! Except for testing purpose, this module is not supposed to be used by the
!! main code but only by the other advection module. More precisly, an final user
!! must only used the generic "advec" module wich contain all the interface to
!! solve the advection equation with the particle method, and to choose the
!! remeshing formula, the dimensionnal splitting and everything else. Except for
!! testing purpose, the other advection modules have only to include
!! "advec_common".
!!
!! The module "test_advec" can be used in order to validate the procedures
!! embedded in this module.
!
!> @author
!! Jean-Baptiste Lagaert, LEGI
!
!------------------------------------------------------------------------------

module advec_common_velo

    use precision_tools
    use structure_tools
    use advec_abstract_proc

    implicit none


    ! Information about the particles and their bloc
    public


    ! ===== Public procedures =====
    !----- To interpolate velocity -----
    public                        :: AC_velocity_interpol_group
    public                        :: AC_velocity_interpol_no_com
    public                        :: AC_velocity_determine_communication

    ! ===== Public variables =====

    ! ===== Private variables =====


contains

! ===== Public procedure =====

! ==================================================================================
! ====================     Compute particle velocity (RK2)      ====================
! ==================================================================================

!> Interpolate the velocity field used in a RK2 scheme for particle advection -
!! version for a group of (more of one) line
!!    @param[in]        dt          = time step
!!    @param[in]        direction   = current direction (1 = along X, 2 = along Y and 3 = along Z)
!!    @param[in]        gs          = size of a group (ie number of line it gathers along the two other directions)
!!    @param[in]        ind_group   = coordinate of the current group of lines
!!    @param[in]        p_pos_adim  = adimensionned particle postion
!!    @param[in,out]    p_V         = particle velocity (along the current direction)
!! @details
!!    A RK2 scheme is used to advect the particles : the midlle point scheme. An
!!    intermediary position "p_pos_bis(i) = p_pos(i) + V(i)*dt/2" is computed and then
!!    the numerical velocity of each particles is computed as the interpolation of V  in
!!    this point. This field is used to advect the particles at the seconde order in time :
!!    p_pos(t+dt, i) = p_pos(i) + p_V(i).
!!    The group line indice is used to ensure using unicity of each mpi message tag.
!!    The interpolation is done for a group of lines, allowing to mutualise
!!    communications. Considering a group of Na X Nb lines, communication performed
!!    by this algorithm are around (Na x Nb) bigger than the alogorithm wich
!!    works on a single line but also around (Na x Nb) less frequent.
subroutine AC_velocity_interpol_group(dt, direction, gs, ind_group, p_pos_adim, p_V)

    ! This code involve a recopy of p_V. It is possible to directly use the 3D velocity field but in a such code
    ! a memory copy is still needed to send velocity field to other processus : mpi send contiguous memory values

    use mpi
    use cart_topology   ! info about mesh and mpi topology
    use advec_variables ! contains info about solver parameters and others.

    ! Input/Ouput
    real(WP), intent(in)                            :: dt           ! time step
    integer, intent(in)                             :: direction    ! current direction
    integer, dimension(2),intent(in)                :: gs           ! groupe size
    integer, dimension(2), intent(in)               :: ind_group
    real(WP), dimension(:,:,:), intent(in)          :: p_pos_adim
    real(WP), dimension(:,:,:),intent(inout),target :: p_V
    ! Others, local
    real(WP),dimension(N_proc(direction),gs(1),gs(2)), target   :: p_posV_bis   ! temporaily field to store middle point velocity and position
    real(WP), dimension(N_proc(direction),gs(1),gs(2))          :: weight       ! interpolation weight
    type(real_pter),dimension(N_proc(direction),gs(1),gs(2))    :: Vp, Vm       ! Velocity on previous and next mesh point
    real(WP), dimension(:), allocatable, target                 :: V_buffer     ! Velocity buffer for postion outside of the local subdomain
    integer, dimension(:), allocatable                          :: pos_in_buffer! buffer size
    integer , dimension(gs(1), gs(2))           :: rece_ind_min ! minimal indice of mesh involved in remeshing particles (of my local subdomains)
    integer , dimension(gs(1), gs(2))           :: rece_ind_max ! maximal indice of mesh involved in remeshing particles (of my local subdomains)
    integer                                     :: ind, ind_com ! indices
    integer                                     :: i1, i2       ! indices in the lines group
    integer                                     :: pos, pos_old ! indices of the mesh point wich preceed the particle position
    integer                                     :: proc_gap, gap! distance between my (mpi) coordonate and coordinate of the
                                                                ! processus associated to a given position
    integer, dimension(:), allocatable          :: rece_rank    ! rank of processus wich send me information
    integer, dimension(:), allocatable          :: send_rank    ! rank of processus to wich I send information
    integer                                     :: rankP        ! rank of processus ("source rank" returned by mpi_cart_shift)
    integer, dimension(:), allocatable          :: send_carto   ! cartography of what I have to send
    integer                                     :: ind_1Dtable  ! indice of my current position inside a one-dimensionnal table
    integer                                     :: ind_for_i1   ! where to read the first coordinate (i1) of the current line inside the cartography ?
    type(real_1D_pter),dimension(:),allocatable :: send_buffer  ! to store what I have to send (on a contiguous way)
    !real(WP), dimension(:), allocatable         :: send_buffer  ! to store what I have to send (on a contiguous way)
    integer, dimension(gs(1),gs(2),2)           :: rece_gap     ! distance between me and processus wich send me information
    integer, dimension(2 , 2)                   :: send_gap     ! distance between me and processus to wich I send information
    integer, dimension(2)                       :: rece_gap_abs ! min (resp max) value of rece_gap(:,:,i) with i=1 (resp 2)
    integer                                     :: com_size     ! size of message send/receive
    integer, dimension(:), allocatable          :: size_com     ! size of message send/receive
    integer                                     :: min_size     ! minimal size of cartography(:,proc_gap)
    integer                                     :: max_size     ! maximal size of cartography(:,proc_gap)
    integer                                     :: tag          ! mpi message tag
    integer, dimension(:), allocatable          :: tag_proc     ! mpi message tag
    integer                                     :: ierr         ! mpi error code
    integer, dimension(:), allocatable          :: s_request    ! mpi communication request (handle) of nonblocking send
    integer, dimension(:), allocatable          :: s_request_bis! mpi communication request (handle) of nonblocking send
    integer, dimension(:), allocatable          :: rece_request ! mpi communication request (handle) of nonblocking receive
    integer, dimension(MPI_STATUS_SIZE)         :: rece_status  ! mpi status (for mpi_wait)
    type(int_1D_pter), dimension(:), allocatable:: cartography  ! cartography(proc_gap) contains the set of the lines indice in the block for wich the
                                                                ! current processus requiers data from proc_gap and for each of these lines the range
                                                                ! of mesh points from where it requiers the velocity values.
! XXX Todo : à vérifier mais il semble inutile de stocker send_rank : il n'est
! utilisé que pour les mpi_wait mais une initialisation des request à
! MPI_REQUEST_NULL devrait permettre de s'en débarrasser.

    ! -- Initialisation --
    do i2 = 1, gs(2)
        do i1 = 1, gs(1)
            do ind = 1, N_proc(direction)
                nullify(Vp(ind,i1,i2)%pter)
                nullify(Vm(ind,i1,i2)%pter)
            end do
        end do
    end do
    ! Compute the midlle point
    p_posV_bis = p_pos_adim + (dt/2.0)*p_V/d_sc(direction)

! XXX Todo / Optim
! Ici recopie de la vitesse. On doit la faire car on calcule la vitesse
! interpolée à partir d' "elle-même" : la variable calculée dépend de sa
! précédente valeur en des points qui peuvent différé. Si on l'écrase au fur et
! à mesure on fait des calculs faux.
! On pourrait utiliser directement le champ de vitesse V en entrée pour donner
! p_V en sortie, mais cela pose un problème car selon les directions il faudrait
! changer l'ordre des indices i, i1 et i2. Ce qui est un beau bordel !!
! XXX
    ! Compute range of the set of point where I need the velocity value
    rece_ind_min = floor(p_posV_bis(1,:,:))
    rece_ind_max = floor(p_posV_bis(N_proc(direction),:,:)) + 1

    ! ===== Exchange velocity field if needed =====
    ! It uses non blocking message to do the computations during the communication process
    ! -- What have I to communicate ? --
    rece_gap(:,:,1) = floor(real(rece_ind_min-1)/N_proc(direction))
    rece_gap(:,:,2) = floor(real(rece_ind_max-1)/N_proc(direction))
    rece_gap_abs(1) = minval(rece_gap(:,:,1))
    rece_gap_abs(2) = maxval(rece_gap(:,:,2))
    max_size = 2 + gs(2)*(2+3*gs(1))
    allocate(cartography(rece_gap_abs(1):rece_gap_abs(2)))
    allocate(rece_rank(rece_gap_abs(1):rece_gap_abs(2)))
    call AC_velocity_determine_communication(direction, ind_group, gs, send_gap,  &
    & rece_gap, rece_gap_abs, rece_rank, cartography, max_size)

    ! -- Send messages about what I want --
    allocate(s_request_bis(rece_gap_abs(1):rece_gap_abs(2)))
    allocate(size_com(rece_gap_abs(1):rece_gap_abs(2)))
    allocate(tag_proc(rece_gap_abs(1):rece_gap_abs(2)))
    min_size = 2 + gs(2)
    do proc_gap = rece_gap_abs(1), rece_gap_abs(2)
        if (rece_rank(proc_gap) /= D_rank(direction)) then
            ! Use the cartography to know which lines are concerned
            size_com(proc_gap) = cartography(proc_gap)%pter(2)
            ! Range I want - store into the cartography
            gap = proc_gap*N_proc(direction)
            ! Position in cartography(proc_gap)%pter(:) of the current i1 indice
            ind_for_i1 = min_size
            do i2 = 1, gs(2)
                do ind = ind_for_i1+1, ind_for_i1 + cartography(proc_gap)%pter(2+i2), 2
                    do i1 = cartography(proc_gap)%pter(ind), cartography(proc_gap)%pter(ind+1)
                        ! Interval start from:
                        cartography(proc_gap)%pter(size_com(proc_gap)+1) = max(rece_ind_min(i1,i2), gap+1) ! fortran => indice start from 0
                        ! and ends at:
                        cartography(proc_gap)%pter(size_com(proc_gap)+2) = min(rece_ind_max(i1,i2), gap+N_proc(direction))
                        ! update number of element to receive
                        cartography(proc_gap)%pter(1) = cartography(proc_gap)%pter(1)       &
                                    & + cartography(proc_gap)%pter(size_com(proc_gap)+2)    &
                                    & - cartography(proc_gap)%pter(size_com(proc_gap)+1) + 1
                        size_com(proc_gap) = size_com(proc_gap)+2
                    end do
                end do
                ind_for_i1 = ind_for_i1 + cartography(proc_gap)%pter(2+i2)
            end do
            ! Tag = concatenation of (rank+1), ind_group(1), ind_group(2), direction et unique Id.
            tag_proc(proc_gap) = compute_tag(ind_group, tag_velo_range, direction, proc_gap)
            ! Send message
            call mpi_ISsend(cartography(proc_gap)%pter(1), size_com(proc_gap), MPI_INTEGER, rece_rank(proc_gap), &
                & tag_proc(proc_gap), D_comm(direction), s_request_bis(proc_gap),ierr)
        end if
    end do

    ! -- Send the velocity field to processus which need it --
    allocate(s_request(send_gap(1,1):send_gap(1,2)))
    allocate(send_rank(send_gap(1,1):send_gap(1,2)))
    allocate(send_carto(max_size))
    allocate(send_buffer(send_gap(1,1):send_gap(1,2)))
! XXX Todo : compter le nombre de messages à recevoir puis les traiter dans
! l'ordre où ils arrivent via un MPI_ANY_PROC ? Mais alors il faut lier rang et
! coordonnées ... ce qui signifie ajouter un appel à un mpi_cart_cood ... ou
! envoyer le rand dans la cartographie !!
! A voir ce qui est le mieux.
    do proc_gap = send_gap(1,1), send_gap(1,2)
        call mpi_cart_shift(D_comm(direction), 0, proc_gap, rankP, send_rank(proc_gap), ierr)
        if (send_rank(proc_gap) /= D_rank(direction)) then
            ! I - Receive messages about what I have to send
            ! Ia - Compute reception tag = concatenation of (rank+1), ind_group(1), ind_group(2), direction et unique Id.
            tag = compute_tag(ind_group, tag_velo_range, direction, -proc_gap)
            ! Ib - Receive the message
            call mpi_recv(send_carto(1), max_size, MPI_INTEGER, send_rank(proc_gap), tag, D_comm(direction), rece_status, ierr)
            ! II - Send it
            ! IIa - Create send buffer
            allocate(send_buffer(proc_gap)%pter(send_carto(1)))
            !allocate(send_buffer(send_carto(1)))
            gap = proc_gap*N_proc(direction)
            com_size = 0
            ind_1Dtable = send_carto(2)
            ! Position in cartography(:,proc_gap) of the current i1 indice
            ind_for_i1 = min_size
            do i2 = 1, gs(2)
                do ind = ind_for_i1+1, ind_for_i1 + send_carto(2+i2), 2
                    do i1 = send_carto(ind), send_carto(ind+1)
                        do ind_com = send_carto(ind_1Dtable+1)+gap, send_carto(ind_1Dtable+2)+gap ! indice inside the current line
                            com_size = com_size + 1
                            send_buffer(proc_gap)%pter(com_size) = p_V(ind_com, i1,i2)
                            !send_buffer(com_size) = p_V(ind_com, i1,i2)
                        end do
                        ind_1Dtable = ind_1Dtable + 2
                    end do
                end do
                ind_for_i1 = ind_for_i1 + send_carto(2+i2)
            end do
            ! IIb - Compute send tag
            tag = compute_tag(ind_group, tag_velo_V, direction, proc_gap)
            ! IIc - Send message
            !call mpi_Isend(send_buffer(1), com_size, MPI_DOUBLE_PRECISION, &
            call mpi_ISsend(send_buffer(proc_gap)%pter(1), com_size, MPI_DOUBLE_PRECISION, &
                    & send_rank(proc_gap), tag, D_comm(direction), s_request(proc_gap), ierr)
            !deallocate(send_buffer)
        end if
    end do
    deallocate(send_carto)

    ! -- Non blocking reception of the velocity field --
    ! Allocate the pos_in_buffer to compute V_buffer size and to be able to
    ! allocate it.
    allocate(pos_in_buffer(rece_gap_abs(1):rece_gap_abs(2)))
    pos_in_buffer(rece_gap_abs(1)) = 1
    do proc_gap = rece_gap_abs(1), rece_gap_abs(2)-1
        pos_in_buffer(proc_gap+1)= pos_in_buffer(proc_gap) + cartography(proc_gap)%pter(1)
    end do
    allocate(V_buffer(pos_in_buffer(rece_gap_abs(2)) &
                & + cartography(rece_gap_abs(2))%pter(1)))
    V_buffer = 0
    allocate(rece_request(rece_gap_abs(1):rece_gap_abs(2)))
    do proc_gap = rece_gap_abs(1), rece_gap_abs(2)
        if (rece_rank(proc_gap) /= D_rank(direction)) then
            ! IIa - Compute reception tag
            tag = compute_tag(ind_group, tag_velo_V, direction, -proc_gap)
            ! IIb - Receive message
            call mpi_Irecv(V_buffer(pos_in_buffer(proc_gap)), cartography(proc_gap)%pter(1), MPI_DOUBLE_PRECISION, &
                    & rece_rank(proc_gap), tag, D_comm(direction), rece_request(proc_gap), ierr)
        end if
    end do

    !-- Free som ISsend buffer and some array --
    do proc_gap = rece_gap_abs(1), rece_gap_abs(2)
      !if (associated(cartography(proc_gap)%pter)) deallocate(cartography(proc_gap)%pter)
      deallocate(cartography(proc_gap)%pter)
    end do
    deallocate(cartography) ! We do not need it anymore
! XXX Todo : préférer un call MPI_WAIT_ALL couplé avec une init de s_request_bis
! sur MPI_REQUEST_NULL et enlever la boucle ET le if.
    do proc_gap = rece_gap_abs(1), rece_gap_abs(2)
        if (rece_rank(proc_gap) /= D_rank(direction)) then
            call MPI_WAIT(s_request_bis(proc_gap),rece_status,ierr)
        end if
    end do
    deallocate(s_request_bis)
    deallocate(tag_proc)
    deallocate(size_com)

    ! ===== Compute the interpolated velocity =====
    ! -- Compute the interpolation weight and update the pointers Vp and Vm --
    do i2 = 1, gs(2)
        do i1 = 1, gs(1)
            ! Initialisation of reccurence process
            ind = 1
            pos = floor(p_posV_bis(ind,i1,i2))
            weight(ind,i1,i2) = p_posV_bis(ind,i1,i2)-pos
            ! Vm = V(pos)
            proc_gap = floor(real(pos-1)/N_proc(direction))
            if (rece_rank(proc_gap) == D_rank(direction)) then
                Vm(ind,i1,i2)%pter => p_V(pos-proc_gap*N_proc(direction), i1,i2)
            else
                Vm(ind,i1,i2)%pter => V_buffer(pos_in_buffer(proc_gap))
                pos_in_buffer(proc_gap) = pos_in_buffer(proc_gap) + 1
            end if
            ! Vp = V(pos+1)
            proc_gap = floor(real(pos+1-1)/N_proc(direction))
            if (rece_rank(proc_gap) == D_rank(direction)) then
                Vp(ind,i1,i2)%pter => p_V(pos+1-proc_gap*N_proc(direction), i1,i2)
            else
                Vp(ind,i1,i2)%pter => V_buffer(pos_in_buffer(proc_gap))
                pos_in_buffer(proc_gap) = pos_in_buffer(proc_gap) + 1
            end if
            pos_old = pos

            ! Following indice : we use previous work (already done)
            do ind = 2, N_proc(direction)
                pos = floor(p_posV_bis(ind,i1,i2))
                weight(ind,i1,i2) = p_posV_bis(ind,i1,i2)-pos
                select case(pos-pos_old)
                    case(0)
                        ! The particle belongs to the same segment than the previous one
                        Vm(ind,i1,i2)%pter => Vm(ind-1,i1,i2)%pter
                        Vp(ind,i1,i2)%pter => Vp(ind-1,i1,i2)%pter
                    case(1)
                        ! The particle follows the previous one
                        Vm(ind,i1,i2)%pter => Vp(ind-1,i1,i2)%pter
                        ! Vp = V(pos+1)
                        proc_gap = floor(real(pos+1-1)/N_proc(direction)) ! fortran -> indice starts from 1
                        if (rece_rank(proc_gap) == D_rank(direction)) then
                            Vp(ind,i1,i2)%pter => p_V(pos+1-proc_gap*N_proc(direction), i1,i2)
                        else
                            Vp(ind,i1,i2)%pter => V_buffer(pos_in_buffer(proc_gap))
                            pos_in_buffer(proc_gap) = pos_in_buffer(proc_gap) + 1
                        end if
                    case(2)
                        ! pos = pos_old +2, wich correspond to "extention"
                        ! Vm = V(pos)
                        proc_gap = floor(real(pos-1)/N_proc(direction))
                        if (rece_rank(proc_gap) == D_rank(direction)) then
                            Vm(ind,i1,i2)%pter => p_V(pos-proc_gap*N_proc(direction), i1,i2)
                        else
                            Vm(ind,i1,i2)%pter => V_buffer(pos_in_buffer(proc_gap))
                            pos_in_buffer(proc_gap) = pos_in_buffer(proc_gap) + 1
                        end if
                        ! Vp = V(pos+1)
                        proc_gap = floor(real(pos+1-1)/N_proc(direction))
                        if (rece_rank(proc_gap) == D_rank(direction)) then
                            Vp(ind,i1,i2)%pter => p_V(pos+1-proc_gap*N_proc(direction), i1,i2)
                        else
                            Vp(ind,i1,i2)%pter => V_buffer(pos_in_buffer(proc_gap))
                            pos_in_buffer(proc_gap) = pos_in_buffer(proc_gap) + 1
                        end if
                    case default
                        print*, "unexpected case : pos = ", pos, " , pos_old = ", pos_old, &
                            & " ind = ", ind, " i1 = ", i1, " i2 = ", i2
                end select
                pos_old = pos
                end do ! loop on particle indice inside the current line
        end do ! loop on first coordinate (i1) of a line inside the block of line
    end do ! loop on second coordinate (i2) of a line inside the block of line

    deallocate(pos_in_buffer)   ! We do not need it anymore

    ! -- Compute the interpolate velocity --
    ! Check if communication are done
    do proc_gap = rece_gap_abs(1), rece_gap_abs(2)
        if (rece_rank(proc_gap)/=D_rank(direction)) then
            call mpi_wait(rece_request(proc_gap), rece_status, ierr)
        end if
    end do
    deallocate(rece_request)

    ! Then compute the field
    do i2 = 1, gs(2)
        do i1 = 1, gs(1)
            do ind = 1, N_proc(direction)
                p_posV_bis(ind,i1,i2) = weight(ind,i1,i2)*Vp(ind,i1,i2)%pter + (1.-weight(ind,i1,i2))*Vm(ind,i1,i2)%pter
            end do
        end do
    end do
    p_V = p_posV_bis


    ! ===== Free memory =====
    ! -- Pointeur --
    do i2 = 1, gs(2)
        do i1 = 1, gs(1)
            do ind = 1, N_proc(direction)
                nullify(Vp(ind,i1,i2)%pter)
                nullify(Vm(ind,i1,i2)%pter)
            end do
        end do
    end do
    ! -- Mpi internal buffer for non blocking communication --
    do proc_gap = send_gap(1,1), send_gap(1,2)
        if (send_rank(proc_gap) /= D_rank(direction)) then
            call MPI_WAIT(s_request(proc_gap),rece_status,ierr)
            deallocate(send_buffer(proc_gap)%pter)
        end if
    end do
    deallocate(send_buffer)
    deallocate(s_request)
    ! -- Deallocate dynamic array --
    deallocate(V_buffer)
    deallocate(rece_rank)
    deallocate(send_rank)

end subroutine AC_velocity_interpol_group


!> Determine the set of processes wich will send me information during the velocity interpolation and compute
!! for each of these processes the range of wanted data.
!!    @param[in]    direction       = current direction (1 = along X, 2 = along Y, 3 = along Z)
!!    @param[in]    gp_s            = size of a group (ie number of line it gathers along the two other directions)
!!    @param[in]    ind_group       = coordinate of the current group of lines
!!    @param[in]    ind_group       = coordinate of the current group of lines
!!    @param[out]   send_gap        = gap between my coordinate and the processes of minimal coordinate which will send information to me
!!    @param[in]    rece_gap        = gap between my coordinate and the processes of maximal coordinate which will receive information from me
!!    @param[in]    rece_gap_abs    = min (resp max) value of rece_gap(:,:,i) with i=1 (resp 2)
!!    @param[out]   cartography     = cartography(proc_gap) contains the set of the lines indice in the block for wich the
!!                                    current processus requiers data from proc_gap and for each of these lines the range
!!                                    of mesh points from where it requiers the velocity values.
!! @details
!!    Work on a group of line of size gs(1) x gs(2))
!!    Obtain the list of processus wich need a part of my local velocity field
!!    to interpolate the velocity used in the RK2 scheme to advect its particles.
!!    In the same time, it computes for each processus from which I need a part
!!    of the velocity field, the range of mesh point where I want data and store it
!!    by using some sparse matrix technics (see cartography defined in the
!!    algorithm documentation)
subroutine AC_velocity_determine_communication(direction, ind_group, gs, send_gap,  &
    & rece_gap, rece_gap_abs, rece_rank, cartography, max_size)
! XXX Work only for periodic condition.

    use cart_topology   ! info about mesh and mpi topology
    use advec_variables ! contains info about solver parameters and others.
    use mpi


    ! Input/Ouput
    integer, intent(in)                                 :: direction
    integer, dimension(2), intent(in)                   :: ind_group
    integer, dimension(2), intent(in)                   :: gs
    integer, dimension(gs(1), gs(2), 2), intent(in)     :: rece_gap
    integer, dimension(2), intent(in)                   :: rece_gap_abs ! min (resp max) value of rece_gap(:,:,i) with i=1 (resp 2)
    integer,dimension(rece_gap_abs(1):rece_gap_abs(2))&
        &, intent(out)                                  :: rece_rank    ! rank of processus wich send me information
    integer, dimension(2, 2), intent(out)               :: send_gap
    type(int_1D_pter), dimension( &
        & rece_gap_abs(1):rece_gap_abs(2)), intent(out) :: cartography
    integer, intent(in)                                 :: max_size
    ! Others
    integer                             :: proc_gap         ! gap between a processus coordinate (along the current
                                                            ! direction) into the mpi-topology and my coordinate
    integer, dimension(gs(1), gs(2))    :: rece_gapP        ! gap between the coordinate of the previous processus (in the current direction)
                                                            ! and the processes of maximal coordinate which will receive information from it
    integer, dimension(gs(1), gs(2))    :: rece_gapN        ! same as above but for the next processus
    integer                             :: rankP            ! processus rank for shift (P= previous, N = next)
    integer                             :: send_request_gh  ! mpi status of noindicelocking send
    integer                             :: send_request_gh2 ! mpi status of noindicelocking send
    integer                             :: ierr             ! mpi error code
    integer, dimension(2)               :: tag_table        ! some mpi message tag
    logical, dimension(:,:), allocatable:: test_request     ! for mpi non blocking communication
    integer, dimension(:,:), allocatable:: send_request     ! for mpi non blocking send
    integer                             :: ind1, ind2       ! indice of the current line inside the group
    integer,dimension(2)                :: rece_buffer      ! buffer for reception of rece_max
    integer, dimension(:,:), allocatable:: first, last      ! Storage processus to which I will be the first (or the last) to receive
    integer                             :: min_size         ! begin indice in first and last to stock indice along first dimension of the group line
    integer                             :: gp_size          ! group size
    logical                             :: begin_interval   ! ware we in the start of an interval ?
    logical                             :: not_myself       ! Is the target processus myself ?
    integer, dimension(MPI_STATUS_SIZE) :: statut

    send_gap(1,1) = 3*N(direction)
    send_gap(1,2) = -3*N(direction)
    send_gap(2,:) = 0
    gp_size = gs(1)*gs(2)

    ! ===== Communicate with my neigbors -> obtain ghost ! ====
    ! Inform that about processus from which I need information
    tag_table = compute_tag(ind_group, tag_obtrec_ghost_NP, direction)
    call mpi_ISsend(rece_gap(1,1,1), gp_size, MPI_INTEGER, neighbors(direction,1), tag_table(1), &
        & D_comm(direction), send_request_gh, ierr)
    call mpi_ISsend(rece_gap(1,1,2), gp_size, MPI_INTEGER, neighbors(direction,2), tag_table(2), &
        & D_comm(direction), send_request_gh2, ierr)
    ! Receive the same message form my neighbors
    call mpi_recv(rece_gapN(1,1), gp_size, MPI_INTEGER, neighbors(direction,2), tag_table(1), D_comm(direction), statut, ierr)
    call mpi_recv(rece_gapP(1,1), gp_size, MPI_INTEGER, neighbors(direction,1), tag_table(2), D_comm(direction), statut, ierr)

    ! ===== Compute if I am first or last =====
    min_size = 2 + gs(2)
    ! Initialize first and last to determine if I am the the first or the last processes (considering the current direction)
        ! to require information from this processus
    allocate(first(2,rece_gap_abs(1):rece_gap_abs(2)))
    first(2,:) = 0  ! number of lines for which I am the first
    allocate(last(2,rece_gap_abs(1):rece_gap_abs(2)))
    last(2,:) = 0   ! number of lines for which I am the last
    do proc_gap = rece_gap_abs(1), rece_gap_abs(2)
      first(1,proc_gap) = -proc_gap
      last(1,proc_gap) = -proc_gap
    end do
    ! Compute
    do ind2 = 1, gs(2)
      do ind1 = 1, gs(1)
        ! Proc_gap must belongs to [rece_gap(ind1,ind2,1);rece_gap(ind1,ind2,2)].
        ! I am the first if (proc_gap>rece_gapP(ind1,ind2)-1).
        do proc_gap = rece_gapP(ind1,ind2), rece_gap(ind1,ind2,2)
          first(2,proc_gap) =  first(2,proc_gap)+1
        end do
        ! Proc_gap must belongs to [rece_gap(ind1,ind2,1);rece_gap(ind1,ind2,2)].
        ! I am the last if (proc_gap<rece_gapN(ind1,ind2)+1).
        do proc_gap = rece_gap(ind1,ind2,1), rece_gapN(ind1,ind2)
          last(2,proc_gap) =  last(2,proc_gap)+1
        end do
      end do
    end do

    ! ===== Determine the cartography =====
    do proc_gap = rece_gap_abs(1), rece_gap_abs(2)
      call mpi_cart_shift(D_comm(direction), 0, proc_gap, rankP, rece_rank(proc_gap), ierr)
      !not_myself = (rece_rank(proc_gap) /= D_rank(direction)) ! Is the target processus myself ?
      ! Nothing to do if target processus is myself
      !if (not_myself) then
      if (rece_rank(proc_gap) /= D_rank(direction)) then
        ! Allocate cartography(proc_gap)
        allocate(cartography(proc_gap)%pter(max_size))
        ! Initialize cartography
        cartography(proc_gap)%pter(1) = 0         ! number of velocity values to receive
        cartography(proc_gap)%pter(2) = min_size  ! number of element to send when sending cartography
        ! Update cartography
        do ind2 = 1, gs(2)
          cartography(proc_gap)%pter(2+ind2) = 0    ! 2 x number of interval of concern line into the column i2
          begin_interval = .true.
          do ind1 = 1, gs(1)
            ! Does proc_gap belongs to [rece_gap(i1,i2,1);rece_gap(i1,i2,2)]?
            if((proc_gap>=rece_gap(ind1,ind2,1)).and.(proc_gap<=rece_gap(ind1,ind2,2))) then
              if (begin_interval) then
                cartography(proc_gap)%pter(2+ind2) =  cartography(proc_gap)%pter(2+ind2)+2
                cartography(proc_gap)%pter(cartography(proc_gap)%pter(2)+1) = ind1
                cartography(proc_gap)%pter(2) = cartography(proc_gap)%pter(2) + 2
                cartography(proc_gap)%pter(cartography(proc_gap)%pter(2)) = ind1
                begin_interval = .false.
              else    ! begin_interval
                cartography(proc_gap)%pter(cartography(proc_gap)%pter(2)) = ind1
              end if  ! begin_interval
            else
              begin_interval = .true.
            end if ! rec_gap(1) <= proc_gap <= rece_gap(2)
          end do ! ind1
        end do ! ind2
      else   ! not myself
        allocate(cartography(proc_gap)%pter(1))
        cartography(proc_gap)%pter(1)=0
      end if ! not myself
    end do ! proc_gap

    ! ===== Free Isend buffer from first communication =====
    call MPI_WAIT(send_request_gh,statut,ierr)
    call MPI_WAIT(send_request_gh2,statut,ierr)

    ! ===== Send information about first and last  =====
    tag_table = compute_tag(ind_group, tag_obtrec_NP, direction)
    allocate(send_request(rece_gap_abs(1):rece_gap_abs(2),2))
    allocate(test_request(rece_gap_abs(1):rece_gap_abs(2),2))
    test_request = .false.
    do proc_gap = rece_gap_abs(1), rece_gap_abs(2)
        ! I am the first ?
        if (first(2,proc_gap)>0) then
            if(rece_rank(proc_gap)/= D_rank(direction)) then
                call mpi_ISsend(first(1,proc_gap), 2, MPI_INTEGER, rece_rank(proc_gap), tag_table(1), D_comm(direction), &
                        & send_request(proc_gap,1), ierr)
                test_request(proc_gap,1) = .true.
            else
                send_gap(1,1) = min(send_gap(1,1), -proc_gap)
                send_gap(2,1) = send_gap(2,1) + first(2,proc_gap)
            end if
        end if
        ! I am the last ?
        if (last(2,proc_gap)>0) then
            if(rece_rank(proc_gap)/= D_rank(direction)) then
                call mpi_ISsend(last(1,proc_gap), 2, MPI_INTEGER, rece_rank(proc_gap), tag_table(2), D_comm(direction), &
                        & send_request(proc_gap,2), ierr)
                test_request(proc_gap,2) = .true.
            else
                send_gap(1,2) = max(send_gap(1,2), -proc_gap)
                send_gap(2,2) = send_gap(2,2) + last(2,proc_gap)
            end if
        end if
    end do



    ! ===== Receive information form the first and the last processus which need a part of my local velocity field =====
    do while(send_gap(2,1) < gp_size)
        call mpi_recv(rece_buffer(1), 2, MPI_INTEGER, MPI_ANY_SOURCE, tag_table(1), D_comm(direction), statut, ierr)
        send_gap(1,1) = min(send_gap(1,1), rece_buffer(1))
        send_gap(2,1) = send_gap(2,1) + rece_buffer(2)
    end do
    do while(send_gap(2,2) < gp_size)
        call mpi_recv(rece_buffer(1), 2, MPI_INTEGER, MPI_ANY_SOURCE, tag_table(2), D_comm(direction), statut, ierr)
        send_gap(1,2) = max(send_gap(1,2), rece_buffer(1))
        send_gap(2,2) = send_gap(2,2) + rece_buffer(2)
    end do

    ! ===== Free Isend buffer =====
    do proc_gap = rece_gap_abs(1), rece_gap_abs(2)
        if (test_request(proc_gap,1).eqv. .true.) call MPI_WAIT(send_request(proc_gap,1),statut,ierr)
        if (test_request(proc_gap,2)) call MPI_WAIT(send_request(proc_gap,2),statut,ierr)
    end do
    deallocate(send_request)
    deallocate(test_request)

    ! ===== Deallocate array =====
    deallocate(first)
    deallocate(last)

end subroutine AC_velocity_determine_communication


!> Interpolate the velocity field used in a RK2 scheme for particle advection -
!! version for direction with no domain subdivision ands thus no required
!! communications
!!    @param[in]        dt          = time step
!!    @param[in]        direction   = current direction (1 = along X, 2 = along Y and 3 = along Z)
!!    @param[in]        gs          = size of a group (ie number of line it gathers along the two other directions)
!!    @param[in]        p_pos_adim  = adimensionned particle postion
!!    @param[in,out]    p_V         = particle velocity (along the current direction)
!! @details
!!    A RK2 scheme is used to advect the particles : the midlle point scheme. An
!!    intermediary position "p_pos_bis(i) = p_pos(i) + V(i)*dt/2" is computed and then 
!!    the numerical velocity of each particles is computed as the interpolation of V  in 
!!    this point. This field is used to advect the particles at the seconde order in time : 
!!    p_pos(t+dt, i) = p_pos(i) + p_V(i).
!!    Variant for cases with no required communication.
subroutine AC_velocity_interpol_no_com(dt, direction, gs, p_pos_adim, p_V)

    ! This code involve a recopy of p_V. It is possible to directly use the 3D velocity field but it will also limit the meroy access.

    use cart_topology   ! info about mesh and mpi topology
    use advec_variables ! contains info about solver parameters and others.

    ! Input/Ouput
    real(WP), intent(in)                            :: dt           ! time step
    integer, intent(in)                             :: direction    ! current direction
    integer, dimension(2),intent(in)                :: gs           ! groupe size
    real(WP), dimension(:,:,:), intent(in)          :: p_pos_adim
    real(WP), dimension(:,:,:), intent(inout)       :: p_V
    ! Others, local
    real(WP),dimension(N_proc(direction),gs(1),gs(2))   :: p_posV_bis   ! adimensionned position of the middle point
    integer                                             :: ind          ! indices
    integer                                             :: i1, i2       ! indices in the lines group
    integer                                             :: pos          ! indices of the mesh point wich preceed the particle position

    ! ===== Initialisation =====
    ! Compute the midlle point
    p_posV_bis = p_pos_adim + (dt/2.0)*p_V/d_sc(direction)


    ! ===== Compute the interpolated velocity =====
    ! -- Compute the interpolation weight and update the velocity directly in p_posV_bis --
    do i2 = 1, gs(2)
        do i1 = 1, gs(1)
            do ind = 1, N(direction)

            pos = floor(p_posV_bis(ind,i1,i2))
            p_posV_bis(ind,i1,i2) = p_V(modulo(pos-1,N(direction))+1,i1,i2) + (p_posV_bis(ind,i1,i2)-pos)* &
                & (p_V(modulo(pos,N(direction))+1,i1,i2)-p_V(modulo(pos-1,N(direction))+1,i1,i2))

            end do ! loop on particle indice (ind)
        end do ! loop on first coordinate (i1) of a line inside the block of line
    end do ! loop on second coordinate (i2) of a line inside the block of line

    ! -- Recopy velocity in the right array --
    p_V = p_posV_bis

end subroutine AC_velocity_interpol_no_com


end module advec_common_velo
!> @}
